/*
 *
 * B2IVRInterface.cxx
 * 
 * Open SIP Stack ( OSS )
 *
 * Copyright (C) 2006 Joegen E. Baclor. All Rights Reserved.
 *
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is OpenSIPStack Library.
 *
 * The Initial Developer of the Original Code is Joegen E. Baclor.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * $Log: B2BIVRInterface.cxx,v $
 * Revision 1.8  2007/07/14 07:50:29  joegenbaclor
 * Migrated code to use new STL based string class
 *
 * Revision 1.7  2007/07/03 17:17:37  joegenbaclor
 * Got rid of compile warnings in unix builds
 *
 * Revision 1.6  2007/06/15 09:12:14  joegenbaclor
 * Added IVR Transfer reject event
 *
 * Revision 1.5  2007/06/13 09:22:05  joegenbaclor
 * Implemented basic IVR handler in OpenSBC
 *
 * Revision 1.4  2007/06/12 09:22:41  joegenbaclor
 * Added IVR Call Transfer
 *
 * Revision 1.3  2007/06/11 23:51:48  joegenbaclor
 * Added event queue
 *
 * Revision 1.2  2007/06/08 15:10:30  joegenbaclor
 * Added IVR DTMF Grammar
 *
 * Revision 1.1  2007/06/08 10:32:34  joegenbaclor
 * Initial upload
 *
 *
 */

#include "B2BIVRInterface.h"
#include "B2BUAConnection.h"

#define new PNEW

using namespace B2BUA;

B2BIVRInterface::B2BIVREvent::B2BIVREvent( 
  B2BUAConnection & connection,
  Type type 
)
{
  m_Connection = GCCREATEREF( (&connection), "B2BIVRInterface::B2BIVREvent", "B2BUAConnection" );
  m_Type = type;
}

////////////////////////////////////////////////////////////

B2BIVRInterface::B2BIVRInterface(  
  B2BUserAgent & b2bua 
)  : B2BInterface( b2bua )
{
  m_EventQueue = new EventQueue( PCREATE_NOTIFIER( _OnIVREvent ), 1, 1024 * 2, P_MAX_INDEX, "B2BIVRInterface" );
}

B2BIVRInterface::~B2BIVRInterface()
{
  delete m_EventQueue;
}

BOOL B2BIVRInterface::IVROnOpenChannel(
  B2BUAConnection & connection
)
{
  //connection.IVRPlayFile( "greeting", "welcome.wav" );
  //connection.IVRSetDTMFGrammar( "tel-no", 5, '#' );
  EnqueueEvent( new OnOpenChannel( connection ) );
  return TRUE;
}

void B2BIVRInterface::IVROnEndPlaying(
  B2BUAConnection & connection,
  const OString & identifier
)
{
  EnqueueEvent( new OnEndPlaying(  connection, identifier ) );
}

BOOL B2BIVRInterface::IVROnUserInput(
  B2BUAConnection & connection,
  const OString & str 
)
{
  EnqueueEvent( new OnUserInput( connection, str ));
  return TRUE;
}

void B2BIVRInterface::IVROnEndCollection( 
  B2BUAConnection & connection,
  const OString &grammarId, 
  const OString &grammarBuffer 
)
{
  //PTRACE( 1, "Collected: " << grammarId << " : " << grammarBuffer );
  EnqueueEvent( new OnEndCollection( connection, grammarId, grammarBuffer ) );
}

void B2BIVRInterface::IVROnTransferReject(
  B2BUAConnection & connection,
  const SIPMessage & reject
)
{
  EnqueueEvent( new OnTransferReject( connection, reject ) );
}

void B2BIVRInterface::EnqueueEvent( B2BIVREvent * data )
{
  PWaitAndSignal lock( m_EventQueueMutex );
  m_EventQueue->EnqueueEvent( static_cast<PObject*>(data) );
}

void B2BIVRInterface::_OnIVREvent( EventQueue &, INT data )
{
  if( data == 0 )
    return;

  B2BIVREvent * eventObject = (B2BIVREvent*)data;
  SIPSession::GCRef & ref = eventObject->GetConnection();
  B2BUAConnection * conn = dynamic_cast<B2BUAConnection *>( ref.GetObject() );
  OnIVREvent( conn, eventObject );

  delete eventObject;
}

void B2BIVRInterface::OnIVREvent(
  B2BUAConnection * /*conn*/,
  const B2BIVREvent * /*evt*/
)
{
}

