/*
 *
 * CheckSumMonitor.cxx
 * 
 * Open SIP Stack ( OSS )
 *
 * Copyright (C) 2006 Joegen E. Baclor. All Rights Reserved.
 *
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is OpenSIPStack Library.
 *
 * The Initial Developer of the Original Code is Joegen E. Baclor.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * $Log: CheckSumMonitor.cxx,v $
 * Revision 1.1  2007/03/29 07:46:09  joegenbaclor
 * Added CheckSumMonitor
 *
 *
 */


#include "CheckSumMonitor.h"
#include "ParserTools.h"

#define new PNEW

using namespace SIPParser;
using namespace Tools;

CheckSumMonitor::CheckSumMonitor( 
  const PFilePath & fp
)
{
  m_HasNotifier = FALSE;
  m_Path = fp;
  PFile file;
  if( file.Open( m_Path, PFile::ReadOnly ) )
  {
    PBYTEArray content( file.GetLength() );
    if( file.Read( content.GetPointer( file.GetLength() ), file.GetLength() ) )
    {
      PMessageDigest5::Code checkSum;
      PMessageDigest5 digestor;
      digestor.Start();
      digestor.Process( content.GetPointer( file.GetLength() ), file.GetLength() );
      digestor.Complete( checkSum );
      m_CheckSum = ParserTools::AsHex( checkSum );

      PThread::Create( PCREATE_NOTIFIER( Main ) );
    }
  }
}

CheckSumMonitor::~CheckSumMonitor()
{
  m_ExitSync.Signal();
}

void CheckSumMonitor::Main( PThread &, INT )
{
  while( !m_ExitSync.Wait( 1000 ) )
  {
    PFile file;
    if( file.Open( m_Path, PFile::ReadOnly ) )
    {
      PBYTEArray content( file.GetLength() );
      if( file.Read( content.GetPointer( file.GetLength() ), file.GetLength() ) )
      {
        PMessageDigest5::Code checkSum;
        PMessageDigest5 digestor;
        digestor.Start();
        digestor.Process( content.GetPointer( file.GetLength() ), file.GetLength() );
        digestor.Complete( checkSum );
        PString current = ParserTools::AsHex( checkSum );
        
        if( current != m_CheckSum )
        {
          OnFileChanged();
          m_CheckSum = current;
        }
      }
    }
  }
}

void CheckSumMonitor::OnFileChanged()
{
  if( m_HasNotifier )
    m_Notifier( *this, 0 );
}

