
/*
 *
 * Encryption.cxx
 * 
 * Open SIP Stack ( OSS )
 *
 * Copyright (C) 2006 Joegen E. Baclor. All Rights Reserved.
 *
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is OpenSIPStack Library.
 *
 * The Initial Developer of the Original Code is Joegen E. Baclor.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * $Log: Encryption.cxx,v $
 * Revision 1.4  2007/07/14 07:50:32  joegenbaclor
 * Migrated code to use new STL based string class
 *
 * Revision 1.3  2007/04/11 09:10:36  joegenbaclor
 * Corrected some bugs in encryption
 *
 * Revision 1.1  2007/04/10 06:13:08  joegenbaclor
 * Added new Encryption Mode
 *
 *
 */

#include "Encryption.h"

using namespace Encryption;

PBYTEArray Engine::XOR::m_Key;
PBYTEArray Engine::DiffShift::m_Key;
OString Engine::m_DefKey = "opensipstack.org";

Engine::Type Engine::m_Type = Engine::Type_XOR;
OString Engine::m_Key = Engine::m_DefKey;

BOOL Engine::Encrypt(
  const PBYTEArray & in,
  PBYTEArray & out
)
{
  int keyLen = Engine::m_Key.GetLength();
  PBYTEArray key;
  key.SetSize( keyLen );
  ::memcpy( key.GetPointer( keyLen ), Engine::m_Key.GetPointer( keyLen ), keyLen );
  if( Engine::m_Type == Engine::Type_XOR )
  {
    return Engine::XOR::Encrypt( in, out, key );
  }else if( Engine::m_Type == Engine::Type_DiffShift )
  {
    return Engine::DiffShift::Encrypt( in, out, key );
  }

  return FALSE;
}

BOOL Engine::Encrypt(
  const PBYTEArray & in,
  OString & out
)
{
  int keyLen = Engine::m_Key.GetLength();
  PBYTEArray key;
  key.SetSize( keyLen );
  ::memcpy( key.GetPointer( keyLen ), Engine::m_Key.GetPointer( keyLen ), keyLen );
  if( Engine::m_Type == Engine::Type_XOR )
  {
    return Engine::XOR::Encrypt( in, out, key );
  }else if( Engine::m_Type == Engine::Type_DiffShift )
  {
    return Engine::DiffShift::Encrypt( in, out, key );
  }

  return FALSE;
}

BOOL Engine::Decrypt(
  const PBYTEArray & in,
  PBYTEArray & out
)
{
  int keyLen = Engine::m_Key.GetLength();
  PBYTEArray key;
  key.SetSize( keyLen );
  ::memcpy( key.GetPointer( keyLen ), Engine::m_Key.GetPointer( keyLen ), keyLen );
  if( Engine::m_Type == Engine::Type_XOR )
  {
    return Engine::XOR::Encrypt( in, out, key );
  }else if( Engine::m_Type == Engine::Type_DiffShift )
  {
    return Engine::DiffShift::Decrypt( in, out, key );
  }

  return FALSE;
}

BOOL Engine::Decrypt(
  const PBYTEArray & in,
  OString & out
)
{
  int keyLen = Engine::m_Key.GetLength();
  PBYTEArray key;
  key.SetSize( keyLen );
  ::memcpy( key.GetPointer( keyLen ), Engine::m_Key.GetPointer( keyLen ), keyLen );
  if( Engine::m_Type == Engine::Type_XOR )
  {
    return Engine::XOR::Encrypt( in, out, key );
  }else if( Engine::m_Type == Engine::Type_DiffShift )
  {
    return Engine::DiffShift::Decrypt( in, out, key );
  }

  return FALSE;
}

BOOL Engine::XOR::Encrypt(
  const PBYTEArray & in,
  PBYTEArray & out,
  const PBYTEArray & _key
)
{
  if( _key.GetSize() < 2 )
    return FALSE;

  PBYTEArray key = _key;
  if( PBYTE_ORDER == PBIG_ENDIAN ) 
  {
    key[0] = _key[1];
    key[1] = _key[0];
  }

  PINDEX j = 0;
  ///out.SetSize(0);

  int size = in.GetSize();
  for( PINDEX i = 0; i < size; i+=2 )
  {
    
      out[j++] = in[i] ^ key[0];
      out[j++] = in[i+1] ^ key[1];

      if( i + 2 == in.GetSize() - 1 )
      {
        out[j++] = in[i+2] ^ key[1];
        break;
      }
  }

  return out.GetSize() != 0;
}

BOOL Engine::XOR::Encrypt(
  const PBYTEArray & in,
  OString & out,
  const PBYTEArray & _key
)
{
  if( _key.GetSize() < 2 )
    return FALSE;

  PBYTEArray key = _key;
  if( PBYTE_ORDER == PBIG_ENDIAN ) 
  {
    key[0] = _key[1];
    key[1] = _key[0];
  }

  //out.SetSize(0);

  int size = in.GetSize();
  for( PINDEX i = 0; i < size; i+=2 )
  {
    char c;
    c = static_cast<char>(in[i] ^ key[0]);
    if( (c >= 32 && c <= 126 ) || c == 10 || c == 13 ) 
      out+= c;
    
    c = static_cast<char>(in[i+1] ^ key[1]);
    if( (c >= 32 && c <= 126 ) || c == 10 || c == 13 ) 
      out+= c;

    if( i + 2 == in.GetSize() - 1 )
    {
      c = static_cast<char>(in[i+2] ^ key[1]);
      if( (c >= 32 && c <= 126 ) || c == 10 || c == 13 ) 
        out+=c;
      break;
    }
  }

  return out.GetLength() != 0;
}

BOOL Engine::XOR::Encrypt(
  const PBYTEArray & in,
  PBYTEArray & out
)
{
  static PBYTEArray key;
  key[0] = Engine::m_DefKey[0];
  key[1] = Engine::m_DefKey[1];

  return Encrypt( in, out, m_Key.GetSize() == 0 ? key : m_Key );
}

BOOL Engine::XOR::Encrypt(
  const PBYTEArray & in,
  OString & out
)
{
  static PBYTEArray key;
  key[0] = Engine::m_DefKey[0];
  key[1] = Engine::m_DefKey[1];

  return Encrypt( in, out, m_Key.GetSize() == 0 ? key : m_Key );
}


////////////////////////////////////////////////////////

BOOL Engine::DiffShift::Encrypt(
  const PBYTEArray & in,
  PBYTEArray & out,
  const PBYTEArray & key
)
{
	int i = 0;
	int j = 0;
  int leng = in.GetSize();
  int kleng = key.GetSize();

	for( i = 0, j = 0; i < leng; ++i)
	{
		out[i]=in[i]+(key[j]-'0');
		++j;
		if(j == kleng)
			j = 0;
	}

  return TRUE;
}

BOOL Engine::DiffShift::Encrypt(
  const PBYTEArray & in,
  OString & out,
  const PBYTEArray & key
)
{
  int i = 0;
	int j = 0;
  int leng = in.GetSize();
  int kleng = key.GetSize();

	for( i = 0, j = 0; i < leng; ++i)
	{
          char c = static_cast<char>(in[i]+(key[j]-'0'));
          if( (c >= 32 && c <= 126 ) || c == 10 || c == 13 ) 
		out+=c;
		++j;
		if(j == kleng)
			j = 0;
	}

  return TRUE;
}

BOOL Engine::DiffShift::Decrypt(
  const PBYTEArray & in,
  PBYTEArray & out,
  const PBYTEArray & key
)
{
  int i = 0;
	int j = 0;
  int leng = in.GetSize();
  int kleng = key.GetSize();

	for( i = 0, j = 0; i < leng; ++i)
	{
		out[i]=in[i]-(key[j]-'0');
		++j;
		if(j == kleng)
			j = 0;
	}

  return TRUE;
}

BOOL Engine::DiffShift::Decrypt(
  const PBYTEArray & in,
  OString & out,
  const PBYTEArray & key
)
{
  int i = 0;
	int j = 0;
  int leng = in.GetSize();
  int kleng = key.GetSize();

	for( i = 0, j = 0; i < leng; ++i)
	{
		char c = static_cast<char>(in[i]-(key[j]-'0'));
                if( (c >= 32 && c <= 126 ) || c == 10 || c == 13 ) 
                  out += c;
		++j;
		if(j == kleng)
			j = 0;
	}

  return TRUE;
}

BOOL Engine::DiffShift::Encrypt(
  const PBYTEArray & in,
  PBYTEArray & out
)
{
  static PBYTEArray key;
  key[0] = Engine::m_DefKey[0];
  key[1] = Engine::m_DefKey[1];

  return Encrypt( in, out, m_Key.GetSize() == 0 ? key : m_Key );
}

BOOL Engine::DiffShift::Encrypt(
  const PBYTEArray & in,
  OString & out
)
{
  static PBYTEArray key;
  key[0] = Engine::m_DefKey[0];
  key[1] = Engine::m_DefKey[1];

  return Encrypt( in, out, m_Key.GetSize() == 0 ? key : m_Key );
}

BOOL Engine::DiffShift::Decrypt(
  const PBYTEArray & in,
  PBYTEArray & out
)
{
  static PBYTEArray key;
  key[0] = Engine::m_DefKey[0];
  key[1] = Engine::m_DefKey[1];

  return Decrypt( in, out, m_Key.GetSize() == 0 ? key : m_Key );
}

BOOL Engine::DiffShift::Decrypt(
  const PBYTEArray & in,
  OString & out
)
{
  static PBYTEArray key;
  key[0] = Engine::m_DefKey[0];
  key[1] = Engine::m_DefKey[1];

  return Decrypt( in, out, m_Key.GetSize() == 0 ? key : m_Key );
}





