/*
 *
 * ResourceCounter.cxx
 * 
 * Open SIP Stack ( OSS )
 *
 * Copyright (C) 2006 Joegen E. Baclor. All Rights Reserved.
 *
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is OpenSIPStack Library.
 *
 * The Initial Developer of the Original Code is Joegen E. Baclor.
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * $Log: ResourceCounter.cxx,v $
 * Revision 1.7  2007/07/25 16:11:52  joegenbaclor
 * Raised m_RC_MaxEventQueue from 10 to 30
 *
 * Revision 1.6  2007/07/15 13:32:21  joegenbaclor
 * implemented Print method
 *
 * Revision 1.5  2007/07/15 07:50:04  joegenbaclor
 * tidying up resource counters
 *
 * Revision 1.4  2007/07/14 09:49:48  joegenbaclor
 * Corrected compile errors in linux
 *
 * Revision 1.3  2007/07/10 03:23:05  joegenbaclor
 * Added experimental code to use STL tokenizer instead of internal string tokenizer
 *
 * Revision 1.2  2007/07/09 03:53:53  joegenbaclor
 * Added ResourceCounter to Tools namespace
 * Updated VC8 project files
 *
 * Revision 1.1  2007/07/09 03:25:25  joegenbaclor
 * Added resource counter class
 *
 *
 */
 
#include "ResourceCounter.h"

using namespace Tools;
#define WORD_MAX 65535

// Object counters
PAtomicInteger ResourceCounter::m_RC_ICT = 0;
PAtomicInteger ResourceCounter::m_RC_NICT = 0;
PAtomicInteger ResourceCounter::m_RC_IST = 0;
PAtomicInteger ResourceCounter::m_RC_NIST = 0;
PAtomicInteger ResourceCounter::m_RC_CallSession = 0;
PAtomicInteger ResourceCounter::m_RC_Connection = 0;
PAtomicInteger ResourceCounter::m_RC_Registration = 0;
PAtomicInteger ResourceCounter::m_RC_RTPSession = 0;
PAtomicInteger ResourceCounter::m_RC_EventQueue = 0;
PAtomicInteger ResourceCounter::m_RC_Cache = 0;
PAtomicInteger ResourceCounter::m_RC_Garbage = 0;



PAtomicInteger ResourceCounter::m_RC_MaxConnection     = 1000;
PAtomicInteger ResourceCounter::m_RC_MaxCallSession    = ResourceCounter::m_RC_MaxConnection * 3;
PAtomicInteger ResourceCounter::m_RC_MaxRegistration   = 1000;
PAtomicInteger ResourceCounter::m_RC_MaxRTPSession     = 600;
PAtomicInteger ResourceCounter::m_RC_MaxEventQueue     = 30;
PAtomicInteger ResourceCounter::m_RC_MaxCache          = WORD_MAX;
PAtomicInteger ResourceCounter::m_RC_MaxGarbage        = WORD_MAX;

PAtomicInteger ResourceCounter::m_RC_MaxIST            = ResourceCounter::m_RC_MaxCallSession / 2;
PAtomicInteger ResourceCounter::m_RC_MaxICT            = WORD_MAX;
PAtomicInteger ResourceCounter::m_RC_MaxNICT           = WORD_MAX;
PAtomicInteger ResourceCounter::m_RC_MaxNIST           = WORD_MAX;

// high value
PAtomicInteger ResourceCounter::m_RC_HighICT = 0;
PAtomicInteger ResourceCounter::m_RC_HighNICT = 0;
PAtomicInteger ResourceCounter::m_RC_HighIST = 0;
PAtomicInteger ResourceCounter::m_RC_HighNIST = 0;
PAtomicInteger ResourceCounter::m_RC_HighCallSession = 0;
PAtomicInteger ResourceCounter::m_RC_HighConnection = 0;
PAtomicInteger ResourceCounter::m_RC_HighRegistration = 0;
PAtomicInteger ResourceCounter::m_RC_HighRTPSession = 0;
PAtomicInteger ResourceCounter::m_RC_HighEventQueue = 0;
PAtomicInteger ResourceCounter::m_RC_HighCache = 0;
PAtomicInteger ResourceCounter::m_RC_HighGarbage = 0;

int ResourceCounter::IncrementICT()
{
  ++m_RC_ICT;
  if( m_RC_ICT > m_RC_HighICT )
    m_RC_HighICT = m_RC_ICT;
  return m_RC_ICT;
}

int ResourceCounter::DecrementICT()
{
  return --m_RC_ICT;
}

int ResourceCounter::IncrementNICT()
{
  ++m_RC_NICT;
  if( m_RC_NICT > m_RC_HighNICT )
    m_RC_HighNICT = m_RC_NICT;
  return m_RC_NICT;
}

int ResourceCounter::DecrementNICT()
{
  return --m_RC_NICT;
}

int ResourceCounter::IncrementIST()
{
  ++m_RC_IST;
  if( m_RC_IST > m_RC_HighIST )
    m_RC_HighIST = m_RC_IST;
  return m_RC_IST;
}

int ResourceCounter::DecrementIST()
{
  return --m_RC_IST;
}

int ResourceCounter::IncrementNIST()
{
  ++m_RC_NIST;
  if( m_RC_NIST > m_RC_HighNIST )
    m_RC_HighNIST = m_RC_NIST;
  return m_RC_NIST;
}

int ResourceCounter::DecrementNIST()
{
  return --m_RC_NIST;
}

int ResourceCounter::IncrementCallSession()
{
  ++m_RC_CallSession;
  if( m_RC_CallSession > m_RC_HighCallSession )
    m_RC_HighCallSession = m_RC_CallSession;
  return m_RC_CallSession;
}

int ResourceCounter::DecrementCallSession()
{
  return --m_RC_CallSession;
}

int ResourceCounter::IncrementConnection()
{
  ++m_RC_Connection;
  if( m_RC_Connection > m_RC_HighConnection )
    m_RC_HighConnection = m_RC_Connection;
  return m_RC_Connection;
}

int ResourceCounter::DecrementConnection()
{
  return --m_RC_Connection;
}

int ResourceCounter::IncrementRegistration()
{
  ++m_RC_Registration;
  if( m_RC_Registration > m_RC_HighRegistration )
    m_RC_HighRegistration = m_RC_Registration;
  return m_RC_Registration;
}

int ResourceCounter::DecrementRegistration()
{
  return --m_RC_Registration;
}

int ResourceCounter::IncrementRTPSession()
{
  ++m_RC_RTPSession;
  if( m_RC_RTPSession > m_RC_HighRTPSession )
    m_RC_HighRTPSession = m_RC_RTPSession;
  return m_RC_RTPSession;
}

int ResourceCounter::DecrementRTPSession()
{
  return --m_RC_RTPSession;
}

int ResourceCounter::IncrementEventQueue()
{
  ++m_RC_EventQueue;
  if( m_RC_EventQueue > m_RC_HighEventQueue )
    m_RC_HighEventQueue = m_RC_EventQueue;
  return m_RC_EventQueue;
}

int ResourceCounter::DecrementEventQueue()
{
  return --m_RC_EventQueue;
}

int ResourceCounter::IncrementCache()
{
  ++m_RC_Cache;
  if( m_RC_Cache > m_RC_HighCache )
    m_RC_HighCache = m_RC_Cache;
  return m_RC_Cache;
}

int ResourceCounter::DecrementCache()
{
  return --m_RC_Cache;
}

int ResourceCounter::IncrementGarbage()
{
  ++m_RC_Garbage;
  if( m_RC_Garbage > m_RC_HighGarbage )
    m_RC_HighGarbage = m_RC_Garbage;
  return m_RC_Garbage;
}

int ResourceCounter::DecrementGarbage()
{
  return --m_RC_Garbage;
}

BOOL ResourceCounter::IsICTMaxReached()
{
  return m_RC_ICT >= m_RC_MaxICT;
}

BOOL ResourceCounter::IsNICTMaxReached()
{
  return m_RC_NICT >= m_RC_MaxNICT;
}

BOOL ResourceCounter::IsISTMaxReached()
{
  return m_RC_IST >= m_RC_MaxIST;
}

BOOL ResourceCounter::IsNISTMaxReached()
{
  return m_RC_NIST >= m_RC_MaxNIST;
}

BOOL ResourceCounter::IsCallSessionMaxReached()
{
  return m_RC_CallSession >= m_RC_MaxCallSession;
}

BOOL ResourceCounter::IsConnectionMaxReached()
{
  return m_RC_Connection >= m_RC_MaxConnection;
}

BOOL ResourceCounter::IsRegistrationMaxReached()
{
  return m_RC_Registration >= m_RC_MaxRegistration;
}

BOOL ResourceCounter::IsRTPSessionMaxReached()
{
  return m_RC_RTPSession >= m_RC_MaxRTPSession;
}

BOOL ResourceCounter::IsEventQueueMaxReached()
{
  return m_RC_EventQueue >= m_RC_MaxEventQueue;
}

BOOL ResourceCounter::IsCacheMaxReached()
{
  return m_RC_Cache >= m_RC_MaxCache;
}

BOOL ResourceCounter::IsGarbageMaxReached()
{
  return m_RC_Garbage >= m_RC_MaxGarbage;
}

void ResourceCounter::Print( ostream & strm )
{
  strm << "ICT=" << ResourceCounter::m_RC_ICT << " ";
  strm << "NICT=" << ResourceCounter::m_RC_NICT << " ";
  strm << "IST=" << ResourceCounter::m_RC_IST << " ";
  strm << "NIST=" << ResourceCounter::m_RC_NIST << " ";
  strm << "CALL=" << ResourceCounter::m_RC_CallSession << " ";
  strm << "CONN=" << ResourceCounter::m_RC_Connection << " ";
  strm << "REG=" << ResourceCounter::m_RC_Registration << " ";
  strm << "RTP=" << ResourceCounter::m_RC_RTPSession << " ";
  strm << "QUEUE=" << ResourceCounter::m_RC_EventQueue << " ";
  strm << "CACHE=" << ResourceCounter::m_RC_Cache << " ";
  strm << "GC=" << ResourceCounter::m_RC_Garbage << " ";
}
