/*
 *
 * SIPMessage.cxx
 *
 * Open SIP Stack ( OSS )
 *
 * Copyright (C) 2006 Joegen E. Baclor. All Rights Reserved.
 *
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is OpenSIPStack Library.
 *
 * The Initial Developer of the Original Code is Joegen E. Baclor.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * $Log: SIPMessage.cxx,v $
 * Revision 1.89  2007/07/26 11:52:56  joegenbaclor
 * Finalized AuthenticationPending
 *
 * Revision 1.88  2007/07/26 08:15:53  joegenbaclor
 * Uploading support for AuthenticationPending state in B2BUAConnection.  This fix will eliminatethe need  to generate a diffrent call-id for resending credentials for 401/407
 *
 * Revision 1.87  2007/07/15 07:50:04  joegenbaclor
 * tidying up resource counters
 *
 * Revision 1.85  2007/07/14 07:50:33  joegenbaclor
 * Migrated code to use new STL based string class
 *
 * Revision 1.84  2007/07/06 13:34:11  rcolobong
 * Bug Fix CloneRecordRoute, ClonePath, ClonePAssertedIdentities, and ClonePPreferredIdentities where access violation occur
 * Bug Fix exeption bad_cast occur when getting PAssertedIdentity and PPreferredIdentity
 *
 * Revision 1.83  2007/06/28 15:46:56  joegenbaclor
 * Added IsReposne method
 * Made sure GetCSeqMEthod always return an upper case method
 *
 * Revision 1.82  2007/06/07 12:22:00  rcolobong
 * 1. Bug fix errors related to Consultative Transfer
 * 2. Update handling Replaces during Refer
 * 3. Add new method ConstructReplaces in SIPSession
 *
 * Revision 1.81  2007/05/18 01:24:44  joegenbaclor
 * 1.  Added separate useragent to handle B2BUpperReg backdoor
 * 2.  Modified routing for B2BUpperReg
 * 3.  Corrected bug in ParserTools::GetCompactHeader()
 *
 * Revision 1.80  2007/05/08 12:49:27  joegenbaclor
 * Corrected memory leaks in SIPMessage.
 *
 * Revision 1.79  2007/04/30 04:33:27  joegenbaclor
 * Fixed bug where via in comma delimited form is not parsed properly
 *
 * Revision 1.78  2007/04/16 06:23:32  joegenbaclor
 * Added IsAllowed Method
 *
 * Revision 1.77  2007/04/10 06:13:08  joegenbaclor
 * Added new Encryption Mode
 *
 * Revision 1.76  2007/03/21 05:28:26  joegenbaclor
 * disabled reference counting in SIPMessages
 *
 * Revision 1.75  2007/03/20 10:32:19  joegenbaclor
 * Made sure RemoveAllRoutes parses the route header first before deleting the list.
 *
 * Revision 1.74  2007/03/02 06:22:46  joegenbaclor
 * Corrected some bug in GC.
 * Added CPU Throttle mechanism during heavy load to avoid denial of service
 * Made sure lists are Made unique in SIPHeade clone function
 *
 * Revision 1.73  2007/02/22 06:07:32  joegenbaclor
 * Added check for m_ContactList==NULL in GetContactAt() function.  Thanks Gustavo Curetti.
 *
 * Revision 1.72  2007/02/19 08:49:00  joegenbaclor
 * More work on privacy.
 * 1.  Change PAssertedIdentity base from SIPHeaderF to Contact to allow displaynames
 * 2.  Removed unused variables in ProfileUA
 * 3.  Allowed Privacy to be enabled via Makecall
 * 4.  Improved SIP session privacy support
 *
 * Revision 1.71  2007/02/15 09:37:39  joegenbaclor
 * Added Privacy Header accessors to SIPMessage
 *
 * Revision 1.70  2007/02/13 05:28:03  joegenbaclor
 * Added GetTopRouteURI funtion
 *
 * Revision 1.69  2007/02/13 00:34:45  joegenbaclor
 * Removed trailing extra CRLF in message bodies as reported by Ingmar Shraub
 *
 * Revision 1.68  2007/02/08 12:32:54  joegenbaclor
 * Reverted last TCP changes.  As per SIP Implementors, it is a bad idea to use the TCP outbound socket in the reverse direction
 *
 * Revision 1.67  2007/02/08 07:21:26  joegenbaclor
 * Changed GetContactAt signature to avoid getting a NULL reference
 *
 * Revision 1.66  2007/02/07 05:14:11  joegenbaclor
 * more record route changes
 *
 * Revision 1.65  2007/02/07 04:04:59  joegenbaclor
 * Fixed record route bug
 *
 * Revision 1.64  2007/02/06 16:50:04  joegenbaclor
 * 1.  More work on TCP
 * 2.  Mutexed SIPHeadeF and SIPHeaderC collection members
 * 3.  Correct bug in B2BUACall record routing
 * 4.  Added Strict Router support to SIPSession::SendRequest
 *
 * Revision 1.63  2007/02/04 15:04:09  joegenbaclor
 * Completed transaction layer TCP support
 *
 * Revision 1.62  2007/01/27 23:06:15  joegenbaclor
 * Added RequestURI accessors
 *
 * Revision 1.61  2007/01/24 10:19:55  joegenbaclor
 * Work on TCP support has commenced.
 *
 * Revision 1.60  2006/11/30 08:19:38  joegenbaclor
 * Relicensed opensipstack to MPL/GPL/LGPL triple license scheme
 *
 * Revision 1.59  2006/11/23 09:30:48  joegenbaclor
 * Fixed bug in Route header parsing where the order is reversed by the SIPParser
 *
 * Revision 1.58  2006/11/22 03:16:18  joegenbaclor
 * Removed Record Routes from B2B Outbound invites
 *
 * Revision 1.57  2006/11/21 06:55:19  joegenbaclor
 * Added ability to add/append additional listeners.
 *
 * Revision 1.56  2006/11/18 03:14:28  joegenbaclor
 * Implemented session loop detection mechanism using iid via parameter
 *
 * Revision 1.55  2006/11/17 04:22:22  joegenbaclor
 * Added ability to remember core to use via m_UACoreName
 *
 * Revision 1.54  2006/11/10 08:10:51  rcolobong
 * Use ParserTools GetDefaultHashKey1 and GetDefaultHashKey2 in XORHash rather than using the MACRO GS_HASH_KEY_1/GS_HASH_KEY_2
 *
 * Revision 1.53  2006/11/08 06:46:20  joegenbaclor
 * Added session level NIC/Interface binding
 *
 * Revision 1.52  2006/11/07 09:30:46  joegenbaclor
 * Added new internal header "IFACE-ADDR" to allow transctions to remember the interface address for sending future responses within the transaction lifespan
 *
 * Revision 1.51  2006/11/02 22:53:46  joegenbaclor
 * Added Path header support (RFC 3327)
 *
 * Revision 1.50  2006/10/30 02:42:27  joegenbaclor
 * 1.  Various changes to support RFC 3680.
 * 2.  Removed Presence files.  (To be promoted as a contrib library for ATLSIP)
 *
 * Revision 1.49  2006/10/25 10:51:30  rcolobong
 * Fix bug where m_SupportedList might be Null in GetSupported()
 *
 * Revision 1.48  2006/10/23 15:44:58  joegenbaclor
 * 1.  Numerous bug fixes for REFER
 * 2.  Added new SIP Headers for SIP timers and Attended call transfer
 * 3.  Added loop sanity check to SIP UDP Transport
 * 4.  Corrected parser bug for ProxyRequire header
 *
 * Revision 1.47  2006/10/17 16:20:26  joegenbaclor
 * Fixed bug in parsing ProxyRequire header
 *
 * Revision 1.46  2006/10/17 00:12:24  joegenbaclor
 * Added support for RFC 4028 SIP Session Timers header
 *
 * Revision 1.45  2006/09/02 04:44:05  joegenbaclor
 * Modified FixViaNATParametes to not Pop the internal headers for RCVADDR and RCV Port
 *
 * Revision 1.44  2006/08/18 07:12:24  joegenbaclor
 * 1.  added Function overload for GetContentType
 * 2.  Changed unix make files to not include opensbc in make process
 *
 * Revision 1.43  2006/08/15 15:51:14  joegenbaclor
 * Added sanity check to Contact header where some SIP Stacks may send blank contact in response after
 * receiving a request without a contact presnt
 *
 * Revision 1.42  2006/08/14 01:29:08  joegenbaclor
 * Introduced separate provisions for UAS/UAC Current Request/Reponse.  Previousely the current request/response can be accessed via a single provision SIPSession::Get/Set/CurrentRequest()
 *
 * Revision 1.41  2006/08/11 08:38:42  joegenbaclor
 * Initial upload of Presence package
 *
 * Revision 1.40  2006/08/09 12:17:48  joegenbaclor
 * Added RFC 3428 Support
 *
 * Revision 1.39  2006/08/06 02:32:08  joegenbaclor
 * Corrected bug regarding blank call-ids for UnknownTransaction events
 *
 * Revision 1.38  2006/07/28 04:14:21  joegenbaclor
 * Added function to get the size of Accept header if available
 *
 * Revision 1.37  2006/07/19 02:19:21  joegenbaclor
 * 1.  Bug fixes concerning B2BUA destruction during during 3xx, 4xx, 5xx, 6xx local response
 * 2.  Fixed bug where Route header is never removed from outbound Invite resulting that may result to a loop when
 *      Inbound Invite has a Route header present
 * 3.  Added RemoveAllRoutes() in SIPMessage
 *
 * Revision 1.36  2006/07/13 06:53:05  joegenbaclor
 * 1.  Introduced Sanity checks for incoming SIP Message
 * 2.  Corrected bug in SIPURI where enum scheme is not getting recognized.
 * 3.  Added ENUM support to routing
 * 4.  Introduced global routing flag to indicate when and not to rewrite outbound To URI's
 *
 * Revision 1.35  2006/07/06 05:38:11  joegenbaclor
 * Final last minute changes to CBCOM hash support
 *
 * Revision 1.34  2006/06/30 07:12:21  joegenbaclor
 * *** empty log message ***
 *
 * Revision 1.33  2006/06/28 14:15:00  joegenbaclor
 * Bulk commit for Media Proxy functionalities for OpenB2BUA
 * Marks a minor version upgrade from 1.0.2 to 1.0.3
 *
 * Revision 1.32  2006/06/20 09:52:55  joegenbaclor
 * Added Routing and SDP Handlers to B2BUA
 *
 * Revision 1.31  2006/06/09 07:37:30  joegenbaclor
 * 1.  Added Publish, SIPETag & SIPIfMatch  support in SIPMessage
 * 2.  Started work on StatusPage Resource for OSSAppConfig
 *
 * Revision 1.30  2006/06/07 09:18:33  joegenbaclor
 * Added the ability to SIPMessage to read from a PChannel
 *
 * Revision 1.29  2006/05/19 06:30:37  joegenbaclor
 * 1.  Fixed bug in SIPHeaderB where tag may not be parsed properly if no < > enclosure is present
 * 2.  Various enhancements to OpenSBC and proxy session
 *
 * Revision 1.28  2006/05/17 03:43:07  joegenbaclor
 * Bulk commit after sourceforge cvs migration changes
 *
 * Revision 1.27  2006/04/26 00:51:09  joegenbaclor
 * 1.  Added IsMethod() and IsStatusCode() functions to get rid of repetitive code for determining Message class types
 * 2.  Added IsRefer()
 *
 * Revision 1.26  2006/04/21 09:55:10  joegenbaclor
 * Added RFC 3265 header requirements
 *
 * Revision 1.25  2006/04/21 04:12:02  joegenbaclor
 * Added Refer-To header support
 *
 * Revision 1.24  2006/04/07 07:47:20  joegenbaclor
 * More work on OpenPhone/OPAL
 *
 * Revision 1.23  2006/04/05 00:26:04  joegenbaclor
 * 1.  Completed support for STUN NAT traversal method and OPAL style address translation
 * 2.  More bug fixes to ProxySession
 * 3.  Fixed bug in messages where zero size list headers may be interpreted by the parser as a valid header
 * 4.  Started work on OPAL outbound
 *
 * Revision 1.22  2006/03/27 01:25:33  joegenbaclor
 * Some improvements on NICT and NIST state machines
 *
 * Revision 1.21  2006/03/25 18:40:21  joegenbaclor
 * 1.  Worked on some race conditions on Finite State Machines during heavy load.
 * 2.  Intoduced Global Mutex to SIPMessage.  This is inevitable beacuse PDictionary and PList members are not thread-safe
 * 3,  Some minor improvements to simple-proxy application
 *
 * Revision 1.20  2006/03/24 15:42:18  joegenbaclor
 * Some bug fixes including possible deadlocks between Session Destruction and ProcessStackEvent
 *
 * Revision 1.19  2006/03/23 01:36:45  joegenbaclor
 * Added local routing through simple registrar
 *
 * Revision 1.18  2006/03/21 14:11:51  joegenbaclor
 * 1.  Added rport support for NAT Traversal
 * 2.  Added NAT Keep alive mechanism in SIPSession
 * 3.  Some Improvements in various classes affecting NAT like Via, SIPTransport etc.
 *
 * Revision 1.17  2006/03/17 13:51:00  joegenbaclor
 * 1.  Implemented UAC Call Redirection
 * 2.  Implemented SendCancel() method in CallSession
 * 3.  Introduced ClearInternalHeaders() function in SIP Message to erase internal headers inheritted from previous transaction.
 *
 * Revision 1.16  2006/03/15 16:40:46  joegenbaclor
 * 1.  More Dialog Support.
 * 2.  Improved SIPUserAgent::AppendCommonHeaders().
 * 3.  Fixed SDP parser bugs
 * 5.  Renamed SDP Classes to avoid collision with OPAL classes with the same name
 * 4.  Worked on OpalOSSConnection RTP
 *
 * Revision 1.15  2006/03/14 05:53:58  joegenbaclor
 * Modified SetBody() so that if the string is empty m_ContentType is also blanked out
 *
 * Revision 1.14  2006/03/14 03:53:53  joegenbaclor
 * Initial source upload for BETA Release Candidate
 *
 *
 *
 */

#include "SIPMessage.h"
#include "SIPTransport.h"
#include "SIPUserAgent.h"
#include "Encryption.h"

#define new PNEW
#define USE_MSG_REF_COUNT 0

using namespace SIPParser;
using namespace SIPTransports;

PAtomicInteger SIPMessage::m_InternalIdBase;
#if LOG_INSTANCE_COUNTER
PAtomicInteger SIPMessage::m_InstanceCounter = 0;
#endif

PMutex SIPMessage::m_GlobalMutex;
#define GlobalLock() PWaitAndSignal global_lock( SIPMessage::m_GlobalMutex )

SIPMessage::SIPMessage()
{
#if LOG_INSTANCE_COUNTER
  ++SIPMessage::m_InstanceCounter;
#endif

  m_IsFinal = FALSE;

  if( SIPMessage::m_InternalIdBase == INT_MAX)
    SIPMessage::m_InternalIdBase = 0;

  m_InternalId = ++SIPMessage::m_InternalIdBase;

  m_Socket = NULL;
  m_TransactionCanDeleteSocket = FALSE;
  ZeroObjects();

}

SIPMessage::SIPMessage(
  const SIPMessage & msg
)
{
#if LOG_INSTANCE_COUNTER
  ++SIPMessage::m_InstanceCounter;
#endif

  GlobalLock();

  m_IsFinal = FALSE;

  if( SIPMessage::m_InternalIdBase == INT_MAX )
    SIPMessage::m_InternalIdBase = 0;

  m_InternalId = ++SIPMessage::m_InternalIdBase;

  m_Socket = NULL;
  m_TransactionCanDeleteSocket = FALSE;

  ZeroObjects();

  operator=( msg );
}

SIPMessage::SIPMessage(
  const OString & msg
)
{
#if LOG_INSTANCE_COUNTER
  ++SIPMessage::m_InstanceCounter;
#endif

  GlobalLock();

  m_IsFinal = FALSE;

  if( SIPMessage::m_InternalIdBase == INT_MAX )
    SIPMessage::m_InternalIdBase = 0;

  m_InternalId = ++SIPMessage::m_InternalIdBase;

  m_Socket = NULL;
  m_TransactionCanDeleteSocket = FALSE;
  ZeroObjects();
  operator=(msg);
}

SIPMessage::SIPMessage(
  const PBYTEArray & msg
)
{
#if LOG_INSTANCE_COUNTER
  ++SIPMessage::m_InstanceCounter;
#endif

  GlobalLock();

  m_IsFinal = FALSE;

  if( SIPMessage::m_InternalIdBase == INT_MAX )
    SIPMessage::m_InternalIdBase = 0;

  m_InternalId = ++SIPMessage::m_InternalIdBase;
  m_Socket = NULL;
  m_TransactionCanDeleteSocket = TRUE;
  ZeroObjects();
  operator=(msg);
}

SIPMessage::~SIPMessage()
{
#if LOG_INSTANCE_COUNTER
  --SIPMessage::m_InstanceCounter;
#endif

  GlobalLock();
#if USE_MSG_REF_COUNT
  ReleaseRef();
#else
  Cleanup();
#endif
}

BOOL SIPMessage::Read(
  PChannel & inputStream
)
{
  // Do this to force a Read() by the PChannelBuffer outside of the
  // ios::lock() mutex which would prevent simultaneous reads and writes.
  inputStream.SetReadTimeout( PMaxTimeInterval );
#if defined(__MWERKS__) || (__GNUC__ >= 3) || (_MSC_VER >= 1300) || defined(SOLARIS)
  if (inputStream.rdbuf()->pubseekoff(0, ios_base::cur) == streampos(-1))
#else
  if (inputStream.rdbuf()->seekoff(0, ios::cur, ios::in) == EOF)
#endif                  
    inputStream.clear(ios::badbit);

  if (!inputStream.IsOpen())
    return FALSE;

  // get the message from transport into cmd and parse MIME
  OString startLine;
  MimeHeaderList headers;
 
  if (!inputStream.bad() && !inputStream.eof()) 
  {
    inputStream.SetReadTimeout(3000);
    inputStream >> startLine >> headers;
  }

  if (inputStream.bad()) 
  {
    PTRACE_IF(1, inputStream.GetErrorCode(PChannel::LastReadError) != PChannel::NoError,
              "SIP\tPDU Read failed: " << inputStream.GetErrorText(PChannel::LastReadError));
    return FALSE;
  }

  if(startLine.IsEmpty()) 
  {
    return FALSE;
  }

  
  PString body;
  // get the SDP content
  PINDEX contentLength = headers.GetInteger( "Content-Length", 0 );

  if (contentLength > 0)
  {
    inputStream.read(body.GetPointer(contentLength+1), contentLength);
    body[contentLength] = '\0';
  }

  OStringStream strm;
  strm << startLine << "\r\n";
  strm << headers << "\r\n";

  if( !body.IsEmpty() )
    strm << body << "\r\n\r\n"; 

  Parse( strm.str() );

  return TRUE;
}

BOOL SIPMessage::Write(
  PChannel & outputStream
)
{
  Finalize();
  OString str = AsString();
  return outputStream.WriteString( str );
}

void SIPMessage::Parse(
  const OString & msg
)
{
  GlobalLock();

  OStringArray tokens;
  ParserTools::MsgTokenize( msg, tokens );

  if( tokens.GetSize() == 0 )
    return;

  PINDEX i;
  PINDEX bodyOffSet = 0;

  if( m_RawMessage != NULL )
    delete m_RawMessage;

  m_RawMessage = new OStringArray();

  if( m_StartLine != NULL )
    delete m_StartLine;

  m_StartLine = new StartLine( tokens[0] );

  tokens.RemoveAt(0);

  for( i = 0; i < tokens.GetSize(); i++ )
  {
    if( !tokens[i].IsEmpty() )
      m_RawMessage->AppendString( tokens[i] );
    else
      break;

    bodyOffSet = i;
  }

  if( m_RawMessageBody != NULL )
    delete m_RawMessageBody;

  m_RawMessageBody = new OString();

  BOOL hasBody = FALSE;

  for( i = ++bodyOffSet; i < tokens.GetSize(); i++ )
  {
    if( tokens[i].IsEmpty() )
      ++bodyOffSet;
    else
      break;
  }

  for( i = bodyOffSet; i < tokens.GetSize(); i++ )
  {
    if( !tokens[i].IsEmpty() )
    {
      *m_RawMessageBody += tokens[i] + "\r\n";
      hasBody = TRUE;
    }else
      break;
  }

  if( !hasBody )
  {
    delete m_RawMessageBody;
    m_RawMessageBody = NULL;
  }

  Finalize();

  return;
}

void SIPMessage::Parse(
  const PBYTEArray & _msg
)
{
  PBYTEArray msg = _msg;
  msg.SetSize( msg.GetSize() + 1 );
  msg[msg.GetSize()-1] = '\0';
  OString msg2( (const char*)msg.GetPointer() );
  Parse( msg2 );
}

/*
BOOL SIPMessage::EnterCriticalSection()const
{

  m_CriticalSection.Wait();

  return TRUE;
}

void SIPMessage::LeaveCriticalSection()const
{
    m_CriticalSection.Signal();
}
*/

SIPMessage & SIPMessage::operator=(
  const SIPMessage & msg
)
{
  GlobalLock();

  /// object copying a ref to itself;
  //if( msg.m_InternalId == m_InternalId )
  //  return *this;

  /// compare the pointer instead of internal id
  if( &msg == this )
    return *this;

  /// we can now cleanup the object so we have a clean slate
  Cleanup();

  /// now assign this message the contents of msg
  (const_cast< SIPMessage & >( msg )).AssignContents( *this );
  return *this;
}


SIPMessage & SIPMessage::operator=(
  const OString & msg
)
{
  GlobalLock();
  /// we can now cleanup the object so we have a clean slate
  Cleanup();
  Parse( msg );

  return *this;
}

SIPMessage & SIPMessage::operator=(
  const PBYTEArray & msg
)
{
  GlobalLock();
  Cleanup();
  Parse( msg );
  return *this;
}

/////////////////////////////////////////////////////////////


BOOL SIPMessage::GetStartLine(
  StartLine & startLine
)const
{
   

  if( m_StartLine == NULL )
    return FALSE;

  startLine = *m_StartLine;
  return TRUE;
}

BOOL SIPMessage::GetRequestURI(
  SIPURI & uri
)const
{
  if( !IsRequest() )
    return FALSE;

  RequestLine requestLine;
  
  if( !GetRequestLine( requestLine ) )
    return FALSE;

  uri = requestLine.GetRequestURI();

  return TRUE;
}

BOOL SIPMessage::SetRequestURI(
  const SIPURI & uri
)
{
  RequestLine requestLine;
  
  if( !GetRequestLine( requestLine ) )
    return FALSE;

  requestLine.SetRequestURI( uri );

  SetStartLine( requestLine );
  
  return TRUE;
}

OString SIPMessage::GetMethod()const
{
  GlobalLock();
  if( !IsRequest() )
    return FALSE;

  StartLine startLine;
  GetStartLine( startLine );
  return startLine.GetRequestLine()->GetMethod();
}

OString SIPMessage::GetStartLine()const
{
  if( m_StartLine == NULL )
    return OString::Empty();

  return m_StartLine->AsString();
}

void SIPMessage::SetStartLine(
  const StartLine & startLine
)
{
  if( m_StartLine != NULL )
    delete m_StartLine;

  m_StartLine = (StartLine*)startLine.Clone();
}

void SIPMessage::SetStartLine(
  const StatusLine & statusLine
)
{
  StartLine startLine( statusLine );
  SetStartLine( startLine );
}

void SIPMessage::SetStartLine(
  const RequestLine & requestLine
)
{
  StartLine startLine( requestLine );
  SetStartLine( startLine );
}

WORD SIPMessage::GetStatusCode()const
{
  if( IsRequest() )
    return 0;

 StartLine startLine;
 GetStartLine( startLine );

 return startLine.GetStatusLine()->GetStatusCode();
}


BOOL SIPMessage::GetStatusLine(
  StatusLine & statusLine
)const
{
  if( IsRequest() )
    return FALSE;

  StartLine startLine;
  if( !GetStartLine( startLine ) )
    return FALSE;

  StatusLine * sline = startLine.GetStatusLine();
  //PAssertNULL( sline );

  statusLine = *sline;
  return TRUE;
}

BOOL SIPMessage::GetRequestLine(
  RequestLine & requestLine
)const
{
  if( !IsRequest() )
    return FALSE;

  StartLine startLine;
  if( !GetStartLine( startLine ) )
    return FALSE;

  RequestLine * rline = startLine.GetRequestLine();
  //PAssertNULL( rline );

  requestLine = *rline;
  return TRUE;
}

////////////////////////////////////////////////////////////


void SIPMessage::PrintOn( ostream & strm )const
{
  GlobalLock();

  ///PAssert( m_IsFinal, "SIPMessage needs to call Finalize() before this operation" );
  (const_cast< SIPMessage * >( this ))->Finalize();

  if( m_StartLine != NULL )
    strm << *m_StartLine ;

  strm << "\r\n";

  if( m_From != NULL )
    strm << *m_From << "\r\n";

  if( m_To != NULL )
    strm << *m_To << "\r\n";

   if( m_ReferTo != NULL )
    strm << *m_ReferTo << "\r\n";

  /// special headers
  if( m_ViaList != NULL && m_ViaList->GetSize() != 0 )
  {
    for( PINDEX i = 0; i < m_ViaList->GetSize(); i++ )
      strm << (*m_ViaList)[i] << "\r\n";
  }

  if( m_CSeq != NULL )
    strm << *m_CSeq << "\r\n";

  if( m_CallId != NULL )
    strm << *m_CallId << "\r\n";

  if( m_ContactList != NULL && m_ContactList->GetSize() != 0 )
  {
    /// some broken stacks send blank Contact body if its now present in the request
    OStringStream _c;
    _c << *m_ContactList;
    OString c( _c );
    if( !c.Trim().IsEmpty() )
      strm << *m_ContactList << "\r\n";
  }

  /// Route Headers
  if( m_RouteList != NULL && m_RouteList->GetSize() != 0 )
    strm << *m_RouteList << "\r\n";

  if( m_RecordRouteList != NULL && m_RecordRouteList->GetSize() != 0 )
    strm << *m_RecordRouteList << "\r\n";

  if( m_PathList != NULL && m_PathList->GetSize() != 0 )
    strm << *m_PathList << "\r\n";

  if( m_PAssertedIdentityList != NULL && m_PAssertedIdentityList->GetSize() != 0 )
    strm << *m_PAssertedIdentityList << "\r\n";

  if( m_PPreferredIdentityList != NULL && m_PPreferredIdentityList->GetSize() != 0 )
    strm << *m_PPreferredIdentityList << "\r\n";

  if( m_AlertInfo != NULL )
    strm << *m_AlertInfo << "\r\n";

  if( m_CallInfo != NULL )
    strm << *m_CallInfo << "\r\n";

  if( m_ContentDisposition != NULL )
    strm << *m_ContentDisposition << "\r\n";

  if( m_ContentLanguage != NULL )
    strm << *m_ContentLanguage << "\r\n";

  if( m_Date != NULL )
    strm << *m_Date << "\r\n";

  if( m_ErrorInfo != NULL )
    strm << *m_ErrorInfo << "\r\n";

  if( m_MIMEVersion != NULL )
    strm << *m_MIMEVersion << "\r\n";

  if( m_Event != NULL )
    strm << *m_Event << "\r\n";

  if( m_SIPETag != NULL )
    strm << *m_SIPETag << "\r\n";

  if( m_SIPIfMatch != NULL )
    strm << *m_SIPIfMatch << "\r\n";

  if( m_SubscriptionState != NULL )
    strm << *m_SubscriptionState << "\r\n";

  if( m_Organization != NULL )
    strm << *m_Organization << "\r\n";

  if( m_Priority != NULL )
    strm << *m_Priority << "\r\n";

  if( m_Privacy != NULL )
    strm << *m_Privacy << "\r\n";

  if( m_RetryAfter != NULL )
    strm << *m_RetryAfter << "\r\n";

  if( m_Server != NULL )
    strm << *m_Server << "\r\n";

  if( m_Subject != NULL )
    strm << *m_Subject << "\r\n";

  if( m_UserAgent != NULL )
    strm << *m_UserAgent << "\r\n";

  if( m_Warning != NULL )
    strm << *m_Warning << "\r\n";

  /// URI headers
  if( m_ReplyTo != NULL )
    strm << *m_ReplyTo << "\r\n";

  if( m_Replaces != NULL )
    strm << *m_Replaces << "\r\n";

  /// Integer headers
  if( m_Expires != NULL )
    strm << *m_Expires << "\r\n";

  if( m_MaxForwards != NULL )
    strm << *m_MaxForwards << "\r\n";

  if( m_MinExpires != NULL )
    strm << *m_MinExpires << "\r\n";


  if( m_TimeStamp != NULL )
    strm << *m_TimeStamp << "\r\n";

  /// Comma delimited named parameters
  if( m_AuthenticationInfo != NULL )
    strm << *m_AuthenticationInfo << "\r\n";

  if( m_Authorization != NULL )
    strm << *m_Authorization << "\r\n";

  if( m_ProxyAuthenticate != NULL )
    strm << *m_ProxyAuthenticate << "\r\n";

  if( m_ProxyAuthorization != NULL )
    strm << *m_ProxyAuthorization << "\r\n";

  if( m_WWWAuthenticate != NULL )
    strm << *m_WWWAuthenticate << "\r\n";

  /// Comma delimited headers

  if( m_AcceptList != NULL && m_AcceptList->GetSize() != 0 )
    strm << *m_AcceptList << "\r\n";

  if( m_AllowEventsList != NULL && m_AllowEventsList->GetSize() != 0 )
    strm << *m_AllowEventsList << "\r\n";

  if( m_AcceptEncodingList != NULL && m_AcceptEncodingList->GetSize() != 0 )
    strm << *m_AcceptEncodingList << "\r\n";

  if( m_AcceptLanguageList != NULL && m_AcceptLanguageList->GetSize() != 0 )
    strm << *m_AcceptLanguageList << "\r\n";

  if( m_AllowList != NULL && m_AllowList->GetSize() != 0 )
    strm << *m_AllowList << "\r\n";

  if( m_ContentEncodingList != NULL && m_ContentEncodingList->GetSize() != 0 )
    strm << *m_ContentEncodingList << "\r\n";

  if( m_InReplyToList!= NULL && m_InReplyToList->GetSize() != 0 )
    strm << *m_InReplyToList << "\r\n";

  if( m_ProxyRequireList != NULL && m_ProxyRequireList->GetSize() != 0 )
    strm << *m_ProxyRequireList << "\r\n";

  if( m_RequireList != NULL && m_RequireList->GetSize() != 0 )
    strm << *m_RequireList << "\r\n";

  if( m_SupportedList != NULL && m_SupportedList->GetSize() != 0 )
    strm << *m_SupportedList << "\r\n";

  if( m_UnsupportedList != NULL && m_UnsupportedList->GetSize() != 0  )
    strm << *m_UnsupportedList << "\r\n";

  /// session timer headers
  if( m_SessionExpires != NULL )
    strm << *m_SessionExpires << "\r\n";

  if( m_MinSE != NULL )
    strm << *m_MinSE << "\r\n";
  /// Custom Headers

  if( m_CustomHeaders != NULL )
  {
    for( PINDEX i = 0; i < m_CustomHeaders->GetSize(); i++ )
      strm << (*m_CustomHeaders)[i] << "\r\n";
  };

  if( m_ContentType != NULL )
    strm << *m_ContentType << "\r\n";

  if( m_ContentLength != NULL )
  {
    strm << *m_ContentLength << "\r\n";
  }

  strm << "\r\n";


  if( m_RawMessageBody != NULL )
    strm << m_RawMessageBody->Trim() << "\r\n";


}

PObject * SIPMessage::Clone()const
{
  return new SIPMessage( *this );
}

OString SIPMessage::AsString()const
{
  OStringStream strm;
  PrintOn( strm );
  return strm.str();
}

void SIPMessage::Cleanup()
{
  GlobalLock();

   

  if( m_RawMessage != NULL )
    delete m_RawMessage;

  if( m_RawMessageBody != NULL )
    delete m_RawMessageBody;

  if( m_StartLine != NULL )
    delete m_StartLine;

  /// Standad SIP Headers
  if( m_AlertInfo != NULL )
    delete m_AlertInfo;

  if( m_CallId != NULL )
    delete m_CallId;

  if( m_CallInfo != NULL )
    delete m_CallInfo;

  if( m_ContentDisposition != NULL )
    delete m_ContentDisposition;

  if( m_ContentLanguage != NULL )
    delete m_ContentLanguage;

  if( m_ContentLength != NULL )
    delete m_ContentLength;

  if( m_ContentType != NULL )
    delete m_ContentType;

  if( m_Date != NULL )
    delete m_Date;

  if( m_ErrorInfo != NULL )
    delete m_ErrorInfo;

  if( m_MIMEVersion != NULL )
    delete m_MIMEVersion;

  if( m_Organization != NULL )
    delete m_Organization;

  if( m_Event != NULL )
    delete m_Event;

  if( m_SIPETag != NULL )
    delete m_SIPETag;

  if( m_SIPIfMatch != NULL )
    delete m_SIPIfMatch;

  if( m_SubscriptionState != NULL )
    delete m_SubscriptionState;

  if( m_Priority != NULL )
    delete m_Priority;

  if( m_Privacy != NULL )
    delete m_Privacy;

  if( m_RetryAfter != NULL )
    delete m_RetryAfter;

  if( m_Server != NULL )
    delete m_Server;

  if( m_Subject != NULL )
    delete m_Subject;

  if( m_UserAgent != NULL )
    delete m_UserAgent;

  if( m_Warning != NULL )
    delete m_Warning;

  /// URI headers
  if( m_From != NULL )
    delete m_From;

  if( m_To != NULL )
    delete m_To;

  if( m_ReferTo != NULL )
    delete m_ReferTo;

  if( m_Replaces != NULL )
    delete m_Replaces;

  if( m_ReplyTo != NULL )
    delete m_ReplyTo;

  /// Integer headers
  if( m_Expires != NULL )
    delete m_Expires;

  if( m_MaxForwards != NULL )
    delete m_MaxForwards;

  if( m_MinExpires != NULL )
    delete m_MinExpires;

  if( m_TimeStamp != NULL )
    delete m_TimeStamp;

  /// Comma delimited named parameters
  if( m_AuthenticationInfo != NULL )
    delete m_AuthenticationInfo;

  if( m_Authorization != NULL )
    delete m_Authorization;

  if( m_ProxyAuthenticate != NULL )
    delete m_ProxyAuthenticate;

  if( m_ProxyAuthorization != NULL )
    delete m_ProxyAuthorization;

  if( m_WWWAuthenticate != NULL )
    delete m_WWWAuthenticate;

  /// Comma delimited headers

  if( m_AcceptList != NULL )
    delete m_AcceptList;

  if( m_AllowEventsList != NULL )
    delete m_AllowEventsList;

  if( m_AcceptEncodingList != NULL )
    delete m_AcceptEncodingList;

  if( m_AcceptLanguageList != NULL )
    delete m_AcceptLanguageList;

  if( m_AllowList != NULL )
    delete m_AllowList;

  if( m_ContentEncodingList != NULL )
    delete m_ContentEncodingList;

  if( m_ContactList != NULL )
    delete m_ContactList;

  if( m_InReplyToList != NULL )
    delete m_InReplyToList;

  if( m_ProxyRequireList != NULL )
    delete m_ProxyRequireList;

  if( m_RequireList != NULL )
    delete m_RequireList;

  if( m_SupportedList != NULL )
    delete m_SupportedList;

  if( m_UnsupportedList != NULL )
    delete m_UnsupportedList;

  /// Route Headers
  if( m_RouteList != NULL )
    delete m_RouteList;

  if( m_RecordRouteList != NULL )
    delete m_RecordRouteList;

  if( m_PathList != NULL )
    delete m_PathList;

  if( m_PAssertedIdentityList != NULL )
    delete m_PAssertedIdentityList;

  if( m_PPreferredIdentityList != NULL )
    delete m_PPreferredIdentityList;

  /// special headers
  if( m_ViaList != NULL )
    delete m_ViaList;

  if( m_CSeq != NULL )
    delete m_CSeq;

  /// Custom Headers
  if( m_CustomHeaders != NULL )
    delete m_CustomHeaders;

  if( m_InternalHeaders != NULL )
    delete m_InternalHeaders;

  ZeroObjects();

}

void SIPMessage::ZeroObjects()
{
  GlobalLock();

  m_RawMessage = NULL;
  m_RawMessageBody = NULL;
  m_StartLine = NULL;

  /// Standad SIP Headers
  m_AlertInfo = NULL;
  m_CallId = NULL;
  m_CallInfo = NULL;
  m_ContentDisposition = NULL;
  m_ContentLanguage = NULL;
  m_ContentLength = NULL;
  m_ContentType = NULL;
  m_Date = NULL;
  m_ErrorInfo = NULL;
  m_MIMEVersion = NULL;
  m_Organization = NULL;
  m_Event = NULL;
  m_SIPETag = NULL;
  m_SIPIfMatch = NULL;
  m_SubscriptionState = NULL;
  m_Priority = NULL;
  m_Privacy = NULL;
  m_RetryAfter = NULL;
  m_Server = NULL;
  m_Subject = NULL;
  m_UserAgent = NULL;
  m_Warning = NULL;

  /// URI headers
  m_From = NULL;
  m_To = NULL;
  m_ReferTo = NULL;
  m_Replaces = NULL;
  m_ReplyTo = NULL;

  /// Integer headers
  m_Expires = NULL;
  m_SessionExpires = NULL;
  m_MinSE = NULL;
  m_MaxForwards = NULL;
  m_MinExpires = NULL;
  m_TimeStamp = NULL;

  /// Comma delimited named parameters
  m_AuthenticationInfo = NULL;
  m_Authorization = NULL;
  m_ProxyAuthenticate = NULL;
  m_ProxyAuthorization = NULL;
  m_WWWAuthenticate = NULL;

  /// Comma delimited headers

  m_AcceptList = NULL;
  m_AllowEventsList = NULL;
  m_AcceptEncodingList = NULL;
  m_AcceptLanguageList = NULL;
  m_AllowList = NULL;
  m_ContentEncodingList = NULL;
  m_ContactList = NULL;
  m_InReplyToList = NULL;
  m_ProxyRequireList = NULL;
  m_RequireList = NULL;
  m_SupportedList = NULL;
  m_UnsupportedList = NULL;

  /// Route Headers
  m_RouteList = NULL;
  m_RecordRouteList = NULL;
  m_PathList = NULL;
  m_PAssertedIdentityList = NULL;
  m_PPreferredIdentityList = NULL;

  /// special headers
  m_ViaList = NULL;
  m_CSeq = NULL;

  /// Custom Headers
  m_CustomHeaders = NULL;

  m_InternalHeaders = NULL;
}

void SIPMessage::AssignContents(
  SIPMessage & msg
)const
{
  GlobalLock();

  (const_cast< SIPMessage * >( this ))->Finalize();

  /// None-clonable///////////////////////////////////////////////
  msg.m_IsFinal = m_IsFinal;
  msg.m_InternalId = m_InternalId;
  msg.m_Socket = m_Socket;
  msg.m_TransactionCanDeleteSocket = m_TransactionCanDeleteSocket;
  ////////////////////////////////////////////////////////////////

  if( m_RawMessage != NULL )
    msg.m_RawMessage = static_cast<OStringArray*>(m_RawMessage->Clone());

  if( m_RawMessageBody != NULL )
    msg.m_RawMessageBody = static_cast<OString*>(m_RawMessageBody->Clone());

  if( m_StartLine != NULL )
    msg.m_StartLine = static_cast<StartLine*>(m_StartLine->Clone());


  /// Standad SIP Headers
  if( m_AlertInfo != NULL )
    msg.m_AlertInfo = static_cast<AlertInfo*>(m_AlertInfo->Clone());

  if( m_CallId != NULL )
    msg.m_CallId = static_cast<CallId*>(m_CallId->Clone());

  if( m_CallInfo != NULL )
    msg.m_CallInfo = static_cast<CallInfo*>(m_CallInfo->Clone());
  
  if( m_ContentDisposition != NULL )
    msg.m_ContentDisposition = static_cast<ContentDisposition*>(m_ContentDisposition->Clone());

  if( m_ContentLanguage != NULL )
    msg.m_ContentLanguage = static_cast<ContentLanguage*>(m_ContentLanguage->Clone());

  if( m_ContentLength != NULL )
    msg.m_ContentLength = static_cast<ContentLength*>(m_ContentLength->Clone());

  if( m_ContentType != NULL )
    msg.m_ContentType = static_cast<ContentType*>(m_ContentType->Clone());

  if( m_Date != NULL )
    msg.m_Date = static_cast<Date*>(m_Date->Clone());

  if( m_ErrorInfo != NULL )
    msg.m_ErrorInfo = static_cast<ErrorInfo*>(m_ErrorInfo->Clone());

  if( m_MIMEVersion != NULL )
    msg.m_MIMEVersion = static_cast<MIMEVersion*>(m_MIMEVersion->Clone());

  if( m_Organization != NULL )
    msg.m_Organization = static_cast<Organization*>(m_Organization->Clone());

  if( m_Event != NULL )
    msg.m_Event = static_cast<Event*>(m_Event->Clone());
  
  if( m_SIPETag != NULL )
    msg.m_SIPETag = static_cast<SIPETag*>(m_SIPETag->Clone());

  if( m_SIPIfMatch != NULL )
    msg.m_SIPIfMatch = static_cast<SIPIfMatch*>(m_SIPIfMatch->Clone());

  if( m_SubscriptionState != NULL )
    msg.m_SubscriptionState = static_cast<SubscriptionState*>(m_SubscriptionState->Clone());

  if( m_Priority != NULL )
    msg.m_Priority = static_cast<Priority*>(m_Priority->Clone());

  if( m_Privacy != NULL )
    msg.m_Privacy = static_cast<Privacy*>(m_Privacy->Clone());

  if( m_RetryAfter != NULL )
    msg.m_RetryAfter = static_cast<RetryAfter*>(m_RetryAfter->Clone());

  if( m_Server != NULL )
    msg.m_Server = static_cast<Server*>(m_Server->Clone());

  if( m_Subject != NULL )
    msg.m_Subject = static_cast<Subject*>(m_Subject->Clone());

  if( m_UserAgent != NULL )
    msg.m_UserAgent = static_cast<UserAgent*>(m_UserAgent->Clone());

  if( m_Warning != NULL )
    msg.m_Warning = static_cast<Warning*>(m_Warning->Clone());

  /// URI headers
  if( m_From != NULL )
    msg.m_From = static_cast<From*>(m_From->Clone());

  if( m_To != NULL )
    msg.m_To = static_cast<To*>(m_To->Clone());

  if( m_ReferTo != NULL )
    msg.m_ReferTo = static_cast<ReferTo*>(m_ReferTo->Clone());

  if( m_Replaces != NULL )
    msg.m_Replaces = static_cast<Replaces*>(m_Replaces->Clone());

  if( m_ReplyTo != NULL )
    msg.m_ReplyTo = static_cast<ReplyTo*>(m_ReplyTo->Clone());

  /// Integer headers
  if( m_Expires != NULL )
    msg.m_Expires = static_cast<Expires*>(m_Expires->Clone());

  if( m_SessionExpires != NULL )
    msg.m_SessionExpires = static_cast<SessionExpires*>(m_SessionExpires->Clone());

  if( m_MinSE != NULL )
    msg.m_MinSE = static_cast<MinSE*>(m_MinSE->Clone());

  if( m_MaxForwards != NULL )
    msg.m_MaxForwards = static_cast<MaxForwards*>(m_MaxForwards->Clone());

  if( m_MinExpires != NULL )
    msg.m_MinExpires = static_cast<MinExpires*>(m_MinExpires->Clone());

  if( m_TimeStamp != NULL )
    msg.m_TimeStamp = static_cast<TimeStamp*>(m_TimeStamp->Clone());

  /// Comma delimited named parameters
  if( m_AuthenticationInfo != NULL )
    msg.m_AuthenticationInfo = static_cast<AuthenticationInfo*>(m_AuthenticationInfo->Clone());
  
  if( m_Authorization != NULL )
    msg.m_Authorization = static_cast<Authorization*>(m_Authorization->Clone());

  if( m_ProxyAuthenticate != NULL )
    msg.m_ProxyAuthenticate = static_cast<ProxyAuthenticate*>(m_ProxyAuthenticate->Clone());

  if( m_ProxyAuthorization != NULL )
    msg.m_ProxyAuthorization = static_cast<ProxyAuthorization*>(m_ProxyAuthorization->Clone());

  if( m_WWWAuthenticate != NULL )
    msg.m_WWWAuthenticate = static_cast<WWWAuthenticate*>(m_WWWAuthenticate->Clone());

  /// Comma delimited headers
  if( m_AcceptList != NULL )
    msg.m_AcceptList = static_cast<AcceptList*>(m_AcceptList->Clone());

  if( m_AllowEventsList != NULL )
    msg.m_AllowEventsList = static_cast<AllowEventsList*>(m_AllowEventsList->Clone());

  if( m_AcceptEncodingList != NULL )
    msg.m_AcceptEncodingList = static_cast<AcceptEncodingList*>(m_AcceptEncodingList->Clone());

  if( m_AcceptLanguageList != NULL )
    msg.m_AcceptLanguageList = static_cast<AcceptLanguageList*>(m_AcceptLanguageList->Clone());

  if( m_AllowList != NULL )
    msg.m_AllowList = static_cast<AllowList*>(m_AllowList->Clone());

  if( m_ContentEncodingList != NULL )
    msg.m_ContentEncodingList = static_cast<ContentEncodingList*>(m_ContentEncodingList->Clone());

  if( m_ContactList )
    msg.m_ContactList = static_cast<ContactList*>(m_ContactList->Clone());

  if( m_InReplyToList != NULL )
    msg.m_InReplyToList = static_cast<InReplyToList*>(m_InReplyToList->Clone());

  if( m_ProxyRequireList != NULL )
    msg.m_ProxyRequireList = static_cast<ProxyRequireList*>(m_ProxyRequireList->Clone());

  if( m_RequireList != NULL )
    msg.m_RequireList = static_cast<RequireList*>(m_RequireList->Clone());

  if( m_SupportedList != NULL )
    msg.m_SupportedList = static_cast<SupportedList*>(m_SupportedList->Clone());

  if( m_UnsupportedList != NULL )
    msg.m_UnsupportedList = static_cast<UnsupportedList*>(m_UnsupportedList->Clone());

  /// Route Headers
  if( m_RouteList != NULL )
    msg.m_RouteList = static_cast<RouteList*>(m_RouteList->Clone());

  if( m_RecordRouteList != NULL )
    msg.m_RecordRouteList = static_cast<RecordRouteList*>(m_RecordRouteList->Clone());

  if( m_PathList != NULL )
    msg.m_PathList = static_cast<PathList*>(m_PathList->Clone());

  if( m_PAssertedIdentityList != NULL )
    msg.m_PAssertedIdentityList = static_cast<ContactList*>(m_PAssertedIdentityList->Clone());

  if( m_PPreferredIdentityList != NULL )
    msg.m_PPreferredIdentityList = static_cast<ContactList*>(m_PPreferredIdentityList->Clone());

  /// special headers
  if( m_ViaList != NULL )
    msg.m_ViaList = static_cast<ViaList*>(m_ViaList->Clone());

  if( m_CSeq != NULL )
    msg.m_CSeq = static_cast<CSeq*>(m_CSeq->Clone());

  /// Custom Headers
  if( m_CustomHeaders != NULL )
    msg.m_CustomHeaders = static_cast<CustomHeaders*>(m_CustomHeaders->Clone());

  if( m_InternalHeaders != NULL )
    msg.m_InternalHeaders = static_cast<InternalHeaders*>(m_InternalHeaders->Clone());

}

void SIPMessage::Finalize()
{
  GlobalLock();

   // cs( *this );

  if( m_IsFinal )
    return;

  m_IsFinal = TRUE;

  if( m_RawMessage == NULL )
    return;

  OStringArray & msg = *m_RawMessage;

  /// first parse the StartLine if its not set
  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    /// remove it so its not mistaken as a header
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName().Trim().ToLower();
    OString hbody = h.GetMimeValue();
    /// Standad SIP Headers

    if( hname == "alert-info" )
    {
      if( m_AlertInfo == NULL )
      {
        m_AlertInfo = new AlertInfo( h );
      }
    }else if( hname == "call-id" )
    {
      if( m_CallId == NULL )
      {
        m_CallId = new CallId( h );
      }
    }else if( hname == "call-info" )
    {
      if( m_CallInfo == NULL )
      {
        m_CallInfo = new CallInfo( h );
      }
    }else if( hname ==  "content-disposition" )
    {
      if( m_ContentDisposition == NULL )
      {
        m_ContentDisposition = new ContentDisposition( h );
      }
    }else if( hname == "content-language" )
    {
      if( m_ContentLanguage == NULL )
      {
        m_ContentLanguage = new ContentLanguage( h );
      }
    }else if( hname == "content-length" )
    {
      if( m_ContentLength == NULL )
      {
        m_ContentLength = new ContentLength( h );
      }
    }else if( hname == "content-type" )
    {
      if( m_ContentType == NULL )
      {
        m_ContentType = new ContentType( h );
      }
    }else if( hname == "date" )
    {
      if( m_Date == NULL )
      {
        m_Date = new Date( h );
      }
    }else if( hname == "error-info" )
    {
      if( m_ErrorInfo == NULL )
      {
        m_ErrorInfo = new ErrorInfo( h );
      }
    }else if( hname == "mime-version" )
    {
      if( m_MIMEVersion == NULL )
      {
        m_MIMEVersion = new MIMEVersion( h );
      }
    }else if( hname == "organization" )
    {
      if( m_Organization == NULL )
      {
        m_Organization = new Organization( h );
      }
    }else if( hname == "event" )
    {
      if( m_Event == NULL )
      {
        m_Event = new Event( h );
      }
    }else if( hname == "sip-etag" )
    {
      if( m_SIPETag == NULL )
      {
        m_SIPETag = new SIPETag( h );
      }
    }else if( hname == "sip-if-match" )
    {
      if( m_SIPIfMatch == NULL )
      {
        m_SIPIfMatch = new SIPIfMatch( h );
      }
    }else if( hname == "subscription-state" )
    {
      if( m_SubscriptionState == NULL )
      {
        m_SubscriptionState = new SubscriptionState( h );
      }
    }else if( hname == "priority" )
    {
      if( m_Priority == NULL )
      {
        m_Priority = new Priority( h );
      }
    }else if( hname == "privacy" )
    {
      if( m_Privacy == NULL )
      {
        m_Privacy = new Privacy( h );
      }
    }else if( hname == "retry-after" )
    {
      if( m_RetryAfter == NULL )
      {
        m_RetryAfter = new RetryAfter( h );
      }
    }else if( hname == "server" )
    {
      if( m_Server == NULL )
      {
        m_Server = new Server( h );
      }
    }else if( hname == "subject" )
    {
      if( m_Subject == NULL )
      {
        m_Subject = new Subject( h );
      }
    }else if( hname == "user-agent" )
    {
      if( m_UserAgent == NULL )
      {
        m_UserAgent = new UserAgent( h );
      }
    }else if( hname == "warning" )
    {
      if( m_Warning == NULL )
      {
        m_Warning = new Warning( h );
      }
    }else if( hname == "from" )
    {
      /// URI headers
      if( m_From == NULL )
      {
        m_From = new From( h );
      }
    }else if( hname == "to" )
    {
      if( m_To == NULL )
      {
        m_To = new To( h );
      }
    }else if( hname == "refer-to" )
    {
      if( m_ReferTo == NULL )
      {
        m_ReferTo = new ReferTo( h );
      }
    }else if( hname == "replaces" )
    {
      if( m_Replaces == NULL )
      {
        m_Replaces = new Replaces( h );
      }
    }else if( hname == "reply-to" )
    {
      if( m_ReplyTo == NULL )
      {
        m_ReplyTo = new ReplyTo( h );
      }
    }else if( hname == "expires" )
    {
      /// Integer headers
      if( m_Expires == NULL )
      {
        m_Expires = new Expires( h );
      }
    }else if( hname == "session-expires" )
    {
      /// Integer headers
      if( m_SessionExpires == NULL )
      {
        m_SessionExpires = new SessionExpires( h );
      }
    }else if( hname == "min-se" )
    {
      /// Integer headers
      if( m_MinSE == NULL )
      {
        m_MinSE = new MinSE( h );
      }
    }else if( hname == "max-forwards" )
    {
      if( m_MaxForwards == NULL )
      {
        m_MaxForwards = new MaxForwards( h );
      }
    }else if( hname == "min-expires" )
    {
      if( m_MinExpires == NULL )
      {
        m_MinExpires = new MinExpires( h );
      }
    }else if( hname == "time-stamp" )
    {
      if( m_TimeStamp == NULL )
      {
        m_TimeStamp = new TimeStamp( h );
      }
    }else if( hname == "authentication-info" )
    {
      /// Comma delimited named parameters
      if( m_AuthenticationInfo == NULL )
      {
        m_AuthenticationInfo = new AuthenticationInfo( h );
      }
    }else if( hname == "authorization" )
    {
      if( m_Authorization == NULL )
      {
        m_Authorization = new Authorization( h );
      }
    }else if( hname == "proxy-authenticate" )
    {
      if( m_ProxyAuthenticate == NULL )
      {
        m_ProxyAuthenticate = new ProxyAuthenticate( h );
      }
    }else if( hname == "proxy-authorization" )
    {
      if( m_ProxyAuthorization == NULL )
      {
        m_ProxyAuthorization = new ProxyAuthorization( h );
      }
    }else if( hname == "www-authenticate" )
    {
      if( m_WWWAuthenticate == NULL )
      {
        m_WWWAuthenticate = new WWWAuthenticate( h );
      }
    }else if( hname == "accept" )
    {
    /// Comma delimited headers
      if( m_AcceptList == NULL )
      {
        m_AcceptList = new AcceptList();
      }

      Accept accept( h );
      m_AcceptList->Append( accept );

    }else if( hname == "allow-events" )
    {
    /// Comma delimited headers
      if( m_AllowEventsList == NULL )
      {
        m_AllowEventsList = new AllowEventsList();
      }

      AllowEvents allowEvents( h );
      m_AllowEventsList->Append( allowEvents );

    }else if( hname == "accept-encoding" )
    {
      if( m_AcceptEncodingList == NULL )
      {
        m_AcceptEncodingList = new AcceptEncodingList();
      }

      AcceptEncoding acceptEncoding( h );
      m_AcceptEncodingList->Append( acceptEncoding );
    }else if( hname == "accept-language" )
    {
      if( m_AcceptLanguageList == NULL )
      {
        m_AcceptLanguageList = new AcceptLanguageList();
      }

      AcceptLanguage acceptLanguage( h );
      m_AcceptLanguageList->Append( acceptLanguage );

    }else if( hname == "allow" )
    {
      if( m_AllowList == NULL )
      {
        m_AllowList = new AllowList();
      }

      Allow allow(h);
      m_AllowList->Append( allow );

    }else if( hname == "content-encoding" )
    {
      if( m_ContentEncodingList == NULL )
      {
        m_ContentEncodingList = new ContentEncodingList();
      }

      ContentEncoding contentEncoding( h );
      m_ContentEncodingList->Append( contentEncoding );

    }else if( hname == "contact" )
    {
      if( m_ContactList == NULL )
      {
        m_ContactList = new ContactList();
      }

      Contact contact( h );
      m_ContactList->Append( contact );

    }else if( hname == "in-reply-to" )
    {
      if( m_InReplyToList == NULL )
      {
        m_InReplyToList = new InReplyToList();
      }

      InReplyTo inReplyTo( h );
      m_InReplyToList->Append( inReplyTo );

    }else if( hname == "proxy-require" )
    {
      if( m_ProxyRequireList == NULL )
      {
        m_ProxyRequireList = new ProxyRequireList();
      }

      ProxyRequire proxyRequire( h );
      m_ProxyRequireList->Append( proxyRequire );

    }else if( hname == "require" )
    {
      if( m_RequireList == NULL )
      {
        m_RequireList = new RequireList();
      }

      Require require( h );
      m_RequireList->Append( require );

    }else if( hname == "supported" )
    {
      if( m_SupportedList == NULL )
      {
        m_SupportedList = new SupportedList();
      }
      Supported supported( h );
      m_SupportedList->Append( supported );

    }else if( hname == "unsupported" )
    {
      if( m_UnsupportedList == NULL )
      {
        m_UnsupportedList = new UnsupportedList();
      }

      Unsupported unsupported( h );
      m_UnsupportedList->Append( unsupported );

    }else if( hname == "route" )
    {
      /// Route Headers
      if( m_RouteList == NULL )
      {
        m_RouteList = new RouteList();
      }

      OStringArray tokens = h.GetMimeValue().Tokenise( "," );

      for( PINDEX j = 0; j < tokens.GetSize(); j++ )
      {
        Route nh( tokens[j].Trim() );
        m_RouteList->Append( nh, TRUE );
      }

    }else if( hname == "record-route" )
    {
      if( m_RecordRouteList == NULL )
      {
        m_RecordRouteList = new RecordRouteList();
      }

      //RecordRoute recordRoute( h );
      //m_RecordRouteList->Append( recordRoute, TRUE );

      OStringArray tokens = h.GetMimeValue().Tokenise( "," );

      for( PINDEX j = 0; j < tokens.GetSize(); j++ )
      {
        RecordRoute nh( tokens[j].Trim() );
        m_RecordRouteList->Append( nh, TRUE );
      }

    }else if( hname == "path" )
    {
      if( m_PathList == NULL )
      {
        m_PathList = new PathList();
      }

      Path path( h );
      m_PathList->Append( path );

    }else if( hname == "p-asserted-identity" )
    {
      if( m_PAssertedIdentityList == NULL )
      {
        m_PAssertedIdentityList = new ContactList();
      }

      PAssertedIdentity identity( h );
      m_PAssertedIdentityList->Append( identity );

    }else if( hname == "p-preferred-identity" )
    {
      if( m_PPreferredIdentityList == NULL )
      {
        m_PPreferredIdentityList = new ContactList();
      }

      PPreferredIdentity identity( h );
      m_PPreferredIdentityList->Append( identity );

    }else if( hname == "via" )
    {
      /// special headers
      if( m_ViaList == NULL )
      {
        m_ViaList = new Via::Collection();
      }

      OStringArray clist = hbody.Tokenise( ',' );
      if( clist.GetSize() > 1 )
      {
        for( PINDEX z = 0; z < clist.GetSize(); z++ )
        {
          OString token = clist[z].Trim();
          h.SetMimeValue( token );
          m_ViaList->Append( new Via( h ) );
        }
      }else
      {
        m_ViaList->Append( new Via( h ) ); 
      }


    }else if( hname == "cseq" )
    {
      if( m_CSeq == NULL )
      {
        m_CSeq = new CSeq( h );
      }
    }else
    {
      /// Custom Headers
      if( m_CustomHeaders == NULL )
      {
        m_CustomHeaders = new CustomHeaders();
      }

      m_CustomHeaders->Append( new SIPHeader( h ) );
    }
  }



  delete m_RawMessage;
  m_RawMessage = NULL;

}


//////////////////////////////////////////////////////
///OString * m_RawMessageBody;
BOOL SIPMessage::GetBody(
  OString & body
)const
{
  if( !HasBody() )
    return FALSE;
  body = *m_RawMessageBody;
  return TRUE;
}

OString SIPMessage::GetBody()const
{
  OString body;
  GetBody( body );
  return body;
}

void SIPMessage::SetSDP(
  const OString & sdp 
)
{
  SetContentType( "application/sdp" );
  SetBody( sdp );
}

void SIPMessage::SetBody(
  const OString & body
)
{
  if( m_RawMessageBody != NULL )
    delete m_RawMessageBody;

  if( body.IsEmpty() )
  {
    m_RawMessageBody = NULL;
    ContentLength contentLenth( "0" );
    SetContentLength( contentLenth );

    if( m_ContentType != NULL )
    {
      delete m_ContentType;
      m_ContentType = NULL;
    }
    return;
  }

  m_RawMessageBody = new OString( body );

  OString cl( body.GetLength() );
  ContentLength contentLenth( cl );
  SetContentLength( contentLenth );
}

BOOL SIPMessage::HasBody()const
{
  return m_RawMessageBody != NULL;
}

BOOL SIPMessage::HasSDP()const
{
  if( !HasBody() )
    return FALSE;

  ContentType ctype;

  if( !GetContentType( ctype ) )
    return FALSE;

  OString ct = ctype.GetHeaderBody();

  return ct *= "application/sdp";
}

BOOL SIPMessage::AppendRoute(
  const RouteURI & uri,
  PINDEX idx
)
{
  GlobalLock();

   // cs( *this );
  ParseRouteList();

  if( m_RouteList == NULL )
    m_RouteList = new RouteList();

  if( m_RouteList->GetSize() == 0 )
  {
    Route route;
    route.AddURI( uri );
    return m_RouteList->Append( route );
  }else
  {
    return m_RouteList->Append( uri, idx );
  }
}

BOOL SIPMessage::AppendRoute(
  const Route & header,
  BOOL toBottom
)
{
  GlobalLock();

   // cs( *this );
  ParseRouteList();

  if( m_RouteList == NULL )
    m_RouteList = new RouteList();

  return m_RouteList->Append( header, toBottom );
}

RouteList SIPMessage::GetRouteList()const
{
  GlobalLock();

   // cs( *this );
  (const_cast< SIPMessage * >( this ))->ParseRouteList();

  RouteList routeList;

  if( m_RouteList != NULL )
  {
    routeList = RouteList( *m_RouteList );
    routeList.MakeUnique();
  }

  return routeList;
}

BOOL SIPMessage::SetRouteList(
  const RouteList & list
)
{
  GlobalLock();

  RemoveAllRoutes();

  m_RouteList = new RouteList( list );

  return TRUE;
}

BOOL SIPMessage::RemoveAllRoutes()
{
  GlobalLock();

  (const_cast< SIPMessage * >( this ))->ParseRouteList();

  if( m_RouteList != NULL )
  {
    delete m_RouteList;
    m_RouteList = NULL;
  }

  return TRUE;
}

Route SIPMessage::GetRouteAt(
  PINDEX idx
)const
{
  GlobalLock();

   // cs( *this );
  (const_cast< SIPMessage * >( this ))->ParseRouteList();
  return m_RouteList->GetAt( idx );
}

RouteURI SIPMessage::GetRouteTokenAt(
  PINDEX idx1,
  PINDEX idx2
)const
{
  GlobalLock();

   // cs( *this );
  (const_cast< SIPMessage * >( this ))->ParseRouteList();
  return m_RouteList->GetAt( idx1, idx2 );
}

PINDEX SIPMessage::GetRouteSize()const
{
  GlobalLock();

   // cs( *this );
  if( !(const_cast< SIPMessage * >( this ))->ParseRouteList() )
    return 0;
  return m_RouteList->GetSize();
}

RouteURI SIPMessage::PopTopRouteURI()
{
  GlobalLock();

   // cs( *this );
  ParseRouteList();
  return m_RouteList->PopTopURI();
}

/// returns the first route SIP uri
/// take note that this is different than
/// poppin gthe top URI
SIPURI SIPMessage::GetTopRouteURI()const
{
  GlobalLock();
  const_cast<SIPMessage*>(this)->ParseRouteList();
  SIPURI top;
  if( m_RouteList->GetSize() == 0 )
    return top;

  SIPHeaderF & f = m_RouteList->GetAt(0);
  f.GetTopURI( top );
  return top;
}

RouteURI SIPMessage::PopLastRouteURI()
{
  GlobalLock();
  ParseRouteList();
  return m_RouteList->PopLastURI();
}

BOOL SIPMessage::HasRoute()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParseRouteList();
}

BOOL SIPMessage::AppendRecordRoute(
  const RouteURI & uri,
  PINDEX idx
)
{
  GlobalLock();
  ParseRecordRouteList();
  if( m_RecordRouteList == NULL )
    m_RecordRouteList = new RecordRouteList();

  if( m_RecordRouteList->GetSize() == 0 )
  {
    RecordRoute route;
    route.AddURI( uri );
    return m_RecordRouteList->Append( route );
  }else
  {
    return m_RecordRouteList->Append( uri, idx );
  }
}

BOOL SIPMessage::RemoveAllRecordRoutes()
{
  GlobalLock();

  if( m_RecordRouteList != NULL )
  {
    delete m_RecordRouteList;
    m_RecordRouteList = NULL;
  }

  return TRUE;
}


BOOL SIPMessage::AppendRecordRoute(
  const RecordRoute & header,
  BOOL toBottom
)
{
  GlobalLock();
  ParseRecordRouteList();

  if( m_RecordRouteList == NULL )
    m_RecordRouteList = new RecordRouteList();

  return m_RecordRouteList->Append( header, toBottom );
}

RecordRouteList SIPMessage::GetRecordRouteList()const
{
  GlobalLock();
   // cs( *this );
  (const_cast< SIPMessage * >( this ))->ParseRecordRouteList();

  RecordRouteList recordRouteList;

  if( m_RecordRouteList != NULL )
  {
    recordRouteList = RecordRouteList( *m_RecordRouteList );
    recordRouteList.MakeUnique();
  }

  return recordRouteList;
}

BOOL SIPMessage::SetRecordRouteList(
  const RecordRouteList & list
)
{
  GlobalLock();

  if( m_RecordRouteList != NULL )
  {
    delete m_RecordRouteList;
    m_RecordRouteList = NULL;
  }

  m_RecordRouteList = new RecordRouteList( list );

  return TRUE;
}


RecordRoute SIPMessage::GetRecordRouteAt(
  PINDEX idx
)const
{
  GlobalLock();
   // cs( *this );
  (const_cast< SIPMessage * >( this ))->ParseRecordRouteList();
  return m_RecordRouteList->GetAt( idx );
}

RouteURI SIPMessage::GetRecordRouteTokenAt(
  PINDEX idx1,
  PINDEX idx2
)const
{
  GlobalLock();
   // cs( *this );
  (const_cast< SIPMessage * >( this ))->ParseRecordRouteList();
  return m_RecordRouteList->GetAt( idx1, idx2 );
}

PINDEX SIPMessage::GetRecordRouteSize()const
{
  GlobalLock();
   // cs( *this );
  if( !(const_cast< SIPMessage * >( this ))->ParseRecordRouteList() )
    return 0;

  return m_RecordRouteList->GetSize();
}

BOOL SIPMessage::CloneRecordRoutes(
  SIPMessage & msg
)const
{
  GlobalLock();
   // cs( *this );
  (const_cast< SIPMessage * >( this ))->ParseRecordRouteList();

  if( m_RecordRouteList == NULL )
    return FALSE;

  if( m_RecordRouteList->GetSize() == 0 )
    return FALSE;

  if( msg.m_RecordRouteList != NULL )
    delete msg.m_RecordRouteList;

  msg.m_RecordRouteList = (RecordRouteList*)m_RecordRouteList->Clone();

  return TRUE;
}

RouteURI SIPMessage::PopTopRecordRouteURI()
{
  GlobalLock();
   // cs( *this );
  ParseRecordRouteList();
  return m_RecordRouteList->PopTopURI();
}

BOOL SIPMessage::AppendPath(
  const RouteURI & uri,
  PINDEX idx
)
{
  GlobalLock();
   // cs( *this );
  ParsePathList();
  if( m_PathList == NULL )
    m_PathList = new PathList();

  if( m_PathList->GetSize() == 0 )
  {
    Path route;
    route.AddURI( uri );
    return m_PathList->Append( route );
  }else
  {
    return m_PathList->Append( uri, idx );
  }
}

BOOL SIPMessage::AppendPath(
  const Path & header
)
{
  GlobalLock();
  ParsePathList();

  if( m_PathList == NULL )
    m_PathList = new PathList();

  return m_PathList->Append( header );
}

PathList SIPMessage::GetPathList()const
{
  GlobalLock();
   // cs( *this );
  (const_cast< SIPMessage * >( this ))->ParsePathList();

  PathList pathList;

  if( m_PathList != NULL )
  {
    pathList = PathList( *m_PathList );
    pathList.MakeUnique();
  }

  return pathList;
}

BOOL SIPMessage::SetPathList(
  const PathList & list
)
{
  GlobalLock();

  if( m_PathList != NULL )
  {
    delete m_PathList;
    m_PathList = NULL;
  }

  m_PathList = new PathList( list );

  return TRUE;
}


Path SIPMessage::GetPathAt(
  PINDEX idx
)const
{
  GlobalLock();
   // cs( *this );
  (const_cast< SIPMessage * >( this ))->ParsePathList();
  return m_PathList->GetAt( idx );
}

RouteURI SIPMessage::GetPathTokenAt(
  PINDEX idx1,
  PINDEX idx2
)const
{
  GlobalLock();
   // cs( *this );
  (const_cast< SIPMessage * >( this ))->ParsePathList();
  return m_PathList->GetAt( idx1, idx2 );
}

PINDEX SIPMessage::GetPathSize()const
{
  GlobalLock();
   // cs( *this );
  if( !(const_cast< SIPMessage * >( this ))->ParsePathList() )
    return 0;

  return m_PathList->GetSize();
}

BOOL SIPMessage::ClonePaths(
  SIPMessage & msg
)const
{
  GlobalLock();
   // cs( *this );
  (const_cast< SIPMessage * >( this ))->ParsePathList();

  if( m_PathList == NULL )
    return FALSE;

  if( m_PathList->GetSize() == 0 )
    return FALSE;

  if( msg.m_PathList != NULL )
    delete msg.m_PathList;

  msg.m_PathList = (PathList*)m_PathList->Clone();

  return TRUE;
}

RouteURI SIPMessage::PopTopPathURI()
{
  GlobalLock();
  ParsePathList();
  return m_PathList->PopTopURI();
}

BOOL SIPMessage::AppendPAssertedIdentity(
  const PAssertedIdentity & header
)
{
  GlobalLock();
  ParsePAssertedIdentityList();

  if( m_PAssertedIdentityList == NULL )
    m_PAssertedIdentityList = new ContactList();

  return m_PAssertedIdentityList->Append( header );
}

ContactList SIPMessage::GetPAssertedIdentityList()const
{
  GlobalLock();

  (const_cast< SIPMessage * >( this ))->ParsePAssertedIdentityList();

  ContactList identityList;

  if( m_PAssertedIdentityList != NULL )
  {
    identityList = ContactList( *m_PAssertedIdentityList );
    identityList.MakeUnique();
  }

  return identityList;
}

BOOL SIPMessage::SetPAssertedIdentityList(
  const ContactList & list
)
{
  GlobalLock();

  if( m_PAssertedIdentityList != NULL )
  {
    delete m_PAssertedIdentityList;
    m_PAssertedIdentityList = NULL;
  }

  m_PAssertedIdentityList = new ContactList( list );

  return TRUE;
}

PAssertedIdentity SIPMessage::GetPAssertedIdentityAt(
  PINDEX idx
)const
{
  GlobalLock();
   // cs( *this );
  (const_cast< SIPMessage * >( this ))->ParsePAssertedIdentityList();
  return (PAssertedIdentity&)(m_PAssertedIdentityList->GetAt( idx ));
}

PINDEX SIPMessage::GetPAssertedIdentitySize()const
{
  GlobalLock();
   // cs( *this );
  if( !(const_cast< SIPMessage * >( this ))->ParsePAssertedIdentityList() )
    return 0;

  return m_PAssertedIdentityList->GetSize();
}

BOOL SIPMessage::ClonePAssertedIdentities(
  SIPMessage & msg
)const
{
  GlobalLock();
   // cs( *this );
  (const_cast< SIPMessage * >( this ))->ParsePAssertedIdentityList();

  if( m_PAssertedIdentityList == NULL )
    return FALSE;

  if( m_PAssertedIdentityList->GetSize() == 0 )
    return FALSE;

  if( msg.m_PAssertedIdentityList != NULL )
    delete msg.m_PAssertedIdentityList;

  msg.m_PAssertedIdentityList = (ContactList*)m_PAssertedIdentityList->Clone();

  return TRUE;
}

void SIPMessage::RemoveAllPAssertedIdentities()
{
  GlobalLock();
  if( m_PAssertedIdentityList != NULL )
    delete m_PAssertedIdentityList;
  m_PAssertedIdentityList = new ContactList();
}

BOOL SIPMessage::AppendPPreferredIdentity(
  const PPreferredIdentity & header
)
{
  GlobalLock();
   // cs( *this );
  ParsePPreferredIdentityList();

  if( m_PPreferredIdentityList == NULL )
    m_PPreferredIdentityList = new ContactList();

  return m_PPreferredIdentityList->Append( header );
}

ContactList SIPMessage::GetPPreferredIdentityList()const
{
  GlobalLock();

  (const_cast< SIPMessage * >( this ))->ParsePPreferredIdentityList();

  ContactList identityList;

  if( m_PPreferredIdentityList != NULL )
  {
    identityList = ContactList( *m_PPreferredIdentityList );
    identityList.MakeUnique();
  }

  return identityList;
}

BOOL SIPMessage::SetPPreferredIdentityList(
  const ContactList & list
)
{
  GlobalLock();

  if( m_PPreferredIdentityList != NULL )
  {
    delete m_PPreferredIdentityList;
    m_PPreferredIdentityList = NULL;
  }

  m_PPreferredIdentityList = new ContactList( list );

  return TRUE;
}

PPreferredIdentity SIPMessage::GetPPreferredIdentityAt(
  PINDEX idx
)const
{
  GlobalLock();
   // cs( *this );
  (const_cast< SIPMessage * >( this ))->ParsePPreferredIdentityList();
  return (PPreferredIdentity&)(m_PPreferredIdentityList->GetAt( idx ));
}

PINDEX SIPMessage::GetPPreferredIdentitySize()const
{
  GlobalLock();
   // cs( *this );
  if( !(const_cast< SIPMessage * >( this ))->ParsePPreferredIdentityList() )
    return 0;

  return m_PPreferredIdentityList->GetSize();
}

BOOL SIPMessage::ClonePPreferredIdentities(
  SIPMessage & msg
)const
{
  GlobalLock();
   // cs( *this );
  (const_cast< SIPMessage * >( this ))->ParsePPreferredIdentityList();

  if( m_PPreferredIdentityList == NULL )
    return FALSE;

  if( m_PPreferredIdentityList->GetSize() == 0 )
    return FALSE;

  if( msg.m_PPreferredIdentityList != NULL )
    delete msg.m_PPreferredIdentityList;

  msg.m_PPreferredIdentityList = (ContactList*)m_PPreferredIdentityList->Clone();

  return TRUE;
}

void SIPMessage::RemoveAllPPreferredIdentities()
{
  GlobalLock();
  if( m_PPreferredIdentityList != NULL )
    delete m_PPreferredIdentityList;
  m_PPreferredIdentityList = new ContactList();
}

ViaList SIPMessage::GetViaList()const
{
  GlobalLock();
  
  (const_cast< SIPMessage * >( this ))->ParseViaList();

  ViaList viaList;

  if( m_ViaList != NULL )
  {
    viaList = *m_ViaList;
    viaList.MakeUnique();
  }

  return viaList;
}

BOOL SIPMessage::FixViaNATParameters()
{
  GlobalLock();
   // cs( *this );
  (const_cast< SIPMessage * >( this ))->ParseViaList();
  if( m_ViaList == NULL )
    return FALSE;

  if( m_ViaList->GetSize() == 0 )
    return FALSE;

  Via & via = (*m_ViaList)[0];

  SIPHeader a("");
  if( GetInternalHeader( "RCVADDR", a ) && !a.GetHeaderBody().IsEmpty() )
    via.SetReceiveAddress( PIPSocket::Address( a.GetHeaderBody() ) );

  SIPHeader b("");
  if( GetInternalHeader( "RCVPORT", b ) && !b.GetHeaderBody().IsEmpty() )
    via.SetRPort( (WORD)b.GetHeaderBody().AsUnsigned() );

  return TRUE;
}

BOOL SIPMessage::SetViaList(
  const ViaList & list
)
{
  GlobalLock();

  if( m_ViaList != NULL )
  {
    delete m_ViaList;
    m_ViaList = NULL;
  }

  m_ViaList = new ViaList( list );

  return TRUE;
}

//// VIA
BOOL SIPMessage::ParseViaList()
{
  GlobalLock();
   //  cs( *this );

  if( m_ViaList != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  /// first parse the StartLine if its not set
  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    /// remove it so its not mistaken as a header
    msg.RemoveAt( 0 );
  }

  PIntArray viaIdx;
  PINDEX currentIdx = 0;
  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    OString hbody = h.GetMimeValue();
    /// Standad SIP Headers

    if( hname *= "via" )
    {
      /// special headers
      if( m_ViaList == NULL )
      {
        m_ViaList = new Via::Collection();
      }

      /// check if we have a comma delimited list
      OStringArray clist = hbody.Tokenise( ',' );
      if( clist.GetSize() > 1 )
      {
        for( PINDEX z = 0; z < clist.GetSize(); z++ )
        {
          h.SetMimeValue( clist[z].Trim() );
          m_ViaList->Append( new Via( h ) );
        }
      }else
      {
        m_ViaList->Append( new Via( h ) ); 
      }

      viaIdx[currentIdx] = i;
      currentIdx++;
    }
  }

  for( PINDEX j = 0; j < viaIdx.GetSize(); j++ )
    msg.RemoveAt( viaIdx[j] );

  return m_ViaList != NULL;
}

BOOL SIPMessage::AppendVia(
  const Via & header
)
{
  GlobalLock();

  ParseViaList();

  if( m_ViaList == NULL )
  {
    m_ViaList = new Via::Collection();
    m_ViaList->Append( new Via( header ) );
  }else
  {
    
    /// sanity check
    if( m_ViaList->GetSize() > 0 )
    {
      Via & topVia = (*m_ViaList)[0];
      SIPURI topViaURI = topVia.GetURI();
      SIPURI newURI = header.GetURI();
      if( SIPTransport::IsTheSameAddress( topViaURI, newURI, TRUE ) )
        return FALSE;
    }

    ///we append it on top
    Via::Collection oldViaList = *m_ViaList;
    delete m_ViaList;

    m_ViaList = new Via::Collection();
    m_ViaList->Append( new Via( header ) );

    for( PINDEX i = 0; i < oldViaList.GetSize(); i++ )
      m_ViaList->Append( new Via( oldViaList[i] ) );
  }


  return TRUE;
}

BOOL SIPMessage::GetViaAt(
  PINDEX idx,
  Via & via
)const
{
  GlobalLock();
   // cs( *this );
  (const_cast< SIPMessage * >( this ))->ParseViaList();

  if( m_ViaList == NULL )
    return FALSE;

  if( idx > m_ViaList->GetSize() - 1 )
    return FALSE;

  via = (*m_ViaList)[idx];

  return TRUE;
}

BOOL SIPMessage::SetViaAt(
  PINDEX idx,
  const Via & header
)
{
  GlobalLock();

  ParseViaList();

  if( m_ViaList == NULL )
    return FALSE;

  if( idx > m_ViaList->GetSize() - 1 )
    return FALSE;

  (*m_ViaList)[idx] = header;

  return TRUE;
}

void SIPMessage::RemoveAllVias()
{
  GlobalLock();
  ParseViaList();

  if( m_ViaList != NULL )
  {
    delete m_ViaList;
    m_ViaList = NULL;
  }
}

Via SIPMessage::PopTopVia()
{
  GlobalLock();
   // cs( *this );
  ParseViaList();
  Via via;

  if( m_ViaList == NULL )
    return via;

  if( m_ViaList->GetSize() == 0 )
    return via;

  via = (*m_ViaList)[0];

  m_ViaList->RemoveAt(0);
  return via;
}

BOOL SIPMessage::SetViaParameter(
  const OString & name,
  const OString & value
)
{
  GlobalLock();
   // cs( *this );
  ParseViaList();

  if( m_ViaList == NULL )
    return FALSE;

  if( m_ViaList->GetSize() == 0 )
    return FALSE;

  Via & via = (*m_ViaList)[0];
  via.SetParameter( name, value );
  
  return TRUE;
}

PINDEX SIPMessage::GetViaSize()const
{
  GlobalLock();
   // cs( *this );
  (const_cast< SIPMessage * >( this ))->ParseViaList();

  if( m_ViaList == NULL )
    return 0;

  return m_ViaList->GetSize();
}

BOOL SIPMessage::CloneVias( SIPMessage & msg )const
{
  GlobalLock();
   // cs( *this );
  (const_cast< SIPMessage * >( this ))->ParseViaList();

  if( m_ViaList == NULL )
    return FALSE;

  if( m_ViaList->GetSize() == 0 )
    return FALSE;

  if( msg.m_ViaList != NULL )
    delete msg.m_ViaList;

  msg.m_ViaList = new Via::Collection( *m_ViaList );

  msg.m_ViaList->MakeUnique();

  return TRUE;
}

/// Custom Headers
///PLIST( CustomHeaders, SIPHeader );
///CustomHeaders * m_CustomHeaders;

BOOL SIPMessage::ParseCustomHeaders()
{
  GlobalLock();
   //  cs( *this );

  if( m_CustomHeaders != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  /// first parse the StartLine if its not set
  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    /// remove it so its not mistaken as a header
    msg.RemoveAt( 0 );
  }

  PIntArray cIdx;
  PINDEX currentIdx = 0;
  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    const OString & hname = h.GetMimeName();
    /// Standad SIP Headers

    if( !ParserTools::IsKnownHeader( hname ) )
    {
      /// Custom Headers
      if( m_CustomHeaders == NULL )
      {
        m_CustomHeaders = new CustomHeaders();
      }

      m_CustomHeaders->Append( new SIPHeader( h ) );
      cIdx[currentIdx++] = i;
    }

    for( PINDEX j = 0; j < cIdx.GetSize(); j++ )
      msg.RemoveAt( cIdx[j] );
  }

  return m_CustomHeaders != NULL;
}

BOOL SIPMessage::GetCustomHeader(
  const PCaselessString & name,
  SIPHeader & header,
  PINDEX idx
)const
{
  GlobalLock();
   // cs( *this );
  (const_cast< SIPMessage * >( this ))->ParseCustomHeaders();

  if( m_CustomHeaders == NULL )
    return FALSE;

  PINDEX currentIndex = 0;
  BOOL ok = FALSE;
  for( PINDEX i = 0; i < m_CustomHeaders->GetSize(); i++ )
  {
    if( (*m_CustomHeaders)[i].GetHeaderName() *= (const char *)name )
    {
      if( currentIndex == idx )
      {
        header = (*m_CustomHeaders)[i];
        ok = TRUE;
        break;
      }else
        currentIndex++;
    }
  }

  return ok;
}

BOOL SIPMessage::GetCustomHeader(
  SIPHeader & h,
  PINDEX i
)const
{

  GlobalLock();
   // cs( *this );
  (const_cast< SIPMessage * >( this ))->ParseCustomHeaders();

  if( m_CustomHeaders == NULL )
    return FALSE;

  if( i > m_CustomHeaders->GetSize() - 1 )
    return FALSE;

  h = (*m_CustomHeaders)[i];

  return TRUE;
}


void SIPMessage::SetCustomHeader(
  const PCaselessString & name,
  const SIPHeader & header,
  PINDEX idx
)
{
  GlobalLock();
   // cs( *this );
  ParseCustomHeaders();

  if( m_CustomHeaders == NULL )
    return;

  PINDEX currentIndex = 0;
  BOOL ok = FALSE;
  for( PINDEX i = 0; i < m_CustomHeaders->GetSize(); i++ )
  {
    if( (*m_CustomHeaders)[i].GetHeaderName() *= (const char *)name )
    {
      if( currentIndex == idx )
      {
        (*m_CustomHeaders)[i] = header;
        ok = TRUE;
        break;
      }else
        currentIndex++;
    }
  }
}

void SIPMessage::SetCustomHeader(
  const SIPHeader & header,
  PINDEX i
)
{
  GlobalLock();
   // cs( *this );
  ParseCustomHeaders();

  if( m_CustomHeaders == NULL )
    return;

  if( i > m_CustomHeaders->GetSize() - 1 )
    return;

  (*m_CustomHeaders)[i] == header;
}

void SIPMessage::AddCustomHeader(
  const SIPHeader & header
)
{
  GlobalLock();
   // cs( *this );
  ParseCustomHeaders();

  if( m_CustomHeaders == NULL )
    m_CustomHeaders = new CustomHeaders();

  m_CustomHeaders->Append( new SIPHeader( header ) );
}

BOOL SIPMessage::PopCustomHeader(
  const PCaselessString & name,
  SIPHeader & h,
  PINDEX idx
)
{
  GlobalLock();
   // cs( *this );
  ParseCustomHeaders();

  if( m_CustomHeaders == NULL )
    return FALSE;

  PINDEX currentIndex = 0;
  BOOL ok = FALSE;
  for( PINDEX i = 0; i < m_CustomHeaders->GetSize(); i++ )
  {
    if( (*m_CustomHeaders)[i].GetHeaderName() *= (const char *)name )
    {
      if( currentIndex == idx )
      {
        h = (*m_CustomHeaders)[i];
        m_CustomHeaders->RemoveAt(i);
        ok = TRUE;
        break;
      }else
        currentIndex++;
    }
  }

  return ok;
}

BOOL SIPMessage::HasInternalHeader(
  const PCaselessString & name
)const
{
  return !GetInternalHeader( name ).IsEmpty();
}

OString SIPMessage::GetInternalHeader(
  const PCaselessString & name,
  PINDEX i
)const
{
  GlobalLock();
  SIPHeader h("");
  if( !GetInternalHeader( name, h, i ) )
    return OString::Empty();

  return h.GetHeaderBody();
}

BOOL SIPMessage::GetInternalHeader(
  const PCaselessString & name,
  SIPHeader & header,
  PINDEX idx
)const
{
  GlobalLock();
   // cs( *this );
  ///ParseInternalHeaders();

  if( m_InternalHeaders == NULL )
    return FALSE;

  PINDEX currentIndex = 0;
  BOOL ok = FALSE;
  for( PINDEX i = 0; i < m_InternalHeaders->GetSize(); i++ )
  {
    if( (*m_InternalHeaders)[i].GetHeaderName() *= (const char *)name )
    {
      if( currentIndex == idx )
      {
        header = (*m_InternalHeaders)[i];
        ok = TRUE;
        break;
      }else
        currentIndex++;
    }
  }

  return ok;
}

BOOL SIPMessage::GetInternalHeader(
  SIPHeader & h,
  PINDEX i
)const
{

  GlobalLock();
   // cs( *this );
  ///ParseInternalHeaders();

  if( m_InternalHeaders == NULL )
    return FALSE;

  if( i > m_InternalHeaders->GetSize() - 1 )
    return FALSE;

  h = (*m_InternalHeaders)[i];

  return TRUE;
}

void SIPMessage::SetInternalHeader( 
 const PCaselessString & name,
 const OString & value,
 PINDEX index
)
{
  GlobalLock();
  SIPHeader h( (const char *)name );
  h.SetHeaderBody( value );
  SetInternalHeader( name, h, index );
}

void SIPMessage::SetInternalHeader(
  const PCaselessString & name,
  const SIPHeader & header,
  PINDEX idx
)
{
  GlobalLock();

  if( m_InternalHeaders == NULL )
    m_InternalHeaders = new InternalHeaders();

  PINDEX currentIndex = 0;
  BOOL ok = FALSE;
  for( PINDEX i = 0; i < m_InternalHeaders->GetSize(); i++ )
  {
    if( (*m_InternalHeaders)[i].GetHeaderName() *= (const char *)name )
    {
      if( currentIndex == idx )
      {
        (*m_InternalHeaders)[i] = header;
        ok = TRUE;
        break;
      }else
        currentIndex++;
    }
  }

  OString value = header.GetHeaderBody();
  if( !ok )
    AddInternalHeader( (const char *)name, value );
}

void SIPMessage::SetInternalHeader(
  const SIPHeader & header,
  PINDEX i
)
{
  GlobalLock();

  if( m_InternalHeaders == NULL )
    m_InternalHeaders = new InternalHeaders();

  if( i > m_InternalHeaders->GetSize() - 1 )
    return;

  (*m_InternalHeaders)[i] == header;
}

void SIPMessage::AddInternalHeader(
  const OString & name,
  const OString & value
)
{
  GlobalLock();
  SIPHeader h( name, value );
  AddInternalHeader( h );
}

void SIPMessage::ClearInternalHeaders()
{
  GlobalLock();

  if( m_InternalHeaders != NULL )
    m_InternalHeaders->RemoveAll();
}

void SIPMessage::AddInternalHeader(
  const SIPHeader & header
)
{
  GlobalLock();

  if( m_InternalHeaders == NULL )
    m_InternalHeaders = new InternalHeaders();

  m_InternalHeaders->Append( new SIPHeader( header ) );
}

BOOL SIPMessage::PopInternalHeader(
  const PCaselessString & name,
  SIPHeader & h,
  PINDEX idx
)
{
  GlobalLock();
   // cs( *this );
  ///ParseInternalHeaders();

  if( m_InternalHeaders == NULL )
    return FALSE;

  PINDEX currentIndex = 0;
  BOOL ok = FALSE;
  for( PINDEX i = 0; i < m_InternalHeaders->GetSize(); i++ )
  {
    OString hname = (*m_InternalHeaders)[i].GetHeaderName();
    if( hname *= (const char *)name )
    {
      if( currentIndex == idx )
      {
        h = (*m_InternalHeaders)[i];
        m_InternalHeaders->RemoveAt(i);
        ok = TRUE;
        break;
      }else
        currentIndex++;
    }
  }

  return ok;
}

BOOL SIPMessage::GetInternalHeader(
  const PCaselessString & name,
  SIPHeader & h,
  PINDEX idx
)
{
  GlobalLock();
   // cs( *this );
  ///ParseInternalHeaders();

  if( m_InternalHeaders == NULL )
    return FALSE;

  PINDEX currentIndex = 0;
  BOOL ok = FALSE;
  for( PINDEX i = 0; i < m_InternalHeaders->GetSize(); i++ )
  {
    if( (*m_InternalHeaders)[i].GetHeaderName() *= (const char *)name )
    {
      if( currentIndex == idx )
      {
        h = (*m_InternalHeaders)[i];
        ok = TRUE;
        break;
      }else
        currentIndex++;
    }
  }

  return ok;
}


TransactionId SIPMessage::GetTransactionId()const
{
  GlobalLock();
  return TransactionId( *this );
}


BOOL SIPMessage::IsRequest()const
{
  GlobalLock();
  StartLine startLine;
  GetStartLine( startLine );
  return startLine.IsRequest();
}

BOOL SIPMessage::IsMethod(
  const PCaselessString & meth
) const
{
  GlobalLock();
  StartLine startLine;
  GetStartLine( startLine );
  if( !startLine.IsRequest() )
    return FALSE;

  if( startLine.GetRequestLine() == NULL )
    return FALSE;

  return startLine.GetRequestLine()->GetMethod() *= (const char *)meth;
}

BOOL SIPMessage::IsInvite()const
{
   return IsMethod("INVITE" );
}

BOOL SIPMessage::IsAck()const
{
  return IsMethod( "ACK" );
}

BOOL SIPMessage::IsBye()const
{
  return IsMethod( "BYE" );
}

BOOL SIPMessage::IsOptions()const
{
  return IsMethod( "OPTIONS" );
}

BOOL SIPMessage::IsInfo()const
{
  return IsMethod( "INFO" );
}

BOOL SIPMessage::IsRegister()const
{
  return IsMethod( "REGISTER" );
}

BOOL SIPMessage::IsSubscribe()const
{
  return IsMethod( "SUBSCRIBE" );
}

BOOL SIPMessage::IsUnsubscribe()const
{
  GlobalLock();
  
  if( GetCSeqMethod().ToUpper().Trim() != "SUBSCRIBE" )
    return FALSE;

  if( HasExpires() )
  {
    Expires expires;
    GetExpires( expires );
    if( expires.AsInteger() == 0 )
    {
      return TRUE;
    }
  }

  return FALSE;
}

BOOL SIPMessage::IsNotify()const
{
  return IsMethod( "NOTIFY" );
}

BOOL SIPMessage::IsPublish()const
{
  return IsMethod( "PUBLISH" );
}


BOOL SIPMessage::IsRefer()const
{
  return IsMethod( "REFER" );
}

BOOL SIPMessage::IsMessage()const
{
  return IsMethod( "MESSAGE" );
}

BOOL SIPMessage::IsUpdate()const
{
  return IsMethod( "UPDATE" );
}


BOOL SIPMessage::IsUnregister()const
{
  GlobalLock();
  
  ///if( !IsRequest() || !IsRegister() )
  ///  return FALSE;

  /// only check the CSeq method to include evaluation
  /// of 200 Ok response for unregistration.  This
  /// is beneficial for proxies hi-jacking Upper Registrations

  if( GetCSeqMethod().ToUpper().Trim() != "REGISTER" )
    return FALSE;

  if( GetContactSize() > 0 )
  {
    Contact contact;
    if( !GetContactAt( contact, 0 ) )
      return FALSE;
    
    if( contact.GetHeaderBody() *= "*" )
    {
      return TRUE;
    }else /// check if expires == 0
    {
      ContactURI contactURI;
      if( contact.GetURI( contactURI ) )
      {
        OString paramValue;
        contactURI.GetParameter( "expires", paramValue );
        if( paramValue *= "0" )
          return TRUE;
      }
    }
  }

  if( HasExpires() )
  {
    Expires expires;
    GetExpires( expires );
    if( expires.AsInteger() == 0 )
    {
      return TRUE;
    }
  }

  return FALSE;
}

BOOL SIPMessage::IsUnpublish()const
{
  GlobalLock();
  
  if( !IsPublish() )
    return FALSE;

  //check Expires
  Expires expires;
  GetExpires( expires );
  if( expires.AsInteger() == 0 )
    return TRUE;

  return FALSE;
}

BOOL SIPMessage::IsCancel()const
{
  return IsMethod( "CANCEL" );
}

BOOL SIPMessage::IsKeepAlive()const
{
  return IsMethod( "KEEP-ALIVE" );
}

BOOL SIPMessage::IsStatusCode(
  int baseCode
) const
{
  GlobalLock();
  StartLine startLine;
  GetStartLine( startLine );
  if( startLine.IsRequest() )
    return FALSE;

  if( startLine.GetStatusLine() == NULL )
    return FALSE;

  WORD statusCode = startLine.GetStatusLine()->GetStatusCode();

  return statusCode >= baseCode && statusCode < baseCode + 100;
}

BOOL SIPMessage::Is1xx()const
{
  return IsStatusCode( 100 );
}


BOOL SIPMessage::Is2xx()const
{
  return IsStatusCode( 200 );
}

BOOL SIPMessage::Is3xx()const
{
  return IsStatusCode( 300 );
}

BOOL SIPMessage::Is4xx()const
{
  return IsStatusCode( 400 );
}

BOOL SIPMessage::Is5xx()const
{
  return IsStatusCode( 500 );
}

BOOL SIPMessage::Is6xx()const
{
  return IsStatusCode( 600 );
}


///////////////////////////////////////////////////////////


///CREATE_STANDARD_PARSER( Authorization, "authorization" );

BOOL SIPMessage::ParseAuthorization()
{
  GlobalLock();
   // cs( *this );

  if( m_Authorization != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "authorization" )
    {
      if( m_Authorization == NULL )
      {
        m_Authorization = new Authorization( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_Authorization != NULL;
}

BOOL SIPMessage::GetAuthorization(
  Authorization & h
)const
{
  GlobalLock();
  //if( !ParseAuthorization() )
  if( !(const_cast< SIPMessage * >( this ))->ParseAuthorization() )
    return FALSE;

  if( !HasAuthorization() )
    return FALSE;

  h = *m_Authorization;

  return TRUE;
}

void SIPMessage::SetAuthorization(
  const Authorization & h
)
{
  GlobalLock();

  if( m_Authorization != NULL )
    delete m_Authorization;
  m_Authorization = (Authorization*)h.Clone();
}

BOOL SIPMessage::HasAuthorization()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParseAuthorization();
}


////////////////////////////////////////////////////////////


BOOL SIPMessage::ParseAlertInfo()
{
  GlobalLock();
   // cs( *this );

  if( m_AlertInfo != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "alert-info" )
    {
      if( m_AlertInfo == NULL )
      {
        m_AlertInfo = new AlertInfo( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_AlertInfo != NULL;
}

BOOL SIPMessage::GetAlertInfo(
  AlertInfo & h
)const
{
  GlobalLock();
  if( !(const_cast< SIPMessage * >( this ))->ParseAlertInfo() )
    return FALSE;

  if( !HasAlertInfo() )
    return FALSE;

  h = *m_AlertInfo;

  return TRUE;
}

void SIPMessage::SetAlertInfo(
  const AlertInfo & h
)
{
  GlobalLock();

  if( m_AlertInfo != NULL )
    delete m_AlertInfo;
  m_AlertInfo = (AlertInfo*)h.Clone();
}

BOOL SIPMessage::HasAlertInfo()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParseAlertInfo();
}

////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseCallId()
{
  GlobalLock();
   // cs( *this );

  if( m_CallId != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "call-id" )
    {
      if( m_CallId == NULL )
      {
        m_CallId = new CallId( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_CallId != NULL;
}

BOOL SIPMessage::GetCallId(
  CallId & h
)const
{
  GlobalLock();
  if( !(const_cast< SIPMessage * >( this ))->ParseCallId() )
    return FALSE;

  if( !HasCallId() )
    return FALSE;

  h = *m_CallId;

  return TRUE;
}

OString SIPMessage::GetCallId()const
{
  GlobalLock();
  CallId callId;
  if( !GetCallId( callId ) )
    return OString::Empty();

  return callId.GetHeaderBody();
}

void SIPMessage::SetCallId(
  const CallId & h
)
{
  GlobalLock();

  if( m_CallId != NULL )
    delete m_CallId;
  m_CallId = (CallId*)h.Clone();
}

BOOL SIPMessage::HasCallId()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParseCallId();
}

/////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseCallInfo()
{
  GlobalLock();
   // cs( *this );

  if( m_CallInfo != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "call-info" )
    {
      if( m_CallInfo == NULL )
      {
        m_CallInfo = new CallInfo( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_CallInfo != NULL;
}

BOOL SIPMessage::GetCallInfo(
  CallInfo & h
)const
{
  GlobalLock();
  if( !(const_cast< SIPMessage * >( this ))->ParseCallInfo() )
    return FALSE;

  if( !HasCallInfo() )
    return FALSE;

  h = *m_CallInfo;

  return TRUE;
}

void SIPMessage::SetCallInfo(
  const CallInfo & h
)
{
  GlobalLock();

  if( m_CallInfo != NULL )
    delete m_CallInfo;
  m_CallInfo = (CallInfo*)h.Clone();
}

BOOL SIPMessage::HasCallInfo()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParseCallInfo();
}

//////////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseContentDisposition()
{
  GlobalLock();
   // cs( *this );

  if( m_ContentDisposition != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "content-disposition" )
    {
      if( m_ContentDisposition == NULL )
      {
        m_ContentDisposition = new ContentDisposition( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_ContentDisposition != NULL;
}

BOOL SIPMessage::GetContentDisposition(
  ContentDisposition & h
)const
{
  GlobalLock();
  if( !(const_cast< SIPMessage * >( this ))->ParseContentDisposition() )
    return FALSE;

  if( !HasContentDisposition() )
    return FALSE;

  h = *m_ContentDisposition;

  return TRUE;
}

void SIPMessage::SetContentDisposition(
  const ContentDisposition & h
)
{
  GlobalLock();

  if( m_ContentDisposition != NULL )
    delete m_ContentDisposition;
  m_ContentDisposition = (ContentDisposition*)h.Clone();
}

BOOL SIPMessage::HasContentDisposition()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParseContentDisposition();
}

//////////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseContentLanguage()
{
  GlobalLock();
   // cs( *this );

  if( m_ContentLanguage != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "content-language" )
    {
      if( m_ContentLanguage == NULL )
      {
        m_ContentLanguage = new ContentLanguage( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_ContentLanguage != NULL;
}

BOOL SIPMessage::GetContentLanguage(
  ContentLanguage & h
)const
{
  GlobalLock();
  if( !(const_cast< SIPMessage * >( this ))->ParseContentLanguage() )
    return FALSE;

  if( !HasContentLanguage() )
    return FALSE;

  h = *m_ContentLanguage;

  return TRUE;
}

void SIPMessage::SetContentLanguage(
  const ContentLanguage & h
)
{
  GlobalLock();

  if( m_ContentLanguage != NULL )
    delete m_ContentLanguage;
  m_ContentLanguage = (ContentLanguage*)h.Clone();
}

BOOL SIPMessage::HasContentLanguage()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParseContentLanguage();
}

//////////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseContentLength()
{
  GlobalLock();
   // cs( *this );

  if( m_ContentLength != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "content-length" )
    {
      if( m_ContentLength == NULL )
      {
        m_ContentLength = new ContentLength( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_ContentLength != NULL;
}

BOOL SIPMessage::GetContentLength(
  ContentLength & h
)const
{
  GlobalLock();
  if( !(const_cast< SIPMessage * >( this ))->ParseContentLength() )
    return FALSE;

  if( !HasContentLength() )
    return FALSE;

  h = *m_ContentLength;

  return TRUE;
}

void SIPMessage::SetContentLength(
  DWORD len
)
{
  GlobalLock();
  ContentLength h;
  h.SetHeaderBody( OString( len ));
  SetContentLength( h );
}

void SIPMessage::SetContentLength(
  const ContentLength & h
)
{
  GlobalLock();

  if( m_ContentLength != NULL )
    delete m_ContentLength;
  m_ContentLength = (ContentLength*)h.Clone();
}

BOOL SIPMessage::HasContentLength()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParseContentLength();
}

//////////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseContentType()
{
  GlobalLock();
   // cs( *this );

  if( m_ContentType != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "content-type" )
    {
      if( m_ContentType == NULL )
      {
        m_ContentType = new ContentType( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_ContentType != NULL;
}

OString SIPMessage::GetContentType()const
{
  ContentType h;
  if( !GetContentType( h ) )
    return OString::Empty();

  return h.GetHeaderBody();
}

BOOL SIPMessage::GetContentType(
  ContentType & h
)const
{
  GlobalLock();
  if( !(const_cast< SIPMessage * >( this ))->ParseContentType() )
    return FALSE;

  if( !HasContentType() )
    return FALSE;

  h = *m_ContentType;

  return TRUE;
}

void SIPMessage::SetContentType(
  const OString & type
)
{
  GlobalLock();
  ContentType cType( type );
  SetContentType( cType );
}

void SIPMessage::SetContentType(
  const ContentType & h
)
{
  GlobalLock();

  if( m_ContentType != NULL )
    delete m_ContentType;
  m_ContentType = (ContentType*)h.Clone();
}

BOOL SIPMessage::HasContentType()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParseContentType();
}

//////////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseCSeq()
{
  GlobalLock();
   // cs( *this );

  if( m_CSeq != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "cseq" )
    {
      if( m_CSeq == NULL )
      {
        m_CSeq = new CSeq( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_CSeq != NULL;
}

BOOL SIPMessage::GetCSeq(
  CSeq & h
)const
{
  GlobalLock();
  if( !(const_cast< SIPMessage * >( this ))->ParseCSeq() )
    return FALSE;

  if( !HasCSeq() )
    return FALSE;

  h = *m_CSeq;

  return TRUE;
}

OString SIPMessage::GetCSeqMethod()const
{
  GlobalLock();
  CSeq cseq;
  GetCSeq( cseq );
  return cseq.GetMethod().ToUpper();
}

DWORD SIPMessage::GetCSeqNumber()const
{
  GlobalLock();
  CSeq cseq;
  GetCSeq( cseq );
  return cseq.GetSequence();
}

void SIPMessage::SetCSeq(
  const CSeq & h
)
{
  GlobalLock();

  if( m_CSeq != NULL )
    delete m_CSeq;
  m_CSeq = (CSeq*)h.Clone();
}

BOOL SIPMessage::HasCSeq()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParseCSeq();
}

DWORD SIPMessage::IncrementCSeq()
{
  GlobalLock();
  CSeq cseq;
  GetCSeq( cseq );
  SetCSeq( ++cseq );
  return cseq.GetSequence();
}
//////////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseDate()
{
  GlobalLock();
   // cs( *this );

  if( m_Date != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "date" )
    {
      if( m_Date == NULL )
      {
        m_Date = new Date( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_Date != NULL;
}

BOOL SIPMessage::GetDate(
  Date & h
)const
{
  GlobalLock();
  if( !(const_cast< SIPMessage * >( this ))->ParseDate() )
    return FALSE;

  if( !HasDate() )
    return FALSE;

  h = *m_Date;

  return TRUE;
}

void SIPMessage::SetDate(
  const Date & h
)
{
  GlobalLock();

  if( m_Date != NULL )
    delete m_Date;
  m_Date = (Date*)h.Clone();
}

BOOL SIPMessage::HasDate()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParseDate();
}

//////////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseErrorInfo()
{
  GlobalLock();
   // cs( *this );

  if( m_ErrorInfo != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "error-info" )
    {
      if( m_ErrorInfo == NULL )
      {
        m_ErrorInfo = new ErrorInfo( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_ErrorInfo != NULL;
}

BOOL SIPMessage::GetErrorInfo(
  ErrorInfo & h
)const
{
  GlobalLock();
  if( !(const_cast< SIPMessage * >( this ))->ParseErrorInfo() )
    return FALSE;

  if( !HasErrorInfo() )
    return FALSE;

  h = *m_ErrorInfo;

  return TRUE;
}

void SIPMessage::SetErrorInfo(
  const ErrorInfo & h
)
{
  GlobalLock();

  if( m_ErrorInfo != NULL )
    delete m_ErrorInfo;
  m_ErrorInfo = (ErrorInfo*)h.Clone();
}

BOOL SIPMessage::HasErrorInfo()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParseErrorInfo();
}

//////////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseMIMEVersion()
{
  GlobalLock();
   // cs( *this );

  if( m_MIMEVersion != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "mime-version" )
    {
      if( m_MIMEVersion == NULL )
      {
        m_MIMEVersion = new MIMEVersion( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_MIMEVersion != NULL;
}

BOOL SIPMessage::GetMIMEVersion(
  MIMEVersion & h
)const
{
  GlobalLock();
  if( !(const_cast< SIPMessage * >( this ))->ParseMIMEVersion() )
    return FALSE;

  if( !HasMIMEVersion() )
    return FALSE;

  h = *m_MIMEVersion;

  return TRUE;
}

void SIPMessage::SetMIMEVersion(
  const MIMEVersion & h
)
{
  GlobalLock();

  if( m_MIMEVersion != NULL )
    delete m_MIMEVersion;
  m_MIMEVersion = (MIMEVersion*)h.Clone();
}

BOOL SIPMessage::HasMIMEVersion()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParseMIMEVersion();
}

//////////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseOrganization()
{
  GlobalLock();
   // cs( *this );

  if( m_Organization != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "organization" )
    {
      if( m_Organization == NULL )
      {
        m_Organization = new Organization( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_Organization != NULL;
}

BOOL SIPMessage::GetOrganization(
  Organization & h
)const
{
  GlobalLock();
  if( !(const_cast< SIPMessage * >( this ))->ParseOrganization() )
    return FALSE;

  if( !HasOrganization() )
    return FALSE;

  h = *m_Organization;

  return TRUE;
}

void SIPMessage::SetOrganization(
  const Organization & h
)
{
  GlobalLock();

  if( m_Organization != NULL )
    delete m_Organization;
  m_Organization = (Organization*)h.Clone();
}

BOOL SIPMessage::HasOrganization()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParseOrganization();
}

//////////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParsePriority()
{
  GlobalLock();
   // cs( *this );

  if( m_Priority != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "priority" )
    {
      if( m_Priority == NULL )
      {
        m_Priority = new Priority( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_Priority != NULL;
}

BOOL SIPMessage::GetPriority(
  Priority & h
)const
{
  GlobalLock();
  if( !(const_cast< SIPMessage * >( this ))->ParsePriority() )
    return FALSE;

  if( !HasPriority() )
    return FALSE;

  h = *m_Priority;

  return TRUE;
}

void SIPMessage::SetPriority(
  const Priority & h
)
{
  GlobalLock();

  if( m_Priority != NULL )
    delete m_Priority;
  m_Priority = (Priority*)h.Clone();
}

BOOL SIPMessage::HasPriority()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParsePriority();
}


//////////////////////////////////////////////////////////////////////////////
BOOL SIPMessage::ParsePrivacy()
{
  GlobalLock();
   // cs( *this );

  if( m_Privacy != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "privacy" )
    {
      if( m_Privacy == NULL )
      {
        m_Privacy = new Privacy( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_Privacy != NULL;
}


BOOL SIPMessage::GetPrivacy(
  Privacy & h
)const
{
  GlobalLock();
  if( !(const_cast< SIPMessage * >( this ))->ParsePrivacy() )
    return FALSE;

  if( !HasPrivacy() )
    return FALSE;

  h = *m_Privacy;

  return TRUE;
}

void SIPMessage::SetPrivacy(
  const Privacy & h
)
{
  GlobalLock();

  if( m_Privacy != NULL )
    delete m_Privacy;
  m_Privacy = (Privacy*)h.Clone();
}

BOOL SIPMessage::HasPrivacy()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParsePrivacy();
}

//////////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseRetryAfter()
{
  GlobalLock();
   // cs( *this );

  if( m_RetryAfter != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "retry-after" )
    {
      if( m_RetryAfter == NULL )
      {
        m_RetryAfter = new RetryAfter( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_RetryAfter != NULL;
}

BOOL SIPMessage::GetRetryAfter(
  RetryAfter & h
)const
{
  GlobalLock();
  if( !(const_cast< SIPMessage * >( this ))->ParseRetryAfter() )
    return FALSE;

  if( !HasRetryAfter() )
    return FALSE;

  h = *m_RetryAfter;

  return TRUE;
}

void SIPMessage::SetRetryAfter(
  const RetryAfter & h
)
{
  GlobalLock();

  if( m_RetryAfter != NULL )
    delete m_RetryAfter;
  m_RetryAfter = (RetryAfter*)h.Clone();
}

BOOL SIPMessage::HasRetryAfter()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParseRetryAfter();
}

//////////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseServer()
{
  GlobalLock();
   // cs( *this );

  if( m_Server != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "server" )
    {
      if( m_Server == NULL )
      {
        m_Server = new Server( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_Server != NULL;
}

BOOL SIPMessage::GetServer(
  Server & h
)const
{
  GlobalLock();
  if( !(const_cast< SIPMessage * >( this ))->ParseServer() )
    return FALSE;

  if( !HasServer() )
    return FALSE;

  h = *m_Server;

  return TRUE;
}

void SIPMessage::SetServer(
  const Server & h
)
{
  GlobalLock();

  if( m_Server != NULL )
    delete m_Server;
  m_Server = (Server*)h.Clone();
}

BOOL SIPMessage::HasServer()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParseServer();
}

//////////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseSubject()
{
  GlobalLock();
   // cs( *this );

  if( m_Subject != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "subject" )
    {
      if( m_Subject == NULL )
      {
        m_Subject = new Subject( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_Subject != NULL;
}

BOOL SIPMessage::GetSubject(
  Subject & h
)const
{
  GlobalLock();
  if( !(const_cast< SIPMessage * >( this ))->ParseSubject() )
    return FALSE;

  if( !HasSubject() )
    return FALSE;

  h = *m_Subject;

  return TRUE;
}

void SIPMessage::SetSubject(
  const Subject & h
)
{
  GlobalLock();

  if( m_Subject != NULL )
    delete m_Subject;
  m_Subject = (Subject*)h.Clone();
}

BOOL SIPMessage::HasSubject()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParseSubject();
}

//////////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseUserAgent()
{
  GlobalLock();
   // cs( *this );

  if( m_UserAgent != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "user-agent" )
    {
      if( m_UserAgent == NULL )
      {
        m_UserAgent = new UserAgent( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_UserAgent != NULL;
}

BOOL SIPMessage::GetUserAgent(
  UserAgent & h
)const
{
  GlobalLock();
  if( !(const_cast< SIPMessage * >( this ))->ParseUserAgent() )
    return FALSE;

  if( !HasUserAgent() )
    return FALSE;

  h = *m_UserAgent;

  return TRUE;
}

void SIPMessage::SetUserAgent(
  const UserAgent & h
)
{
  GlobalLock();

  if( m_UserAgent != NULL )
    delete m_UserAgent;
  m_UserAgent = (UserAgent*)h.Clone();
}

BOOL SIPMessage::HasUserAgent()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParseUserAgent();
}

//////////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseWarning()
{
  GlobalLock();
   // cs( *this );

  if( m_Warning != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "warning" )
    {
      if( m_Warning == NULL )
      {
        m_Warning = new Warning( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_Warning != NULL;
}

BOOL SIPMessage::GetWarning(
  Warning & h
)const
{
  GlobalLock();
  if( !(const_cast< SIPMessage * >( this ))->ParseWarning() )
    return FALSE;

  if( !HasWarning() )
    return FALSE;

  h = *m_Warning;

  return TRUE;
}

void SIPMessage::SetWarning(
  const Warning & h
)
{
  GlobalLock();

  if( m_Warning != NULL )
    delete m_Warning;
  m_Warning = (Warning*)h.Clone();
}

BOOL SIPMessage::HasWarning()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParseWarning();
}

//////////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseFrom()
{
  GlobalLock();
   // cs( *this );

  if( m_From != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "from" )
    {
      if( m_From == NULL )
      {
        m_From = new From( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_From != NULL;
}

BOOL SIPMessage::GetFrom(
  From & h
)const
{
  GlobalLock();
   // cs( *this );
  if( !(const_cast< SIPMessage * >( this ))->ParseFrom() )
    return FALSE;

  if( !HasFrom() )
    return FALSE;

  h = *m_From;

  return TRUE;
}

SIPURI SIPMessage::GetFromURI()const
{
  GlobalLock();
   // cs( *this );
  From from;

  if( !(const_cast< SIPMessage * >( this ))->GetFrom( from ) )
    return SIPURI();

  return from.GetURI();
}

OString SIPMessage::GetFromTag()const
{
  GlobalLock();
   // cs( *this );
  From from;

  if( !(const_cast< SIPMessage * >( this ))->GetFrom( from ) )
    return OString::Empty();

  OString tag;
  from.GetParameter( "tag", tag );

  /// some implementations dont have <> enclosures so try the uri if its there
  if( tag.IsEmpty() )
    from.GetURI().GetParameter( "tag", tag );
  
  return tag;
}

void SIPMessage::SetFromTag( 
  const OString & tag 
)
{
  GlobalLock();
   // cs( *this );
  From from;
  GetFrom( from );
  from.SetParameter( "tag", tag );
  SetFrom( from );
}

void SIPMessage::SetFrom(
  const From & h
)
{
  GlobalLock();

  if( m_From != NULL )
    delete m_From;
  m_From = (From*)h.Clone();
}

BOOL SIPMessage::HasFrom()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParseFrom();
}

//////////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseTo()
{
  GlobalLock();
   // cs( *this );

  if( m_To != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "to" )
    {
      if( m_To == NULL )
      {
        m_To = new To( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_To != NULL;
}

BOOL SIPMessage::GetTo(
  To & h
)const
{
  GlobalLock();
  if( !(const_cast< SIPMessage * >( this ))->ParseTo() )
    return FALSE;

  if( !HasTo() )
    return FALSE;

  h = *m_To;

  return TRUE;
}

SIPURI SIPMessage::GetToURI()const
{
  GlobalLock();
  To to;

  if( !GetTo( to ) )
    return SIPURI();

  return to.GetURI();
}

OString SIPMessage::GetToTag()const
{
  GlobalLock();
  To to;

  if( !GetTo( to ) )
    return OString::Empty();

  OString tag;
  to.GetParameter( "tag", tag );

  /// some implementations dont have <> enclosures so try the uri if its there
  if( tag.IsEmpty() )
    to.GetURI().GetParameter( "tag", tag );

  return tag;
}

void SIPMessage::SetToTag( 
  const OString & tag 
)
{
  GlobalLock();
   // cs( *this );
  To to;
  GetTo( to );
  to.SetParameter( "tag", tag );
  SetTo( to );
}

void SIPMessage::SetTo(
  const To & h
)
{
  GlobalLock();

  if( m_To != NULL )
    delete m_To;
  m_To = (To*)h.Clone();
}

BOOL SIPMessage::HasTo()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParseTo();
}

//////////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseReplyTo()
{
  GlobalLock();
   // cs( *this );

  if( m_ReplyTo != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "reply-to" )
    {
      if( m_ReplyTo == NULL )
      {
        m_ReplyTo = new ReplyTo( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_ReplyTo != NULL;
}

BOOL SIPMessage::GetReplyTo(
  ReplyTo & h
)const
{
  GlobalLock();
  if( !(const_cast< SIPMessage * >( this ))->ParseReplyTo() )
    return FALSE;

  if( !HasReplyTo() )
    return FALSE;

  h = *m_ReplyTo;

  return TRUE;
}

void SIPMessage::SetReplyTo(
  const ReplyTo & h
)
{
  GlobalLock();

  if( m_ReplyTo != NULL )
    delete m_ReplyTo;
  m_ReplyTo = (ReplyTo*)h.Clone();
}

BOOL SIPMessage::HasReplyTo()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParseReplyTo();
}

//////////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseExpires()
{
  GlobalLock();
   // cs( *this );

  if( m_Expires != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "expires" )
    {
      if( m_Expires == NULL )
      {
        m_Expires = new Expires( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_Expires != NULL;
}

BOOL SIPMessage::GetExpires(
  Expires & h
)const
{
  GlobalLock();
  if( !(const_cast< SIPMessage * >( this ))->ParseExpires() )
    return FALSE;

  if( !HasExpires() )
    return FALSE;

  h = *m_Expires;

  return TRUE;
}

void SIPMessage::SetExpires(
  const Expires & h
)
{
  GlobalLock();

  if( m_Expires != NULL )
    delete m_Expires;
  m_Expires = (Expires*)h.Clone();
}

BOOL SIPMessage::HasExpires()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParseExpires();
}

//////////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseMaxForwards()
{
  GlobalLock();
   // cs( *this );

  if( m_MaxForwards != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "max-forwards" )
    {
      if( m_MaxForwards == NULL )
      {
        m_MaxForwards = new MaxForwards( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_MaxForwards != NULL;
}

BOOL SIPMessage::GetMaxForwards(
  MaxForwards & h
)const
{
  GlobalLock();
  if( !(const_cast< SIPMessage * >( this ))->ParseMaxForwards() )
    return FALSE;

  if( !HasMaxForwards() )
    return FALSE;

  h = *m_MaxForwards;

  return TRUE;
}

void SIPMessage::SetMaxForwards( 
  WORD maxForwards
)
{
  GlobalLock();
  MaxForwards maxF;
  maxF.SetHeaderBody( OString( maxForwards ) );
  SetMaxForwards( maxF );
}

void SIPMessage::SetMaxForwards(
  const MaxForwards & h
)
{
  GlobalLock();

  if( m_MaxForwards != NULL )
    delete m_MaxForwards;
  m_MaxForwards = (MaxForwards*)h.Clone();
}

BOOL SIPMessage::HasMaxForwards()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParseMaxForwards();
}

//////////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseMinExpires()
{
  GlobalLock();
   // cs( *this );

  if( m_MinExpires != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "min-expires" )
    {
      if( m_MinExpires == NULL )
      {
        m_MinExpires = new MinExpires( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_MinExpires != NULL;
}

BOOL SIPMessage::GetMinExpires(
  MinExpires & h
)const
{
  GlobalLock();
  if( !(const_cast< SIPMessage * >( this ))->ParseMinExpires() )
    return FALSE;

  if( !HasMinExpires() )
    return FALSE;

  h = *m_MinExpires;

  return TRUE;
}

void SIPMessage::SetMinExpires(
  const MinExpires & h
)
{
  GlobalLock();

  if( m_MinExpires != NULL )
    delete m_MinExpires;
  m_MinExpires = (MinExpires*)h.Clone();
}

BOOL SIPMessage::HasMinExpires()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParseMinExpires();
}

//////////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseTimeStamp()
{
  GlobalLock();
   // cs( *this );

  if( m_TimeStamp != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "time-stamp" )
    {
      if( m_TimeStamp == NULL )
      {
        m_TimeStamp = new TimeStamp( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_TimeStamp != NULL;
}

BOOL SIPMessage::GetTimeStamp(
  TimeStamp & h
)const
{
  GlobalLock();
  if( !(const_cast< SIPMessage * >( this ))->ParseTimeStamp() )
    return FALSE;

  if( !HasTimeStamp() )
    return FALSE;

  h = *m_TimeStamp;

  return TRUE;
}

void SIPMessage::SetTimeStamp(
  const TimeStamp & h
)
{
  GlobalLock();

  if( m_TimeStamp != NULL )
    delete m_TimeStamp;
  m_TimeStamp = (TimeStamp*)h.Clone();
}

BOOL SIPMessage::HasTimeStamp()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParseTimeStamp();
}

//////////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseAuthenticationInfo()
{
  GlobalLock();
   // cs( *this );

  if( m_AuthenticationInfo != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "authentication-info" )
    {
      if( m_AuthenticationInfo == NULL )
      {
        m_AuthenticationInfo = new AuthenticationInfo( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_AuthenticationInfo != NULL;
}

BOOL SIPMessage::GetAuthenticationInfo(
  AuthenticationInfo & h
)const
{
  GlobalLock();
  if( !(const_cast< SIPMessage * >( this ))->ParseAuthenticationInfo() )
    return FALSE;

  if( !HasAuthenticationInfo() )
    return FALSE;

  h = *m_AuthenticationInfo;

  return TRUE;
}

void SIPMessage::SetAuthenticationInfo(
  const AuthenticationInfo & h
)
{
  GlobalLock();

  if( m_AuthenticationInfo != NULL )
    delete m_AuthenticationInfo;
  m_AuthenticationInfo = (AuthenticationInfo*)h.Clone();
}

BOOL SIPMessage::HasAuthenticationInfo()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParseAuthenticationInfo();
}

//////////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseProxyAuthenticate()
{
  GlobalLock();
   // cs( *this );

  if( m_ProxyAuthenticate != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "proxy-authenticate" )
    {
      if( m_ProxyAuthenticate == NULL )
      {
        m_ProxyAuthenticate = new ProxyAuthenticate( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_ProxyAuthenticate != NULL;
}

BOOL SIPMessage::GetProxyAuthenticate(
  ProxyAuthenticate & h
)const
{
  GlobalLock();
  if( !(const_cast< SIPMessage * >( this ))->ParseProxyAuthenticate() )
    return FALSE;

  if( !HasProxyAuthenticate() )
    return FALSE;

  h = *m_ProxyAuthenticate;

  return TRUE;
}

void SIPMessage::SetProxyAuthenticate(
  const ProxyAuthenticate & h
)
{
  GlobalLock();

  if( m_ProxyAuthenticate != NULL )
    delete m_ProxyAuthenticate;
  m_ProxyAuthenticate = (ProxyAuthenticate*)h.Clone();
}

BOOL SIPMessage::HasProxyAuthenticate()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParseProxyAuthenticate();
}

//////////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseProxyAuthorization()
{
  GlobalLock();
   // cs( *this );

  if( m_ProxyAuthorization != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "proxy-authorization" )
    {
      if( m_ProxyAuthorization == NULL )
      {
        m_ProxyAuthorization = new ProxyAuthorization( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_ProxyAuthorization != NULL;
}

BOOL SIPMessage::GetProxyAuthorization(
  ProxyAuthorization & h
)const
{
  GlobalLock();
  if( !(const_cast< SIPMessage * >( this ))->ParseProxyAuthorization() )
    return FALSE;

  if( !HasProxyAuthorization() )
    return FALSE;

  h = *m_ProxyAuthorization;

  return TRUE;
}

void SIPMessage::SetProxyAuthorization(
  const ProxyAuthorization & h
)
{
  GlobalLock();

  if( m_ProxyAuthorization != NULL )
    delete m_ProxyAuthorization;
  m_ProxyAuthorization = (ProxyAuthorization*)h.Clone();
}

BOOL SIPMessage::HasProxyAuthorization()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParseProxyAuthorization();
}

//////////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseWWWAuthenticate()
{
  GlobalLock();
   // cs( *this );

  if( m_WWWAuthenticate != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "www-authenticate" )
    {
      if( m_WWWAuthenticate == NULL )
      {
        m_WWWAuthenticate = new WWWAuthenticate( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_WWWAuthenticate != NULL;
}

BOOL SIPMessage::GetWWWAuthenticate(
  WWWAuthenticate & h
)const
{
  GlobalLock();
  if( !(const_cast< SIPMessage * >( this ))->ParseWWWAuthenticate() )
    return FALSE;

  if( !HasWWWAuthenticate() )
    return FALSE;

  h = *m_WWWAuthenticate;

  return TRUE;
}

void SIPMessage::SetWWWAuthenticate(
  const WWWAuthenticate & h
)
{
  GlobalLock();

  if( m_WWWAuthenticate != NULL )
    delete m_WWWAuthenticate;
  m_WWWAuthenticate = (WWWAuthenticate*)h.Clone();
}

BOOL SIPMessage::HasWWWAuthenticate()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParseWWWAuthenticate();
}


////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseAcceptList()
{
  GlobalLock();
   // cs( *this );

  if( m_AcceptList != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "accept" )
    {
      if( m_AcceptList == NULL )
      {
        m_AcceptList = new AcceptList();
      }

      Accept nh( h );

      m_AcceptList->Append( nh );
      msg.RemoveAt( i );
      break;
    }
  }

  return m_AcceptList != NULL;
}

BOOL SIPMessage::AppendAccept(
  const OString & token,
  PINDEX idx
)
{
  GlobalLock();
  ParseAcceptList();
    return FALSE;

  if( m_AcceptList == NULL )
  {
    m_AcceptList = new AcceptList();
  }

  return m_AcceptList->Append( token, idx );
}

BOOL SIPMessage::AppendAccept(
  const Accept & header
)
{
  GlobalLock();
  ParseAcceptList();

  if( m_AcceptList == NULL )
  {
    m_AcceptList = new AcceptList();
  }

  m_AcceptList->Append( header );

  return TRUE;
}

PINDEX SIPMessage::GetAcceptSize()const
{
  GlobalLock();
  (const_cast< SIPMessage * >( this ))->ParseAcceptList();
  return m_AcceptList->GetSize();
}

Accept & SIPMessage::GetAcceptAt(
  PINDEX idx
)const
{
  GlobalLock();
  (const_cast< SIPMessage * >( this ))->ParseAcceptList();
  return (Accept&)m_AcceptList->GetAt( idx );
}

OString SIPMessage::GetAcceptTokenAt(
  PINDEX idx1,
  PINDEX idx2
)const
{
  GlobalLock();
  (const_cast< SIPMessage * >( this ))->ParseAcceptList();

  if( m_AcceptList == NULL )
  {
    return OString::Empty();
  }

  return m_AcceptList->GetTokenAt( idx1, idx2 );
}

BOOL SIPMessage::RemoveAllAccept()
{
  GlobalLock();

  if( m_AcceptList != NULL )
  {
    delete m_AcceptList;
    m_AcceptList = NULL;
  }

  return TRUE;
}

///////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseAcceptEncodingList()
{
  GlobalLock();
   // cs( *this );

  if( m_AcceptEncodingList != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "accept-encoding" )
    {
      if( m_AcceptEncodingList == NULL )
      {
        m_AcceptEncodingList = new AcceptEncodingList();
      }

      AcceptEncoding nh( h );

      m_AcceptEncodingList->Append( nh );
      msg.RemoveAt( i );
      break;
    }
  }

  return m_AcceptEncodingList != NULL;
}

BOOL SIPMessage::AppendAcceptEncoding(
  const OString & token,
  PINDEX idx
)
{
  GlobalLock();
  ParseAcceptEncodingList();
    return FALSE;

  if( m_AcceptEncodingList == NULL )
  {
    m_AcceptEncodingList = new AcceptEncodingList();
  }

  return m_AcceptEncodingList->Append( token, idx );
}

BOOL SIPMessage::AppendAcceptEncoding(
  const AcceptEncoding & header
)
{
  GlobalLock();
  ParseAcceptEncodingList();

  if( m_AcceptEncodingList == NULL )
  {
    m_AcceptEncodingList = new AcceptEncodingList();
  }

  m_AcceptEncodingList->Append( header );

  return TRUE;
}

AcceptEncoding & SIPMessage::GetAcceptEncodingAt(
  PINDEX idx
)const
{
  GlobalLock();
  (const_cast< SIPMessage * >( this ))->ParseAcceptEncodingList();
  return (AcceptEncoding&)m_AcceptEncodingList->GetAt( idx );
}

OString SIPMessage::GetAcceptEncodingTokenAt(
  PINDEX idx1,
  PINDEX idx2
)const
{
  GlobalLock();
  (const_cast< SIPMessage * >( this ))->ParseAcceptEncodingList();

  if( m_AcceptEncodingList == NULL )
  {
    return OString::Empty();
  }

  return m_AcceptEncodingList->GetTokenAt( idx1, idx2 );
}

BOOL SIPMessage::RemoveAllAcceptEncoding()
{
  GlobalLock();

  if( m_AcceptEncodingList != NULL )
  {
    delete m_AcceptEncodingList;
    m_AcceptEncodingList = NULL;
  }

  return TRUE;
}

///////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseAcceptLanguageList()
{
  GlobalLock();
   // cs( *this );

  if( m_AcceptLanguageList != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "accept-language" )
    {
      if( m_AcceptLanguageList == NULL )
      {
        m_AcceptLanguageList = new AcceptLanguageList();
      }

      AcceptLanguage nh( h );

      m_AcceptLanguageList->Append( nh );
      msg.RemoveAt( i );
      break;
    }
  }

  return m_AcceptLanguageList != NULL;
}

BOOL SIPMessage::AppendAcceptLanguage(
  const OString & token,
  PINDEX idx
)
{
  GlobalLock();
  ParseAcceptLanguageList();
    return FALSE;

  if( m_AcceptLanguageList == NULL )
  {
    m_AcceptLanguageList = new AcceptLanguageList();
  }

  return m_AcceptLanguageList->Append( token, idx );
}

BOOL SIPMessage::AppendAcceptLanguage(
  const AcceptLanguage & header
)
{
  GlobalLock();
  ParseAcceptLanguageList();

  if( m_AcceptLanguageList == NULL )
  {
    m_AcceptLanguageList = new AcceptLanguageList();
  }

  m_AcceptLanguageList->Append( header );

  return TRUE;
}

AcceptLanguage & SIPMessage::GetAcceptLanguageAt(
  PINDEX idx
)const
{
  GlobalLock();
  (const_cast< SIPMessage * >( this ))->ParseAcceptLanguageList();
  return (AcceptLanguage&)m_AcceptLanguageList->GetAt( idx );
}

OString SIPMessage::GetAcceptLanguageTokenAt(
  PINDEX idx1,
  PINDEX idx2
)const
{
  GlobalLock();
  (const_cast< SIPMessage * >( this ))->ParseAcceptLanguageList();

  if( m_AcceptLanguageList == NULL )
  {
    return OString::Empty();
  }

  return m_AcceptLanguageList->GetTokenAt( idx1, idx2 );
}

BOOL SIPMessage::RemoveAllAcceptLanguage()
{
  GlobalLock();

  if( m_AcceptLanguageList != NULL )
  {
    delete m_AcceptLanguageList;
    m_AcceptLanguageList = NULL;
  }

  return TRUE;
}

///////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseAllowList()
{
  GlobalLock();
   // cs( *this );

  if( m_AllowList != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "allow" )
    {
      if( m_AllowList == NULL )
      {
        m_AllowList = new AllowList();
      }

      Allow nh( h );

      m_AllowList->Append( nh );
      msg.RemoveAt( i );
      break;
    }
  }

  return m_AllowList != NULL;
}

BOOL SIPMessage::AppendAllow(
  const OString & token,
  PINDEX idx
)
{
  GlobalLock();
  ParseAllowList();
    return FALSE;

  if( m_AllowList == NULL )
  {
    m_AllowList = new AllowList();
  }

  return m_AllowList->Append( token, idx );
}

BOOL SIPMessage::AppendAllow(
  const Allow & header
)
{
  GlobalLock();
  ParseAllowList();

  if( m_AllowList == NULL )
  {
    m_AllowList = new AllowList();
  }

  m_AllowList->Append( header );

  return TRUE;
}

BOOL SIPMessage::IsAllowed( SIPMessage::Methods meth )const
{
  OString method = SIPMessage::GetMethodString( meth );
  
  GlobalLock();
  (const_cast< SIPMessage * >( this ))->ParseAllowList();
  
  if( m_AllowList == NULL )
    return FALSE;
  
  for( PINDEX i = 0; i < m_AllowList->GetSize(); i++ )
  {
    Allow allow = (Allow&)m_AllowList->GetAt( i );
    for( PINDEX j = 0; j < allow.GetSize(); j++ )
    {
      OString allowMeth = allow.GetEntry( j );
      if( allowMeth *= method )
        return TRUE;
    }
  }
  
  return FALSE;
  
}
 

Allow & SIPMessage::GetAllowAt(
  PINDEX idx
)const
{
  GlobalLock();
  (const_cast< SIPMessage * >( this ))->ParseAllowList();
  return (Allow&)m_AllowList->GetAt( idx );
}

OString SIPMessage::GetAllowTokenAt(
  PINDEX idx1,
  PINDEX idx2
)const
{
  GlobalLock();
  (const_cast< SIPMessage * >( this ))->ParseAllowList();

  if( m_AllowList == NULL )
  {
    return OString::Empty();
  }

  return m_AllowList->GetTokenAt( idx1, idx2 );
}

BOOL SIPMessage::RemoveAllAllow()
{
  GlobalLock();

  if( m_AllowList != NULL )
  {
    delete m_AllowList;
    m_AllowList = NULL;
  }

  return TRUE;
}


///////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseContentEncodingList()
{
  GlobalLock();
   // cs( *this );

  if( m_ContentEncodingList != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "content-encoding" )
    {
      if( m_ContentEncodingList == NULL )
      {
        m_ContentEncodingList = new ContentEncodingList();
      }

      ContentEncoding nh( h );

      m_ContentEncodingList->Append( nh );
      msg.RemoveAt( i );
      break;
    }
  }

  return m_ContentEncodingList != NULL;
}

BOOL SIPMessage::AppendContentEncoding(
  const OString & token,
  PINDEX idx
)
{
  GlobalLock();
  ParseContentEncodingList();
    return FALSE;

  if( m_ContentEncodingList == NULL )
  {
    m_ContentEncodingList = new ContentEncodingList();
  }

  return m_ContentEncodingList->Append( token, idx );
}

BOOL SIPMessage::AppendContentEncoding(
  const ContentEncoding & header
)
{
  GlobalLock();
  ParseContentEncodingList();

  if( m_ContentEncodingList == NULL )
  {
    m_ContentEncodingList = new ContentEncodingList();
  }

  m_ContentEncodingList->Append( header );

  return TRUE;
}


ContentEncoding & SIPMessage::GetContentEncodingAt(
  PINDEX idx
)const
{
  GlobalLock();
  (const_cast< SIPMessage * >( this ))->ParseContentEncodingList();
  return (ContentEncoding&)m_ContentEncodingList->GetAt( idx );
}

OString SIPMessage::GetContentEncodingTokenAt(
  PINDEX idx1,
  PINDEX idx2
)const
{
  GlobalLock();
  (const_cast< SIPMessage * >( this ))->ParseContentEncodingList();

  if( m_ContentEncodingList == NULL )
  {
    return OString::Empty();
  }

  return m_ContentEncodingList->GetTokenAt( idx1, idx2 );
}

BOOL SIPMessage::RemoveAllContentEncoding()
{
  GlobalLock();

  if( m_ContentEncodingList != NULL )
  {
    delete m_ContentEncodingList;
    m_ContentEncodingList = NULL;
  }

  return TRUE;
}

///////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseContactList()
{
  GlobalLock();
   // cs( *this );

  if( m_ContactList != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "contact" )
    {
      if( m_ContactList == NULL )
      {
        m_ContactList = new ContactList();
      }

      Contact nh( h );

      m_ContactList->Append( nh );
      msg.RemoveAt( i );
      break;
    }
  }

  return m_ContactList != NULL;
}

BOOL SIPMessage::AppendContact(
  const OString & token,
  PINDEX idx
)
{
  GlobalLock();
  ParseContactList();
    return FALSE;

  if( m_ContactList == NULL )
  {
    m_ContactList = new ContactList();
  }

  return m_ContactList->Append( token, idx );
}

BOOL SIPMessage::AppendContact(
  const ContactURI & contactURI
)
{
  GlobalLock();
  Contact contact;
  contact.AddURI( contactURI );
  return AppendContact( contact );
}

BOOL SIPMessage::AppendContact(
  const Contact & header
)
{
  GlobalLock();
  ParseContactList();

  if( m_ContactList == NULL )
  {
    m_ContactList = new ContactList();
  }

  m_ContactList->Append( header );

  return TRUE;
}

SIPURI SIPMessage::GetTopContactURI()const
{
  GlobalLock();
  Contact contact ;
  if( !GetContactAt( contact, 0 ) )
    return SIPURI();

  ContactURI curi;
  contact.GetURI( curi );
  return curi.GetURI();
}


void SIPMessage::SetTopContactURI(
  const SIPURI & contactURI
)
{
  GlobalLock();
  ContactURI curi;
  curi.SetURI( contactURI );
  Contact c;
  if( GetContactAt( c, 0 ) )
    c.SetURI( curi, 0 );
}

SIPURI SIPMessage::GetContactTopURI()const
{
  GlobalLock();
  Contact c;
  if( !GetContactAt( c, 0 ) )
    return SIPURI();

  ContactURI curi;
  if( ! c.GetURI( curi, 0 )  )
    return SIPURI();

  return curi.GetURI();
}

BOOL SIPMessage::GetContactAt(
  Contact & c,
  PINDEX idx
)const
{
  GlobalLock();
  (const_cast< SIPMessage * >( this ))->ParseContactList();

  if( m_ContactList == NULL )
    return FALSE;

  if( idx >= m_ContactList->GetSize() )
    return FALSE;

  c = (Contact&)m_ContactList->GetAt( idx );

  return TRUE;
}

PINDEX SIPMessage::GetContactSize()const
{
  GlobalLock();
  (const_cast< SIPMessage * >( this ))->ParseContactList();

  if( m_ContactList == NULL )
    return 0;

  SIPHeaderC::Collection & contactList = m_ContactList->GetCollection();
  return contactList.GetSize();
}

OString SIPMessage::GetContactTokenAt(
  PINDEX idx1,
  PINDEX idx2
)const
{
  GlobalLock();
  (const_cast< SIPMessage * >( this ))->ParseContactList();

  if( m_ContactList == NULL )
  {
    return OString::Empty();
  }

  return m_ContactList->GetTokenAt( idx1, idx2 );
}

BOOL SIPMessage::RemoveAllContact()
{
  GlobalLock();

  if( m_ContactList != NULL )
  {
    delete m_ContactList;
    m_ContactList = NULL;
  }

  return TRUE;
}

BOOL SIPMessage::SetContactList(
  const ContactList & contactList
)
{
  GlobalLock();

  if( m_ContactList != NULL )
  {
    delete m_ContactList;
    m_ContactList = NULL;
  }

  m_ContactList = new ContactList( contactList );

  return TRUE;
}

ContactList SIPMessage::GetContactList()const
{
  GlobalLock();
   // cs( *this );
  (const_cast< SIPMessage * >( this ))->ParseContactList();

  ContactList contactList;

  if( m_ContactList != NULL )
  {
    contactList = ContactList( *m_ContactList );
    contactList.MakeUnique();
  }

  return contactList;
}


///////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseInReplyToList()
{
  GlobalLock();
   // cs( *this );

  if( m_InReplyToList != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "in-reply-to" )
    {
      if( m_InReplyToList == NULL )
      {
        m_InReplyToList = new InReplyToList();
      }

      InReplyTo nh( h );

      m_InReplyToList->Append( nh );
      msg.RemoveAt( i );
      break;
    }
  }

  return m_InReplyToList != NULL;
}

BOOL SIPMessage::AppendInReplyTo(
  const OString & token,
  PINDEX idx
)
{
  GlobalLock();
  ParseInReplyToList();
    return FALSE;

  if( m_InReplyToList == NULL )
  {
    m_InReplyToList = new InReplyToList();
  }

  return m_InReplyToList->Append( token, idx );
}

BOOL SIPMessage::AppendInReplyTo(
  const InReplyTo & header
)
{
  GlobalLock();
  ParseInReplyToList();

  if( m_InReplyToList == NULL )
  {
    m_InReplyToList = new InReplyToList();
  }

  m_InReplyToList->Append( header );

  return TRUE;
}

InReplyTo & SIPMessage::GetInReplyToAt(
  PINDEX idx
)const
{
  GlobalLock();
  (const_cast< SIPMessage * >( this ))->ParseInReplyToList();
  return (InReplyTo&)m_InReplyToList->GetAt( idx );
}

OString SIPMessage::GetInReplyToTokenAt(
  PINDEX idx1,
  PINDEX idx2
)const
{
  GlobalLock();
  (const_cast< SIPMessage * >( this ))->ParseInReplyToList();

  if( m_InReplyToList == NULL )
  {
    return OString::Empty();
  }

  return m_InReplyToList->GetTokenAt( idx1, idx2 );
}

BOOL SIPMessage::RemoveAllInReplyTo()
{
  GlobalLock();

  if( m_InReplyToList != NULL )
  {
    delete m_InReplyToList;
    m_InReplyToList = NULL;
  }

  return TRUE;
}

///////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseProxyRequireList()
{
  GlobalLock();
   // cs( *this );

  if( m_ProxyRequireList != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "proxy-require" )
    {
      if( m_ProxyRequireList == NULL )
      {
        m_ProxyRequireList = new ProxyRequireList();
      }

      ProxyRequire nh( h );

      m_ProxyRequireList->Append( nh );
      msg.RemoveAt( i );
      break;
    }
  }

  return m_ProxyRequireList != NULL;
}

BOOL SIPMessage::AppendProxyRequire(
  const OString & token,
  PINDEX idx
)
{
  GlobalLock();
  ParseProxyRequireList();
    return FALSE;

  if( m_ProxyRequireList == NULL )
  {
    m_ProxyRequireList = new ProxyRequireList();
  }

  return m_ProxyRequireList->Append( token, idx );
}

BOOL SIPMessage::AppendProxyRequire(
  const ProxyRequire & header
)
{
  GlobalLock();
  ParseProxyRequireList();

  if( m_ProxyRequireList == NULL )
  {
    m_ProxyRequireList = new ProxyRequireList();
  }

  m_ProxyRequireList->Append( header );

  return TRUE;
}

ProxyRequire & SIPMessage::GetProxyRequireAt(
  PINDEX idx
)const
{
  GlobalLock();
  (const_cast< SIPMessage * >( this ))->ParseProxyRequireList();
  return (ProxyRequire&)m_ProxyRequireList->GetAt( idx );
}

OString SIPMessage::GetProxyRequireTokenAt(
  PINDEX idx1,
  PINDEX idx2
)const
{
  GlobalLock();
  (const_cast< SIPMessage * >( this ))->ParseProxyRequireList();

  if( m_ProxyRequireList == NULL )
  {
    return OString::Empty();
  }

  return m_ProxyRequireList->GetTokenAt( idx1, idx2 );
}

BOOL SIPMessage::RemoveAllProxyRequire()
{
  GlobalLock();

  if( m_ProxyRequireList != NULL )
  {
    delete m_ProxyRequireList;
    m_ProxyRequireList = NULL;
  }

  return TRUE;
}

///////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseRequireList()
{
  GlobalLock();
   // cs( *this );

  if( m_RequireList != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "require" )
    {
      if( m_RequireList == NULL )
      {
        m_RequireList = new RequireList();
      }

      Require nh( h );

      m_RequireList->Append( nh );
      msg.RemoveAt( i );
      break;
    }
  }

  return m_RequireList != NULL;
}

BOOL SIPMessage::AppendRequire(
  const OString & token,
  PINDEX idx
)
{
  GlobalLock();
  ParseRequireList();
    return FALSE;

  if( m_RequireList == NULL )
  {
    m_RequireList = new RequireList();
  }

  return m_RequireList->Append( token, idx );
}

BOOL SIPMessage::AppendRequire(
  const Require & header
)
{
  GlobalLock();
  ParseRequireList();

  if( m_RequireList == NULL )
  {
    m_RequireList = new RequireList();
  }

  m_RequireList->Append( header );

  return TRUE;
}

Require & SIPMessage::GetRequireAt(
  PINDEX idx
)const
{
  GlobalLock();
  (const_cast< SIPMessage * >( this ))->ParseRequireList();
  return (Require&)m_RequireList->GetAt( idx );
}

OString SIPMessage::GetRequireTokenAt(
  PINDEX idx1,
  PINDEX idx2
)const
{
  GlobalLock();
  (const_cast< SIPMessage * >( this ))->ParseRequireList();

  if( m_RequireList == NULL )
  {
    return OString::Empty();
  }

  return m_RequireList->GetTokenAt( idx1, idx2 );
}

BOOL SIPMessage::RemoveAllRequire()
{
  GlobalLock();

  if( m_RequireList != NULL )
  {
    delete m_RequireList;
    m_RequireList = NULL;
  }

  return TRUE;
}

///////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseSupportedList()
{
  GlobalLock();
   // cs( *this );

  if( m_SupportedList != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "supported" )
    {
      if( m_SupportedList == NULL )
      {
        m_SupportedList = new SupportedList();
      }

      Supported nh( h );

      m_SupportedList->Append( nh );
      msg.RemoveAt( i );
      break;
    }
  }

  return m_SupportedList != NULL;
}

BOOL SIPMessage::AppendSupported(
  const OString & token,
  PINDEX idx
)
{
  GlobalLock();
  ParseSupportedList();
    return FALSE;

  if( m_SupportedList == NULL )
  {
    m_SupportedList = new SupportedList();
  }

  return m_SupportedList->Append( token, idx );
}

BOOL SIPMessage::AppendSupported(
  const Supported & header
)
{
  GlobalLock();
  ParseSupportedList();

  if( m_SupportedList == NULL )
  {
    m_SupportedList = new SupportedList();
  }

  m_SupportedList->Append( header );

  return TRUE;
}

OStringArray SIPMessage::GetSupported()const
{
  GlobalLock();

  (const_cast< SIPMessage * >( this ))->ParseSupportedList();

  OStringArray supportedList;

  if( m_SupportedList == NULL ) 
    return supportedList; // return empty array

  for( PINDEX i = 0; i < m_SupportedList->GetSize(); i++ )
  {
    Supported & s = GetSupportedAt( i );
    for( PINDEX j = 0; j < s.GetSize(); j++ )
    {
      supportedList.AppendString( s.GetEntry( j ) );
    }
  }

  return supportedList;
}

BOOL SIPMessage::IsSupportedExtension( 
  const OString & extensionName 
)const
{
  GlobalLock();

  OStringArray supported = GetSupported();
  if( supported.GetSize() > 0 )
  {
    for( PINDEX i = 0; i < supported.GetSize(); i++ )
    {
      if( supported[i] *= extensionName )
        return TRUE;
    }
  }

  return FALSE;
}

Supported & SIPMessage::GetSupportedAt(
  PINDEX idx
)const
{
  GlobalLock();
  (const_cast< SIPMessage * >( this ))->ParseSupportedList();
  return (Supported&)m_SupportedList->GetAt( idx );
}

OString SIPMessage::GetSupportedTokenAt(
  PINDEX idx1,
  PINDEX idx2
)const
{
  GlobalLock();
  (const_cast< SIPMessage * >( this ))->ParseSupportedList();

  if( m_SupportedList == NULL )
  {
    return OString::Empty();
  }

  return m_SupportedList->GetTokenAt( idx1, idx2 );
}

BOOL SIPMessage::RemoveAllSupported()
{
  GlobalLock();

  if( m_SupportedList != NULL )
  {
    delete m_SupportedList;
    m_SupportedList = NULL;
  }

  return TRUE;
}





///////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseUnsupportedList()
{
  GlobalLock();
   // cs( *this );

  if( m_UnsupportedList != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "unsupported" )
    {
      if( m_UnsupportedList == NULL )
      {
        m_UnsupportedList = new UnsupportedList();
      }

      Unsupported nh( h );

      m_UnsupportedList->Append( nh );
      msg.RemoveAt( i );
      break;
    }
  }

  return m_UnsupportedList != NULL;
}

BOOL SIPMessage::AppendUnsupported(
  const OString & token,
  PINDEX idx
)
{
  GlobalLock();
  ParseUnsupportedList();
    return FALSE;

  if( m_UnsupportedList == NULL )
  {
    m_UnsupportedList = new UnsupportedList();
  }

  return m_UnsupportedList->Append( token, idx );
}

BOOL SIPMessage::AppendUnsupported(
  const Unsupported & header
)
{
  GlobalLock();
  ParseUnsupportedList();

  if( m_UnsupportedList == NULL )
  {
    m_UnsupportedList = new UnsupportedList();
  }

  m_UnsupportedList->Append( header );

  return TRUE;
}

Unsupported & SIPMessage::GetUnsupportedAt(
  PINDEX idx
)const
{
  GlobalLock();
  (const_cast< SIPMessage * >( this ))->ParseUnsupportedList();
  return (Unsupported&)m_UnsupportedList->GetAt( idx );
}

OString SIPMessage::GetUnsupportedTokenAt(
  PINDEX idx1,
  PINDEX idx2
)const
{
  GlobalLock();
  (const_cast< SIPMessage * >( this ))->ParseUnsupportedList();

  if( m_UnsupportedList == NULL )
  {
    return OString::Empty();
  }

  return m_UnsupportedList->GetTokenAt( idx1, idx2 );
}

BOOL SIPMessage::RemoveAllUnsupported()
{
  GlobalLock();

  if( m_UnsupportedList != NULL )
  {
    delete m_UnsupportedList;
    m_UnsupportedList = NULL;
  }

  return TRUE;
}

///////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseRouteList()
{
  GlobalLock();
   // cs( *this );

  if( m_RouteList != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "route" )
    {
      if( m_RouteList == NULL )
      {
        m_RouteList = new RouteList();
      }

      OStringArray tokens = h.GetMimeValue().Tokenise( "," );

      for( PINDEX j = 0; j < tokens.GetSize(); j++ )
      {
        Route nh( tokens[j].Trim() );
        m_RouteList->Append( nh, TRUE );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_RouteList != NULL;
}

///////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseRecordRouteList()
{
  GlobalLock();
   // cs( *this );

  if( m_RecordRouteList != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "record-route" )
    {
      if( m_RecordRouteList == NULL )
      {
        m_RecordRouteList = new RecordRouteList();
      }

      OStringArray tokens = h.GetMimeValue().Tokenise( "," );

      for( PINDEX j = 0; j < tokens.GetSize(); j++ )
      {
        RecordRoute nh( tokens[j].Trim() );
        m_RecordRouteList->Append( nh, TRUE );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_RecordRouteList != NULL;
}

////////////////////////////////////////////////////

BOOL SIPMessage::ParsePathList()
{
  GlobalLock();
   // cs( *this );

  if( m_PathList != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "Path" )
    {
      if( m_PathList == NULL )
      {
        m_PathList = new PathList();
      }

      Path nh( h );

      m_PathList->Append( nh );
      msg.RemoveAt( i );
      break;
    }
  }

  return m_PathList != NULL;
}

BOOL SIPMessage::ParsePAssertedIdentityList()
{
  GlobalLock();
   // cs( *this );

  if( m_PAssertedIdentityList != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "P-Asserted-Identity" )
    {
      if( m_PAssertedIdentityList == NULL )
      {
        m_PAssertedIdentityList = new ContactList();
      }

      PAssertedIdentity nh( h );

      m_PAssertedIdentityList->Append( nh );
      msg.RemoveAt( i );
      break;
    }
  }

  return m_PAssertedIdentityList != NULL;
}

BOOL SIPMessage::ParsePPreferredIdentityList()
{
  GlobalLock();
   // cs( *this );

  if( m_PPreferredIdentityList != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "P-Preferred-Identity" )
    {
      if( m_PPreferredIdentityList == NULL )
      {
        m_PPreferredIdentityList = new ContactList();
      }

      PPreferredIdentity nh( h );

      m_PPreferredIdentityList->Append( nh );
      msg.RemoveAt( i );
      break;
    }
  }

  return m_PPreferredIdentityList != NULL;
}

////////////////////////////////////////////////////
BOOL SIPMessage::CreateResponse(
  SIPMessage & response,
  WORD statusCode,
  const OString & reasonPhrase,
  BOOL noTag
)const
{
  GlobalLock();
  PAssert( IsRequest(), "Attempting to create a response to a response!!!" );

  StatusLine statusLine( statusCode );
  if( !reasonPhrase.IsEmpty() )
    statusLine.SetReasonPhrase( reasonPhrase );
  StartLine startLine( statusLine );
  response.SetStartLine( startLine );

  To to;
  if( !GetTo(to) )
    return FALSE;
  if( statusCode != 100 && !noTag )
  {
    if( ( IsInvite() || IsSubscribe() ) && GetToTag().IsEmpty() )
    {
      /// only set the tag parameter if its empty
      OString tag = ParserTools::GenTagParameter();
      to.AddParameter( "tag", tag );
    }
  }
  response.SetTo( to );


  /** Clone the from header */
  From from;
  if( !GetFrom( from ) )
    return FALSE;
  response.SetFrom( from );

  /** Clone the CSEQ */
  CSeq cseq;
  if( !GetCSeq( cseq ) )
    return FALSE;
  response.SetCSeq( cseq );

  /** Clone the call ID */
  CallId callId;
  if( !GetCallId( callId ) )
    return FALSE;
  response.SetCallId( callId );

  /*
  Via via;
  if( GetViaSize() == 0 )
    return FALSE;
  GetViaAt( 0, via );
  
  response.AppendVia( via );
  */

  ViaList viaList = GetViaList();
  if( viaList.GetSize() == 0 )
    return FALSE;
  response.SetViaList( viaList );
  
  /// clone the record route if there is any
  RecordRouteList recordRouteList = GetRecordRouteList();
  if( recordRouteList.GetSize() > 0 )
    response.SetRecordRouteList( recordRouteList );

  response.SetContentLength( 0 );

  response.SetEncryption( IsEncrypted() );

  response.Finalize();

  return TRUE;
}


OString SIPMessage::GetMethodString( Methods meth )
{
  static const char * methods[NumMethods] =
  {
      "ACK",
      "BYE",
      "CANCEL",
      "INFO",
      "INVITE",
      "MESSAGE",
      "NOTIFY",
      "OPTIONS",
      "PRACK",
      "PUBLISH",
      "REFER",
      "REGISTER",
      "SUBSCRIBE",
      "UPDATE"
  };

  return methods[meth];
}

BOOL SIPMessage::SetSendAddress(
  const SIPURI & address
)
{
  GlobalLock(); 
  OString port = "5060";
  if( !address.GetPort().IsEmpty() )
    port = address.GetPort();

  OString proto;
  if( !address.GetParameter( "transport", proto ) )
    proto = "udp";

  SetInternalHeader( "SENDADDR", address.GetHost() );
  SetInternalHeader( "SENDPORT", port );
  SetInternalHeader( "SENDPROTO", proto );

  return TRUE;
}


/////////////////////////////////////////////

/////////////////////////
BOOL SIPMessage::ParseReferTo()
{
  GlobalLock();
   // cs( *this );

  if( m_ReferTo != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "Refer-To" )
    {
      if( m_ReferTo == NULL )
      {
        m_ReferTo = new ReferTo( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_ReferTo != NULL;
}

BOOL SIPMessage::GetReferTo(
  ReferTo & h
)const
{
  GlobalLock();
  if( !(const_cast< SIPMessage * >( this ))->ParseReferTo() )
    return FALSE;

  if( !HasReferTo() )
    return FALSE;

  h = *m_ReferTo;

  return TRUE;
}

void SIPMessage::SetReferTo(
  const ReferTo & h
)
{
  GlobalLock();

  if( m_ReferTo != NULL )
    delete m_ReferTo;
  m_ReferTo = (ReferTo*)h.Clone();
}

BOOL SIPMessage::HasReferTo()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParseReferTo();
}

//////////////////////////////////////////////////////////////////////////////

BOOL SIPMessage::ParseEvent()
{
  GlobalLock();
   // cs( *this );

  if( m_Event != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "event" )
    {
      if( m_Event == NULL )
      {
        m_Event = new Event( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_Event != NULL;
}

BOOL SIPMessage::GetEvent(
  Event & h
)const
{
  GlobalLock();
  if( !(const_cast< SIPMessage * >( this ))->ParseEvent() )
    return FALSE;

  if( !HasEvent() )
    return FALSE;

  h = *m_Event;

  return TRUE;
}

void SIPMessage::SetEvent(
  const Event & h
)
{
  GlobalLock();

  if( m_Event != NULL )
    delete m_Event;
  m_Event = (Event*)h.Clone();
}

BOOL SIPMessage::HasEvent()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParseEvent();
}

//////////////////////////////////////////////////////////////////

BOOL SIPMessage::GetSIPETag(
  SIPETag & h
)const
{
  GlobalLock();
  if( !(const_cast< SIPMessage * >( this ))->ParseSIPETag() )
    return FALSE;

  if( !HasSIPETag() )
    return FALSE;

  h = *m_SIPETag;

  return TRUE;
}

OString SIPMessage::GetSIPETag()const
{
  SIPETag h;
  if( !GetSIPETag( h ) )
    return OString::Empty();

  return h.GetHeaderBody();
}

void SIPMessage::SetSIPETag(
  const SIPETag & h
)
{
  GlobalLock();

  if( m_SIPETag != NULL )
    delete m_SIPETag;
  m_SIPETag = (SIPETag*)h.Clone();
}

void SIPMessage::SetSIPETag(
  const OString & etag
)
{
  SIPETag h( etag );
  SetSIPETag( h );
}

BOOL SIPMessage::HasSIPETag()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParseSIPETag();
}
///////////////////////////////////////////////////////////////////////////
BOOL SIPMessage::GetSIPIfMatch(
  SIPIfMatch & h
)const
{
  GlobalLock();
  if( !(const_cast< SIPMessage * >( this ))->ParseSIPIfMatch() )
    return FALSE;

  if( !HasSIPIfMatch() )
    return FALSE;

  h = *m_SIPIfMatch;

  return TRUE;
}

OString SIPMessage::GetSIPIfMatch()const
{
  SIPIfMatch h;
  if( !GetSIPIfMatch( h ) )
    return OString::Empty();

  return h.GetHeaderBody();
}

void SIPMessage::SetSIPIfMatch(
  const SIPIfMatch & h
)
{
  GlobalLock();
  if( m_SIPIfMatch != NULL )
    delete m_SIPIfMatch;
  m_SIPIfMatch = (SIPIfMatch*)h.Clone();
}

void SIPMessage::SetSIPIfMatch(
  const OString & etag
)
{
  SIPIfMatch h( etag );
  SetSIPIfMatch( h );
}

BOOL SIPMessage::HasSIPIfMatch()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParseSIPIfMatch();
}
///////////////////////////////////////////////////////////////////////////
BOOL SIPMessage::ParseAllowEventsList()
{
  GlobalLock();
   // cs( *this );

  if( m_AllowEventsList != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "allow-events" )
    {
      if( m_AllowEventsList == NULL )
      {
        m_AllowEventsList = new AllowEventsList();
      }

      AllowEvents nh( h );

      m_AllowEventsList->Append( nh );
      msg.RemoveAt( i );
      break;
    }
  }

  return m_AcceptList != NULL;
}

BOOL SIPMessage::AppendAllowEvents(
  const OString & token,
  PINDEX idx
)
{
  GlobalLock();
  ParseAllowEventsList();
    return FALSE;

  if( m_AllowEventsList == NULL )
  {
    m_AllowEventsList = new AllowEventsList();
  }

  return m_AllowEventsList->Append( token, idx );
}

BOOL SIPMessage::AppendAllowEvents(
  const AllowEvents & header
)
{
  GlobalLock();
  ParseAllowEventsList();

  if( m_AllowEventsList == NULL )
  {
    m_AllowEventsList = new AllowEventsList();
  }

  m_AllowEventsList->Append( header );

  return TRUE;
}

AllowEvents & SIPMessage::GetAllowEventsAt(
  PINDEX idx
)const
{
  GlobalLock();
  (const_cast< SIPMessage * >( this ))->ParseAllowEventsList();
  return (AllowEvents&)m_AllowEventsList->GetAt( idx );
}

OString SIPMessage::GetAllowEventsTokenAt(
  PINDEX idx1,
  PINDEX idx2
)const
{
  GlobalLock();
  (const_cast< SIPMessage * >( this ))->ParseAllowEventsList();

  if( m_AllowEventsList == NULL )
  {
    return OString::Empty();
  }

  return m_AllowEventsList->GetTokenAt( idx1, idx2 );
}

BOOL SIPMessage::RemoveAllAllowEvents()
{
  GlobalLock();

  if( m_AllowEventsList != NULL )
  {
    delete m_AllowEventsList;
    m_AllowEventsList = NULL;
  }

  return TRUE;
}

///////////////////////////////////////////////////////

BOOL SIPMessage::ParseSubscriptionState()
{
  GlobalLock();
   // cs( *this );

  if( m_SubscriptionState != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "subscription-state" )
    {
      if( m_SubscriptionState == NULL )
      {
        m_SubscriptionState = new SubscriptionState( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_SubscriptionState != NULL;
}

BOOL SIPMessage::GetSubscriptionState(
  SubscriptionState & h
)const
{
  GlobalLock();
  if( !(const_cast< SIPMessage * >( this ))->ParseSubscriptionState() )
    return FALSE;

  if( !HasSubscriptionState() )
    return FALSE;

  h = *m_SubscriptionState;

  return TRUE;
}

void SIPMessage::SetSubscriptionState(
  const SubscriptionState & h
)
{
  GlobalLock();

  if( m_SubscriptionState != NULL )
    delete m_SubscriptionState;
  m_SubscriptionState = (SubscriptionState*)h.Clone();
}

BOOL SIPMessage::HasSubscriptionState()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParseSubscriptionState();
}


BOOL SIPMessage::ParseSIPETag()
{
  GlobalLock();
   // cs( *this );

  if( m_SIPETag != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "sip-etag" )
    {
      if( m_SIPETag == NULL )
      {
        m_SIPETag = new SIPETag( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_SIPETag != NULL;
}

BOOL SIPMessage::ParseSIPIfMatch()
{
  GlobalLock();
   // cs( *this );

  if( m_SIPIfMatch != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "sip-if-match" )
    {
      if( m_SIPIfMatch == NULL )
      {
        m_SIPIfMatch = new SIPIfMatch( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_SIPIfMatch != NULL;
}

BOOL SIPMessage::IsEncrypted()const
{
  return GetInternalHeader( "ENCRYPTED" ) == "TRUE";
}

BOOL SIPMessage::SetEncryption( 
  BOOL willHash
)
{
  if( GetInternalHeader( "ENCRYPTED" ) != "TRUE" && willHash )
  {
    SetInternalHeader( "ENCRYPTED", "TRUE"  );
  }else if( GetInternalHeader( "ENCRYPTED" ) == "TRUE" && !willHash )
  {
    SIPHeader encrypted( "ENCRYPTED" );
    PopInternalHeader( "ENCRYPTED", encrypted );
  }

  return TRUE;
}

BOOL SIPMessage::GetEncrypted( 
  PBYTEArray & out 
)const
{
  OStringStream _buff;
  _buff << *this;
  OString buff( _buff );
  
  PBYTEArray in;
  in.SetSize( buff.GetSize() );
  ::memcpy( in.GetPointer(), buff.c_str(), buff.GetSize() );
  
  return Encryption::Engine::Encrypt( in, out );
}

OString SIPMessage::GetInterfaceAddress()const
{
  OString ifaceAddress;
  if( HasInternalHeader( "IFACE-ADDR" ) )
    ifaceAddress = GetInternalHeader( "IFACE-ADDR" );
  return ifaceAddress;
}

void SIPMessage::SetInterfaceAddress(
  const OString & ifaceAddress
)
{
  if( ifaceAddress.IsEmpty() )
    return;

  if( HasInternalHeader( "IFACE-ADDR" ) )
    SetInternalHeader( "IFACE-ADDR", ifaceAddress );
  else
    AddInternalHeader( "IFACE-ADDR", ifaceAddress );
}

WORD SIPMessage::GetInterfacePort()const
{
  OString port = "5060";
  if( HasInternalHeader( "IFACE-PORT" ) )
    port = GetInternalHeader( "IFACE-PORT" );
  return port.AsUnsigned();
}

void SIPMessage::SetInterfacePort(
  WORD port
)
{
  if( port == 0 )
    return;

  if( HasInternalHeader( "IFACE-PORT" ) )
    SetInternalHeader( "IFACE-PORT", OString( port ) );
  else
    AddInternalHeader( "IFACE-PORT", OString( port ) );
}

OString SIPMessage::GetUACoreName()const
{
  OString coreName;
  if( HasInternalHeader( "UA-CORE-NAME" ) && coreName.IsEmpty() )
    coreName = GetInternalHeader( "UA-CORE-NAME" );
  return coreName;
}

void SIPMessage::SetUACoreName(
  const OString & coreName
)
{
  if( coreName.IsEmpty() )
    return;

  if( HasInternalHeader( "UA-CORE-NAME" ) )
    SetInternalHeader( "UA-CORE-NAME", coreName );
  else
    AddInternalHeader( "UA-CORE-NAME", coreName );
}


BOOL SIPMessage::IsValid( BOOL minimal )const
{
  GlobalLock();

  StartLine startLine;
  if( !GetStartLine( startLine ) )
    return FALSE;

  if( GetCSeqMethod().IsEmpty() )
    return FALSE;

  if( GetCallId().IsEmpty() )
    return FALSE;

  if( !minimal )
  {
    From from;
    if( !GetFrom( from ) )
      return FALSE;
    else if( from.GetURI().GetHost().IsEmpty() )
      return FALSE;

    To to;
    if( !GetTo( to ) )
      return FALSE;
    else if( to.GetURI().GetHost().IsEmpty() )
      return FALSE;

    if( GetViaSize() == 0 )
      return FALSE;

  }

  return TRUE;
  
}


BOOL SIPMessage::GetMinSE(
  MinSE & h
)const
{
  GlobalLock();
  if( !(const_cast< SIPMessage * >( this ))->ParseMinSE() )
    return FALSE;

  if( !HasMinSE() )
    return FALSE;

  h = *m_MinSE;

  return TRUE;
}

void SIPMessage::SetMinSE(
  const MinSE & h
)
{
  GlobalLock();

  if( m_MinSE != NULL )
    delete m_MinSE;

  m_MinSE = (MinSE*)h.Clone();
}

BOOL SIPMessage::HasMinSE()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParseMinSE();
}

BOOL SIPMessage::GetSessionExpires(
  SessionExpires & h
)const
{
  GlobalLock();
  if( !(const_cast< SIPMessage * >( this ))->ParseSessionExpires() )
    return FALSE;

  if( !HasSessionExpires() )
    return FALSE;

  h = *m_SessionExpires;

  return TRUE;
}

void SIPMessage::SetSessionExpires(
  const SessionExpires & h
)
{
  GlobalLock();

  if( m_SessionExpires != NULL )
    delete m_SessionExpires;

  m_SessionExpires = (SessionExpires*)h.Clone();
}

BOOL SIPMessage::HasSessionExpires()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParseSessionExpires();
}

BOOL SIPMessage::ParseMinSE()
{
  GlobalLock();

  if( m_MinSE != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "Min-SE" )
    {
      if( m_MinSE== NULL )
      {
         m_MinSE = new MinSE( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_MinSE!= NULL;
}

BOOL SIPMessage::ParseSessionExpires()
{
  GlobalLock();

  if( m_SessionExpires != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "Session-Expires" )
    {
      if( m_SessionExpires== NULL )
      {
        m_SessionExpires= new SessionExpires( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_SessionExpires!= NULL;
}


/////////////////////////
BOOL SIPMessage::ParseReplaces()
{
  GlobalLock();
   // cs( *this );

  if( m_Replaces != NULL )
    return TRUE;

  if( m_RawMessage == NULL )
    return FALSE;

  OStringArray & msg = *m_RawMessage;

  if( m_StartLine == NULL )
  {
    m_StartLine = new StartLine( msg[0] );
    msg.RemoveAt( 0 );
  }

  for( PINDEX i = 0; i < msg.GetSize(); i++ )
  {
    MimeHeader h( msg[i] );
    OString hname = h.GetMimeName();
    if( hname *= "replaces" )
    {
      if( m_Replaces == NULL )
      {
        m_Replaces = new Replaces( h );
      }

      msg.RemoveAt( i );
      break;
    }
  }

  return m_Replaces != NULL;
}

BOOL SIPMessage::GetReplaces(
  Replaces & h
)const
{
  GlobalLock();
  if( !(const_cast< SIPMessage * >( this ))->ParseReplaces() )
    return FALSE;

  if( !HasReplaces() )
    return FALSE;

  h = *m_Replaces;

  return TRUE;
}


void SIPMessage::SetReplaces(
  const Replaces & h
)
{
  GlobalLock();

  if( m_Replaces != NULL )
    delete m_Replaces;
  m_Replaces = (Replaces*)h.Clone();
}

BOOL SIPMessage::HasReplaces()const
{
  GlobalLock();
  return (const_cast< SIPMessage * >( this ))->ParseReplaces();
}

