/*
 *
 * SoftPhoneInterface.cxx
 * 
 * Open SIP Stack ( OSS )
 *
 * Copyright (C) 2006 Joegen E. Baclor. All Rights Reserved.
 *
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is OpenSIPStack Library.
 *
 * The Initial Developer of the Original Code is Joegen E. Baclor.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * $Log: SoftPhoneInterface.cxx,v $
 * Revision 1.40  2007/07/15 12:33:24  joegenbaclor
 * Corrected bug in calling OnIncomingSDPOffer where it is called for the second leg instead of the first leg call session.  This is due to the change in placement of the method form receipt of the invite to SetupOutbound
 *
 * Revision 1.39  2007/07/14 09:49:48  joegenbaclor
 * Corrected compile errors in linux
 *
 * Revision 1.38  2007/07/14 07:50:34  joegenbaclor
 * Migrated code to use new STL based string class
 *
 * Revision 1.37  2007/06/19 09:25:25  ijpinzon
 * Invoke InitializeSIP before registration and calling
 *
 * Revision 1.36  2007/05/30 07:34:46  ijpinzon
 * 1. Exposed setting of silence detection mode
 * 2. Exposed setting of audio jitter delay
 *
 * Revision 1.35  2007/05/22 05:21:17  ijpinzon
 * 1. Transferred setting of mediaFormatMask and mediaFormatOrder to SoftPhoneManager::ApplyMediaInfo()
 * 2. Exposed setting of encryption mode and encryption key in SetEnableEncryption
 *
 * Revision 1.34  2007/05/08 07:06:40  ijpinzon
 * 1. Implemented ring device
 * 2. Extended tone playback functions to support playing on a custom sound channel
 *
 * Revision 1.33  2007/04/23 11:31:49  ijpinzon
 * 1.  Modified SetVolume to support the new volume control in SoftphoneManager
 * 2.  Implemented GetVolume for SoftphoneInterface
 *
 * Revision 1.32  2007/04/22 07:49:41  joegenbaclor
 * Corrected bug in RFC 2833 tramsission where the ned event flag is never sent
 *
 * Revision 1.31  2007/04/13 09:41:55  joegenbaclor
 * Added Allow header in INVITE and REGISTER
 *
 * Revision 1.30  2007/04/10 06:13:08  joegenbaclor
 * Added new Encryption Mode
 *
 * Revision 1.29  2007/03/07 04:39:05  joegenbaclor
 * Added remote URI in disconnect event info
 *
 * Revision 1.28  2007/03/04 03:00:34  joegenbaclor
 * Removed Linux compile warnings
 *
 * Revision 1.27  2007/02/19 11:31:14  joegenbaclor
 * more privacy work
 *
 * Revision 1.26  2007/02/12 12:36:57  joegenbaclor
 * Some aesthetic work
 *
 * Revision 1.25  2007/02/11 10:44:38  joegenbaclor
 * Added SetOriginAddress when translating SDP
 *
 * Revision 1.24  2007/02/10 03:03:20  joegenbaclor
 * Added SetInterfaceAddress() function to SoftPhoneManager
 *
 * Revision 1.23  2007/02/04 15:04:11  joegenbaclor
 * Completed transaction layer TCP support
 *
 * Revision 1.22  2007/02/02 06:55:47  joegenbaclor
 * Added OnTerminated Event and syncrhonous Unregister
 *
 * Revision 1.21  2007/01/26 01:45:57  joegenbaclor
 * Changed name of GetAccountBalance to RPCGetAccountBalance
 *
 * Revision 1.20  2007/01/14 13:12:57  joegenbaclor
 * Fixed compile warnings
 *
 * Revision 1.19  2007/01/01 05:12:53  joegenbaclor
 * Introduced termiantion flag to avoid crashes on exit.
 * TO DO: Maybe the best thing to do here is to have a WaitForTermination()
 * function exposed for the SIPStack
 *
 * Revision 1.18  2006/12/19 23:38:34  joegenbaclor
 * Fixed bug in softphone proxy authentication where passwords <= 3 in length may not be properly handled
 *
 * Revision 1.17  2006/12/18 10:31:35  joegenbaclor
 * Fixed ACM G.723.1 one way audio bug
 *
 * Revision 1.16  2006/12/17 23:36:04  joegenbaclor
 * Added cleanup method
 *
 * Revision 1.15  2006/12/14 09:49:58  joegenbaclor
 * Tweaked cleanup routines
 *
 * Revision 1.14  2006/12/13 10:01:26  joegenbaclor
 * More work on Softphone
 *
 * Revision 1.13  2006/12/12 23:53:34  joegenbaclor
 * More work on DTMF encoding
 *
 * Revision 1.12  2006/12/12 05:26:10  joegenbaclor
 * Modified DTMF player to use PDTMFEncoder
 *
 * Revision 1.11  2006/12/11 14:35:43  joegenbaclor
 * Allowed setting of empty STUN address
 *
 * Revision 1.10  2006/12/11 10:27:37  joegenbaclor
 * moved config object creation to constructor
 *
 * Revision 1.9  2006/12/08 09:56:13  joegenbaclor
 * Added Hold functions
 *
 * Revision 1.8  2006/12/07 09:12:22  joegenbaclor
 * 1.  Changed registration expire value from milliseconds to seconds
 *
 * Revision 1.7  2006/12/06 10:17:12  joegenbaclor
 * More SF enhancement
 *
 * Revision 1.6  2006/12/06 06:40:56  joegenbaclor
 * STUN related fixes
 *
 * Revision 1.5  2006/12/06 01:59:22  joegenbaclor
 * More Softphone support
 *
 * Revision 1.4  2006/12/04 08:55:32  joegenbaclor
 * Added log hook to transport
 *
 * Revision 1.3  2006/12/04 07:32:50  joegenbaclor
 * Added RFC3680 methods to Opal classes
 *
 * Revision 1.2  2006/12/03 07:32:56  joegenbaclor
 * Softphone is now configured from INI files
 *
 * Revision 1.1  2006/12/02 01:12:13  joegenbaclor
 * Initial upload of SoftPhoneInterface classes
 *
 *
 */


#include "SoftPhoneInterface.h"
#include "SoftPhone.h"
#include "Encryption.h"
#include "ptclib/dtmf.h"



using namespace SF;
using namespace MS;

#define new PNEW

static SoftPhoneManager * m_SoftPhone = NULL;
static PConfig * m_Config = NULL;
static PMutex m_ConfigMutex;
static PConfig * m_Contacts = NULL;
static PConfig * m_Accounts = NULL;
static PConfig * m_CallHistory = NULL;

static MS::VoiceFile * m_DialToneFile = NULL;
static MS::VoiceFile * m_RingBackToneFile = NULL;
static MS::VoiceFile * m_BusyToneFile = NULL;

SoftPhoneInterface::SoftPhoneInterface()
{
  m_NATType = NAT::SIPSTUNClient::NumNatTypes;
  SIPTransport::AttachWriteNotifier( PCREATE_NOTIFIER( _OnWritePacketLog ) );
  SIPTransport::AttachReadNotifier( PCREATE_NOTIFIER( _OnReadPacketLog ) );

  m_Config = new PConfig( "default.ini", "Account" );
  m_Contacts = new PConfig( "contacts.ini", "Contacts" );
  m_Accounts = new PConfig( "accounts.ini", "Accounts" );
  m_CallHistory = new PConfig( "calls.ini", "Calls" );
  m_SoftPhone = new SoftPhoneManager( this );

  m_PlayBusyTone = FALSE;
  m_PlayDialTone = FALSE;
  m_PlayRingBackTone = FALSE;

  static PDTMFEncoder dt;
  for( PINDEX i = 0; i < 10; i++ )
    dt.GenerateDialTone();

  m_DialToneFile = new MS::VoiceFile( dt, MS::VoiceFile::Wav_8khz_Mono, "TONE_US" );

  static PDTMFEncoder rt;
  rt.GenerateRingBackTone();
  m_RingBackToneFile = new MS::VoiceFile( rt, MS::VoiceFile::Wav_8khz_Mono, "RING_US" );

  static PDTMFEncoder bt;
  bt.GenerateBusyTone();
  m_BusyToneFile = new MS::VoiceFile( bt, MS::VoiceFile::Wav_8khz_Mono, "BUSY_US" );

}

SoftPhoneInterface::~SoftPhoneInterface()
{
  CleanUp();
}

void SoftPhoneInterface::Terminate()
{
  if( m_SoftPhone != NULL )
    m_SoftPhone->Terminate();
}
void SoftPhoneInterface::CleanUp()
{
  if( m_DialToneFile != NULL )
  {
    delete m_DialToneFile;
    m_DialToneFile = NULL;
  }

  if( m_RingBackToneFile != NULL )
  {
    delete m_RingBackToneFile;
    m_RingBackToneFile = NULL;
  }


  if( m_BusyToneFile != NULL )
  {
    delete m_BusyToneFile;
    m_BusyToneFile = NULL;
  }

  if( m_SoftPhone != NULL )
  {
    delete m_SoftPhone;
    m_SoftPhone = NULL;
  }
  
  if( m_Config != NULL )
  {
    delete m_Config;
    m_Config = NULL;
  }

  if( m_Contacts != NULL )
  {
    delete m_Contacts;
    m_Contacts = NULL;
  }

  if( m_Accounts != NULL )
  {
    delete m_Accounts;
    m_Accounts = NULL;
  }

  if( m_CallHistory != NULL )
  {
    delete m_CallHistory;
    m_CallHistory = NULL;
  }
  
}

PConfig * SoftPhoneInterface::GetMainConfig(){ return m_Config; };
PConfig * SoftPhoneInterface::GetContacts(){ return m_Contacts; };
PConfig * SoftPhoneInterface::GetAccounts(){ return m_Accounts; };
PConfig * SoftPhoneInterface::GetCallHistory(){ return m_CallHistory; };
SoftPhoneManager * SoftPhoneInterface::GetPhoneManager(){ return m_SoftPhone; };

/// Initialize SIP and PC Sound System
BOOL SoftPhoneInterface::Initialize()
{
  m_SoftPhone->SetDisplayName( m_Config->GetString( "Account", "AccountName", "" ) );
  
  SIPURI accountAddress( (const char *)m_Config->GetString( "Account", "AccountAddress", "" ) );
  OString accountHost = accountAddress.GetHost();
  OString userName = accountAddress.GetUser();

  m_SoftPhone->SetSIPDomain( accountHost );
  m_SoftPhone->SetUserName( userName );
  
  SIPURI registrarAddress( (const char *)m_Config->GetString( "Account", "ServerAddress", "" ) );
  if( accountHost.IsEmpty() && !registrarAddress.GetHost().IsEmpty() )
    m_SoftPhone->SetSIPDomain( registrarAddress.GetHost() );

  m_SoftPhone->SetRegistrarAddress( registrarAddress.AsString() );
  m_SoftPhone->SetProxyAddress( m_Config->GetString( "Account", "OutboundProxy", "" ) );
  m_SoftPhone->SetRegistrarUserName( m_Config->GetString( "Account", "AuthenticationUser", "" ) );
  m_SoftPhone->SetRegistrarPassword( m_SoftPhone->GetRegistrarPassword() );
  m_SoftPhone->SetProxyUserName( m_Config->GetString( "Account", "ProxyAuthenticationUser", "" ) );
  m_SoftPhone->SetProxyPassword( m_SoftPhone->GetProxyPassword() );
  m_SoftPhone->SetRecordDevice( m_Config->GetString( "Audio", "Microphone", PSoundChannel::GetDefaultDevice( PSoundChannel::Recorder ) ) );
  m_SoftPhone->SetPlayDevice( m_Config->GetString( "Audio", "Speaker", PSoundChannel::GetDefaultDevice( PSoundChannel::Player ) ) );
  m_SoftPhone->SetEncryption( m_Config->GetBoolean( "Account", "Encrypted", FALSE ) );
  m_SoftPhone->SetPrivacy( m_Config->GetBoolean( "Account", "Privacy", FALSE ) );
  m_SoftPhone->SetRegistrarTTL( m_Config->GetInteger( "Account", "RegistrarTTL", 3600 ) );

  PStringArray formatMask;

  for( PINDEX i = 0; i < 20; i++ )
  {
    PStringStream codecSlot;
    codecSlot << "Codec" << i;
    OString codec = m_Config->GetString( "Audio", codecSlot, "" );
    if( !codec.IsEmpty() )
      formatMask.AppendString( codec );

  }

  if( formatMask.GetSize() == 0 )
  {

    m_Config->SetString( "Audio", "Codec0", "iLBC-13k3" );
    formatMask.AppendString( "iLBC-13k3" );

    m_Config->SetString( "Audio", "Codec1", "GSM-06.10" );
    formatMask.AppendString( "GSM-06.10" );
  }
    
  /// Initialize OPAL PC Sound System
  m_SoftPhone->Initialise();


  /// Initialize the SIP Stack
  m_SoftPhone->InitializeSIP();

  return TRUE;
}

/// Perform Registration
BOOL SoftPhoneInterface::DoLogin()
{
  m_SoftPhone->InitializeSIP();
  return m_SoftPhone->SendRegister();
}

/// Perform unregistration
BOOL SoftPhoneInterface::DoLogout()
{
  return m_SoftPhone->SendUnregister();
}


/// add a contact to the address book
BOOL SoftPhoneInterface::AddContact(
  const OString & displayName, 
  const OString & sipURI, 
  BOOL queryPresence
)
{
  PWaitAndSignal lock( m_ConfigMutex );

  PStringStream record;
  record << "\"" << displayName << "\" "; 
  record << "<" <<  sipURI << ">";
  record << ";query-presence=" << queryPresence;

  PMessageDigest5 digestor;
  PMessageDigest5::Code digest; 
  digestor.Start();
  digestor.Process( displayName );
  digestor.Complete( digest );
  MD5::MD5Hash hashKey( digest );
  
  m_Contacts->SetString( hashKey.AsString(), record );

  return TRUE;
}
  
/// returns a contact by index
BOOL SoftPhoneInterface::GetContact( 
  PINDEX index, 
  OString & displayName, 
  OString & sipURI, 
  BOOL & queryPresence
)const
{
  PWaitAndSignal lock( m_ConfigMutex );

  PStringToString contacts = m_Contacts->GetAllKeyValues();
  OString contact = m_Contacts->GetString( contacts.GetKeyAt( index ), "" );

  if( !contact.IsEmpty() )
  {
    PStringStream record;
    record << "From: " << contact;
    From from( (const char *)record );

    displayName = from.GetDisplayName();
    SIPParser::ParserTools::UnQuote( displayName );

    sipURI = from.GetURI().AsString( FALSE );

    OString _queryPresence;
    if( from.GetParameter( "query-presence", _queryPresence ) )
      queryPresence = (BOOL)_queryPresence.AsInteger();
  }

  return TRUE;
}

BOOL SoftPhoneInterface::GetContactByName(
  const OString & accountName, 
  OString & contactAddress, 
  BOOL & queryPresence
)const
{
  PWaitAndSignal lock( m_ConfigMutex );

  PMessageDigest5 digestor;
  PMessageDigest5::Code digest; 
  digestor.Start();
  digestor.Process( accountName );
  digestor.Complete( digest );
  MD5::MD5Hash hashKey( digest );

  PStringStream record;
  record << "From: " << m_Contacts->GetString( hashKey.AsString(), "" );
  From from( (const char *)record );

  contactAddress = from.GetURI().AsString( FALSE );

  OString query;
  if( from.GetParameter( "query-presence", query ) )
    queryPresence = (BOOL)query.AsInteger();

  return TRUE;
}

/// returns the size of the address book
PINDEX SoftPhoneInterface::GetContactSize()const
{
  PWaitAndSignal lock( m_ConfigMutex );
  PStringToString contacts = m_Contacts->GetAllKeyValues();
  return contacts.GetSize();
}

/// deletes a contact by name
BOOL SoftPhoneInterface::DeleteContact( 
  const OString & contact
)
{
  PWaitAndSignal lock( m_ConfigMutex );

  PMessageDigest5 digestor;
  PMessageDigest5::Code digest; 
  digestor.Start();
  digestor.Process( contact );
  digestor.Complete( digest );
  MD5::MD5Hash hashKey( digest );

  m_Contacts->DeleteKey( hashKey.AsString() );

  return TRUE;
}

/// deletes a contact using the index number 
BOOL SoftPhoneInterface::DeleteContactByIndex(
  PINDEX index
)
{
  PWaitAndSignal lock( m_ConfigMutex );
  PStringToString contacts = m_Contacts->GetAllKeyValues();
  m_Contacts->DeleteKey( contacts.GetKeyAt( index ) );
  return TRUE;
}
  
/// returns the name of this account.
/// this is also the displayname sent
OString SoftPhoneInterface::GetAccountName()const
{
  PWaitAndSignal lock( m_ConfigMutex );
  return (const char *)m_Config->GetString( "Account", "AccountName", "" ); 
}

/// sets the name of this account.
/// this is also the displayname sent
void SoftPhoneInterface::SetAccountName(
  const OString & accountName
)
{
  PWaitAndSignal lock( m_ConfigMutex );
  m_Config->SetString( "Account", "AccountName", accountName );
  m_SoftPhone->SetDisplayName( accountName );
}

/// returns the SIP URI for the softphone
/// this is also the SIP URI the softphone
/// sends for the TO header
OString SoftPhoneInterface::GetAccountAddress()const
{
  PWaitAndSignal lock( m_ConfigMutex );
  return (const char *)m_Config->GetString( "Account", "AccountAddress", "" ); 
}

/// sets the SIP URI for the softphone
/// this is also the SIP URI the softphone
/// sends for the TO header
void SoftPhoneInterface::SetAccountAddress(
  const OString & accountAddress
)
{
  PWaitAndSignal lock( m_ConfigMutex );

  m_Config->SetString( "Account", "AccountAddress", accountAddress );
  SIPURI uri( accountAddress );
  m_SoftPhone->SetUserName( uri.GetUser() );
  m_SoftPhone->SetSIPDomain( uri.GetHost() );
}

/// returns the address of the Registrar
/// this is the address where registrations 
/// will be sent
OString SoftPhoneInterface::GetSIPServerAddress()const
{
  PWaitAndSignal lock( m_ConfigMutex );
  return (const char *)m_Config->GetString( "Account", "ServerAddress", "" );
}

/// returns the address of the Registrar
/// this is the address where registrations 
/// will be sent
void SoftPhoneInterface::SetSIPServerAddress( 
  const OString & serverAddress 
)
{
  PWaitAndSignal lock( m_ConfigMutex );
  m_Config->SetString( "Account", "ServerAddress", serverAddress );
  m_SoftPhone->SetRegistrarAddress( serverAddress );
}


/// returns the address for the outbound proxy
OString SoftPhoneInterface::GetOutboundProxy()const
{
  PWaitAndSignal lock( m_ConfigMutex );
  return (const char *)m_Config->GetString( "Account", "OutboundProxy", "" ); 
}

/// returns the address for the outbound proxy
void SoftPhoneInterface::SetOutboundProxy(
  const OString & proxyAddress
)
{
  m_Config->SetString( "Account", "OutboundProxy", proxyAddress );
  m_SoftPhone->SetProxyAddress( proxyAddress );
}

/// returns the user name sent by the softphone
/// when sending authentication and authorization
/// responses
OString SoftPhoneInterface::GetAuthenticationUser()const
{
  PWaitAndSignal lock( m_ConfigMutex );
  return (const char *)m_Config->GetString( "Account", "AuthenticationUser", "" ); 
}

/// sets the user name sent by the softphone
/// when sending authentication and authorization
/// responses
void SoftPhoneInterface::SetAuthenticationUser(
  const OString & authUser
)
{
  PWaitAndSignal lock( m_ConfigMutex );
  m_Config->SetString( "Account", "AuthenticationUser", authUser );
  m_SoftPhone->SetRegistrarUserName( authUser );
}

/// returns the password sent by the softphone
/// when sending authentication and authorization
/// responses
OString SoftPhoneInterface::GetAuthenticationPassword()const
{
  PWaitAndSignal lock( m_ConfigMutex );
  OString pwd = PBase64::Decode( m_Config->GetString( "Account", "AuthenticationPassword", "-----" ) ); 
  return pwd.Left( pwd.GetLength() - 5 );
}

/// sets the password sent by the softphone
/// when sending authentication and authorization
/// responses
void SoftPhoneInterface::SetAuthenticationPassword(
  const OString & authPassword
)
{
  PWaitAndSignal lock( m_ConfigMutex );
  m_SoftPhone->SetRegistrarPassword( authPassword );
  OString pwd = PBase64::Encode( authPassword + "-----" );
  m_Config->SetString( "Account", "AuthenticationPassword", pwd );
}

/// returns the user name sent by the softphone
/// when sending authentication and authorization
/// responses
OString SoftPhoneInterface::GetProxyAuthenticationUser()const
{
  PWaitAndSignal lock( m_ConfigMutex );
  return (const char *)m_Config->GetString( "Account", "ProxyAuthenticationUser", "" ); 
}

/// sets the user name sent by the softphone
/// when sending authentication and authorization
/// responses
void SoftPhoneInterface::SetProxyAuthenticationUser(
  const OString & authUser
)
{
  PWaitAndSignal lock( m_ConfigMutex );
  m_Config->SetString( "Account", "ProxyAuthenticationUser", authUser );
  m_SoftPhone->SetProxyUserName( authUser );
}

/// returns the password sent by the softphone
/// when sending authentication and authorization
/// responses
OString SoftPhoneInterface::GetProxyAuthenticationPassword()const
{
  PWaitAndSignal lock( m_ConfigMutex );
  OString pwd = PBase64::Decode( m_Config->GetString( "Account", "ProxyAuthenticationPassword", "-----" ) ); 
  return pwd.Left( pwd.GetLength() - 5 );
}

/// sets the password sent by the softphone
/// when sending authentication and authorization
/// responses
void SoftPhoneInterface::SetProxyAuthenticationPassword(
  const OString & authPassword
)
{
  PWaitAndSignal lock( m_ConfigMutex );
  m_SoftPhone->SetProxyPassword( authPassword );
  OString pwd = PBase64::Encode( authPassword + "-----" );
  m_Config->SetString( "Account", "ProxyAuthenticationPassword", pwd );
}

/// sets the STUN server to be used for NAT bindings
NAT::SIPSTUNClient::NatTypes SoftPhoneInterface::SetSTUNServer(
  const OString & server,
  WORD udpPortBase,
  WORD udpPortMax,
  WORD rtpPortBase,
  WORD rtpPortMax
)
{
  PWaitAndSignal lock( m_ConfigMutex );

  if( server.IsEmpty() )
  {
    m_Config->SetString( "Transport", "STUNServer", "" );
    return m_NATType;
  }
  
  if( m_NATType != NAT::SIPSTUNClient::NumNatTypes && server == GetSTUNServerAddress() )
    return m_NATType;

  PStringStream log;
  log << "----------------" << PTimer::Tick() << "----------------\r\n";
  log << "Querying STUN server at " << server << ".\r\nThis may take a while ... \r\n\r\n";
  Event_WritePacketLog( (const char *)log ); 
  m_Config->SetString( "Transport", "STUNServer", server );
  m_NATType = m_SoftPhone->SetSTUNServer( server, udpPortBase, udpPortMax, rtpPortBase, rtpPortMax );

  return m_NATType;
}

    ///  returns the STUN server address to be used for NAT bindings
OString SoftPhoneInterface::GetSTUNServerAddress()const
{
  PWaitAndSignal lock( m_ConfigMutex );
  return (const char *)m_Config->GetString( "Transport", "STUNServer", "" );
}

/// adds a new account to the account database.
BOOL SoftPhoneInterface::AddAccount(
  const OString & accountName,
  const OString & accountAddress,
  const OString & authenticationUser, 
  const OString & authenticationPassword,
  const OString & proxyAuthenticationUser, 
	const OString & proxyAuthenticationPassword, 
  const OString & serverAddress,
  const OString & outboundProxy, 
  BOOL isDefaultAccount, 
  BOOL isAutoLogin
)
{
  PWaitAndSignal lock( m_ConfigMutex );
 
  PStringStream record;
  record << "\"" << accountName << "\" "; 
  record << "<" <<  accountAddress << ">";
  record << ";auth-user=" << authenticationUser;
  record << ";auth-password=" << authenticationPassword;
  record << ";proxy-auth-user=" << proxyAuthenticationUser;
  record << ";proxy-auth-password=" << proxyAuthenticationPassword;
  record << ";sip-server=" << serverAddress;
  record << ";outbound-proxy=" << outboundProxy;
  record << ";is-default-account=" << isDefaultAccount;
  record << ";is-auto-login=" << isAutoLogin;

  PMessageDigest5 digestor;
  PMessageDigest5::Code digest; 
  digestor.Start();
  digestor.Process( accountName + accountAddress );
  digestor.Complete( digest );
  MD5::MD5Hash hashKey( digest );
  OString key = hashKey.AsString();

  if( isDefaultAccount )
  {
    PStringToString accounts = m_Accounts->GetAllKeyValues();
    for( PINDEX i = 0; i < accounts.GetSize(); i++ )
    {
      OString key = accounts.GetKeyAt( i );
      PStringStream account;
      account << "From: ";
      account << accounts.GetDataAt( i );
      From from( (const char *)account );
      from.SetParameter( "is-default-account" , "0" );
      m_Accounts->SetString( key, from.GetHeaderBody() );
    }
  }

  m_Accounts->SetString( key, record );

  return TRUE;
}


/// returns the size of the account database
PINDEX SoftPhoneInterface::GetAccountSize()const
{
  PWaitAndSignal lock( m_ConfigMutex );
  PStringToString accounts = m_Accounts->GetAllKeyValues();
  return accounts.GetSize();
}

/// returns an account using its index in the 
/// account database
BOOL SoftPhoneInterface::GetAccountByIndex(
  PINDEX index,
  OString & key, 
  OString & accountName, 
  OString & accountAddress, 
  OString & serverAddress, 
  OString & outboundProxy, 
  OString & authenticationUser, 
  OString & authenticationPassword,
  OString & proxyAuthenticationUser, 
	OString & proxyAuthenticationPassword, 
  BOOL & isDefaultAccount, 
  BOOL & isAutoLogin
)const
{
  PWaitAndSignal lock( m_ConfigMutex );
  PStringToString accounts = m_Accounts->GetAllKeyValues();
  key = (const char *)accounts.GetKeyAt( index );
  return GetAccountByKey( 
    key, 
    accountName, 
    accountAddress, 
    serverAddress, 
    outboundProxy, 
    authenticationUser, 
    authenticationPassword, 
    proxyAuthenticationUser, 
	  proxyAuthenticationPassword,
    isDefaultAccount, 
    isAutoLogin );
}

/// returns an account using its key in the 
/// account database which might have been known
/// by the application from previous call to GetAccountByIndex
BOOL SoftPhoneInterface::GetAccountByKey(
  const OString & key, 
  OString & accountName, 
  OString & accountAddress, 
  OString & serverAddress, 
  OString & outboundProxy, 
  OString & authenticationUser, 
  OString & authenticationPassword, 
  OString & proxyAuthenticationUser, 
	OString & proxyAuthenticationPassword, 
  BOOL & isDefaultAccount, 
  BOOL & isAutoLogin
)const
{
  PWaitAndSignal lock( m_ConfigMutex );
  OString account = m_Accounts->GetString( key, "" );

  if( !account.IsEmpty() )
  {
    PStringStream record;
    record << "From: " << account;
    From from((const char *)record);

    accountName = from.GetDisplayName();
    accountName = ParserTools::UnQuote( accountName );
    accountAddress = from.GetURI().AsString();
    from.GetParameter( "auth-user", authenticationUser );
    from.GetParameter( "auth-password", authenticationPassword );
    from.GetParameter( "proxy-auth-user", proxyAuthenticationUser );
    from.GetParameter( "proxy-auth-password", proxyAuthenticationPassword );
    from.GetParameter( "sip-server", serverAddress );
    from.GetParameter( "outbound-proxy", outboundProxy );

    OString defAccount;
    from.GetParameter( "is-default-account", defAccount );
    isDefaultAccount = (BOOL)defAccount.AsInteger();

    OString isAuto;
    from.GetParameter( "is-auto-login", isAuto );
    isAutoLogin = (BOOL)isAuto.AsInteger();
    
    return TRUE;
  }

  return FALSE;
  
}

/// deletes an account using its key in the 
/// account database which might have been known
/// by the application from previous call to GetAccountByIndex
BOOL SoftPhoneInterface::DeleteAccountByKey(
  const OString & key
)
{
  PWaitAndSignal lock( m_ConfigMutex );
  m_Accounts->DeleteKey( key );
  return TRUE;
}

/// deletes an account using its index in the 
/// account database
BOOL SoftPhoneInterface::DeleteAccountByIndex(
  PINDEX index
)
{
  PWaitAndSignal lock( m_ConfigMutex );
  PStringToString accounts = m_Accounts->GetAllKeyValues();
  OString key = accounts.GetKeyAt( index );
  return DeleteAccountByKey( key );
}


/// returns the detected record devices in a string array
/// (microphones)
PStringList SoftPhoneInterface::PopAvailableRecordDevices()const
{
  return PSoundChannel::GetDeviceNames( PSoundChannel::Recorder );
}

/// returns the detected play devices in a string array
/// (speakers)
PStringList SoftPhoneInterface::PopAvailablePlayDevices()const
{
  return PSoundChannel::GetDeviceNames( PSoundChannel::Player );
}

/// returns the detected video input devices in a string array
/// (video camera)
PStringList SoftPhoneInterface::PopAvailableVideoInputDevices()const
{
  return PVideoInputDevice::GetDriversDeviceNames("*");
}

/// returns the available codecs in a string array
PStringArray SoftPhoneInterface::PopAvailableAudioCodecs()
{
  PStringArray formats = m_SoftPhone->GetMediaFormats();;
  PStringArray final;

  for( PINDEX i = 0; i < formats.GetSize(); i++ )
  {
    if( formats[i].Find( "H.26" ) == P_MAX_INDEX )
      final.AppendString( formats[i] );
  }

  return final;
}

PStringToString SoftPhoneInterface::PopCallHistory(PINDEX count)
{
  PWaitAndSignal lock( m_ConfigMutex );

  PStringToString history;

  if( m_CallHistory == NULL )
    return history;

  PStringToString keys = m_CallHistory->GetAllKeyValues();
  
  count = count > keys.GetSize() ? keys.GetSize() : count;
  PINDEX j = keys.GetSize() - 1;


  for( PINDEX i = count; i > 0; i-- )
  {
    OString time = keys.GetKeyAt( j );
    OString uri = keys.GetDataAt( j-- );
    history.SetAt( time, uri );
  }

  
  return history;
}


/// returns the name of the active microphone
/// device 
OString SoftPhoneInterface::GetDefaultMicrophone()const
{
  PWaitAndSignal lock( m_ConfigMutex );
  return (const char *)m_Config->GetString( "Audio", "Microphone", PSoundChannel::GetDefaultDevice( PSoundChannel::Recorder ) ); 
}

/// sets the name of the active microphone
/// device 
void SoftPhoneInterface::SetDefaultMicrophone(
  const OString & deviceName
)
{
  PWaitAndSignal lock( m_ConfigMutex );
  m_Config->SetString( "Audio", "Microphone", deviceName );
  m_SoftPhone->SetRecordDevice( deviceName );
  /// Reinitialize the PC Sound System
  m_SoftPhone->Initialise();
}

/// returns the name of the active speaker device
OString SoftPhoneInterface::GetDefaultSpeaker()const
{
  PWaitAndSignal lock( m_ConfigMutex );
  return (const char *)m_Config->GetString( "Audio", "Speaker", PSoundChannel::GetDefaultDevice( PSoundChannel::Player ) ); 
}

void SoftPhoneInterface::SetDefaultSpeaker(
  const OString & deviceName
)
{
  PWaitAndSignal lock( m_ConfigMutex );
  m_Config->SetString( "Audio", "Speaker", deviceName );
  m_SoftPhone->SetPlayDevice( deviceName );
  /// Reinitialize the PC Sound System
  m_SoftPhone->Initialise();
}

/// Returns the name of the active ring device
OString SoftPhoneInterface::GetDefaultRingDevice()const
{
  PWaitAndSignal lock( m_ConfigMutex );
  return (const char *)m_Config->GetString( "Audio", "Ring", PSoundChannel::GetDefaultDevice( PSoundChannel::Player ) ); 
}

void SoftPhoneInterface::SetDefaultRingDevice(
  const OString & deviceName
)
{
  PWaitAndSignal lock( m_ConfigMutex );
  m_Config->SetString( "Audio", "Ring", deviceName );
}

void SoftPhoneInterface::SetVolume( 
  unsigned volume,
  const BOOL isPlayerDevice
)
{
  if ( volume > 100 )
    volume = 100;

  if ( m_SoftPhone )
  {
    if( isPlayerDevice )
      m_SoftPhone->SetPlayDeviceVolume( volume );
    else
      m_SoftPhone->SetRecordDeviceVolume( volume );
  }
}

void SoftPhoneInterface::GetVolume( 
  unsigned & volume,
  const BOOL isPlayerDevice
)
{ 
  volume = 0;

  if ( m_SoftPhone )
  {
    if( isPlayerDevice )
      m_SoftPhone->GetPlayDeviceVolume( volume );
    else
      m_SoftPhone->GetRecordDeviceVolume( volume );
  }
}


/// returns the codec based on its slot number.
/// slot == 1 means highest priority in the codec list 
OString SoftPhoneInterface::GetAudioCodec( 
  PINDEX slot
)const
{
  PWaitAndSignal lock( m_ConfigMutex );
  PStringStream  codecSlot;
  codecSlot << "Codec" << slot;
  return (const char *)m_Config->GetString( "Audio", codecSlot, "" ); 
}

/// sets the codec based on its slot number.
/// slot == 1 means highest priority in the codec list. 
void SoftPhoneInterface::SetAudioCodec(
  PINDEX slot,
  const OString & codecName
)
{
  PWaitAndSignal lock( m_ConfigMutex );
  PStringStream  codecSlot;
  codecSlot << "Codec" << slot;
  m_Config->SetString( "Audio", codecSlot, codecName );
}

/// Plays an audible dtmf tone on the softphone's default audio player device.
/// Take note that this will not send a DTMF tone over the wire.
/// Use SendRFC2833Tone instead.
void SoftPhoneInterface::PlayDTMF(
  const int tone
)
{
  PSoundChannel playChannel( GetDefaultSpeaker(), PSoundChannel::Player );
  PlayDTMF( tone, playChannel );
}

/// Plays an audible dtmf tone on a specified sound channel.
/// Take note that this will not send a DTMF tone over the wire.
/// Use SendRFC2833Tone instead.
void SoftPhoneInterface::PlayDTMF(
  const int tone,
  PSoundChannel & playChannel
)
{
  if ( playChannel.IsOpen() )
  {
    switch( tone )
    {
      case 0:
        MS::DTMFTone_0.PlaySound( playChannel );
        break;
      case 1:
        MS::DTMFTone_1.PlaySound( playChannel );
        break;
      case 2:
        MS::DTMFTone_2.PlaySound( playChannel );
        break;
      case 3:
        MS::DTMFTone_3.PlaySound( playChannel );
        break;
      case 4:
        MS::DTMFTone_4.PlaySound( playChannel );
        break;
      case 5:
        MS::DTMFTone_5.PlaySound( playChannel );
        break;
      case 6:
        MS::DTMFTone_6.PlaySound( playChannel );
        break;
      case 7:
        MS::DTMFTone_7.PlaySound( playChannel );
        break;
      case 8:
        MS::DTMFTone_8.PlaySound( playChannel );
        break;
      case 9:
        MS::DTMFTone_9.PlaySound( playChannel );
        break;
      case 10:
        MS::DTMFTone_10.PlaySound( playChannel );
        break;
      case 11:
        MS::DTMFTone_11.PlaySound( playChannel );
        break;
    }
  }
}

/// Plays an audible dtmf tone on the softphone's default audio player device.
/// Take note that this will not send a DTMF tone over the wire.
/// Use SendRFC2833Tone instead.
void SoftPhoneInterface::PlayDTMF(
  const OString & tone
)
{
  PSoundChannel playChannel( GetDefaultSpeaker(), PSoundChannel::Player );
  PlayDTMF( tone, playChannel );
}


/// Plays an audible dtmf tone on a specified sound channel.
/// Take not that this will not send a DTMF tone over the wire.
/// Use SendRFC2833Tone instead.
void SoftPhoneInterface::PlayDTMF(
  const OString & tone,
  PSoundChannel & playChannel
)
{
  if ( playChannel.IsOpen() )
  {
    for( PINDEX i = 0; i < tone.GetLength(); i++ )
    {
      char c = tone[i];
      switch( c )
      {
        case '0':
          MS::DTMFTone_0.PlaySound( playChannel );
          break;
        case '1':
          MS::DTMFTone_1.PlaySound( playChannel );
          break;
        case '2':
          MS::DTMFTone_2.PlaySound( playChannel );
          break;
        case '3':
          MS::DTMFTone_3.PlaySound( playChannel );
          break;
        case '4':
          MS::DTMFTone_4.PlaySound( playChannel );
          break;
        case '5':
          MS::DTMFTone_5.PlaySound( playChannel );
          break;
        case '6':
          MS::DTMFTone_6.PlaySound( playChannel );
          break;
        case '7':
          MS::DTMFTone_7.PlaySound( playChannel );
          break;
        case '8':
          MS::DTMFTone_8.PlaySound( playChannel );
          break;
        case '9':
          MS::DTMFTone_9.PlaySound( playChannel );
          break;
        case '*':
          MS::DTMFTone_10.PlaySound( playChannel );
          break;
        case '#':
          MS::DTMFTone_11.PlaySound( playChannel );
          break;
      }
    }
  }
}

/// makes an outbound call to this address.
/// address should contain a valid SIP URI
void SoftPhoneInterface::MakeCall( 
  const OString & _address
)
{
  PString address = _address.c_str();
  if( WillRequireRegistration() && !m_SoftPhone->GetSIPEndPoint()->IsRegistered() )
  {
    PStringStream log;
    log << "----------------" << PTimer::Tick() << "----------------\r\n";
    log << "Outbound call is not permitted while status is Offline\r\n";
    Event_WritePacketLog( (const char *)log );
    Event_CallDisconnected( "400 Call Not Allowed, " + OString((const char *)address) );
    return;
  }

  OString dest = address;

  if (address.IsEmpty())
  {
    Event_CallDisconnected( "400 Call Not Allowed, " + OString((const char *)address) );
    return;
  }

  //check if this it has a host part
  //if not append a domain
  if( address.Find( "@" ) == P_MAX_INDEX )
  {
    
    PRegularExpression regex("[0-9]*[.][0-9]*[.][0-9]*[.][0-9]*");
    PINDEX pos,len;
    if( !address.FindRegEx(regex, pos, len)  ) 
    {
      /// not in ip address notation
      SIPURI uriAddress( (const char *)address );
      PIPSocket::Address ip;
      WORD port = 0;
      if( ParserTools::IsNumeric( (const char *)address ) || !SIPTransport::Resolve( uriAddress, ip, port ) )
      {
        /// not an ip address
        /// must be a telephone number
        /// or an alias

        
        if( !m_SoftPhone->GetSIPDomain().IsEmpty() )
        {
          SIPURI destURI((const char *)m_SoftPhone->GetSIPDomain());
          destURI.SetUser( (const char *)address );
          dest = destURI.AsString();
        }else if( !m_SoftPhone->GetRegistrarAddress().IsEmpty() )
        {
          /// get the domain part of the registrar

          SIPURI destURI((const char *)m_SoftPhone->GetRegistrarAddress());
          destURI.SetUser( (const char *)address );
          dest = destURI.AsString();
        }else if( !m_SoftPhone->GetProxyAddress().IsEmpty() )
        {
          SIPURI destURI((const char *)m_SoftPhone->GetProxyAddress());
          destURI.SetUser( (const char *)address );
          dest = destURI.AsString();
        }
      }
    }
  }

  if( dest.ToLower().Left( 4 ) != "sip:" )
    dest = "sip:" + OString((const char *)address);

  PTime now;
  To to;
  to.SetURI( dest );
  m_CallHistory->SetString( now.AsString(), to.AsString() );

  m_SoftPhone->InitializeSIP();
  m_SoftPhone->MakeCall( address );
}

/// hangup the current call
void SoftPhoneInterface::HangupCall()
{
  m_SoftPhone->HangUpCall();
}

/// answers the current call
void SoftPhoneInterface::AnswerCall()
{
  m_SoftPhone->AnswerCall();
}

/// Sends a plain text MESSAGE to remoteAddress.
/// remoteAddress should be a valid SIP URI.
void SoftPhoneInterface::SendInstantMessage(
  const OString & remoteAddress, 
  const OString & message
)
{
  m_SoftPhone->InitializeSIP();
  m_SoftPhone->SendInstantMessage( remoteAddress, message );
}

/// Returns the call history inside a string array.
/// count contains the number of calls to be returned
PStringToString SoftPhoneInterface::GetCallHistory( 
  PINDEX count 
)const
{
  PWaitAndSignal lock( m_ConfigMutex );
  PStringToString history;
  PStringToString keys = m_CallHistory->GetAllKeyValues();

  count = count > keys.GetSize() ? keys.GetSize() : count;
  PINDEX j = keys.GetSize() - 1;

  for( PINDEX i = count; i > 0; i-- )
  {
    PStringStream strm;
    history.SetAt( keys.GetKeyAt( j ),  keys.GetDataAt( j ) );
    j--;
  }

  return history;
}


/// Sends an RFC 2833 DTMF tone over the wire
void SoftPhoneInterface::SendRFC2833Tone(
  const OString & userInput, 
  int duration
)
{

  if( duration < 100 )
    duration = 100;

  for( PINDEX i = 0; i < userInput.GetLength(); i++ )
  {
    if( userInput[i] == ',' )
    {
      PThread::Sleep( 100 );
      continue;
    }

    m_SoftPhone->SendUserInput( userInput[i], duration, TRUE );
    PThread::Sleep( duration );
  }
}

/// Sends a SIP INFO DTMF tone over the wire
void SoftPhoneInterface::SendINFOTone(
  const OString & userInput, 
  int duration
)
{
  if( duration < 100 )
    duration = 100;

  for( PINDEX i = 0; i < userInput.GetLength(); i++ )
  {
    if( userInput[i] == ',' )
    {
      PThread::Sleep( 100 );
      continue;
    }

    m_SoftPhone->SendUserInput( userInput[i], duration, FALSE );
    PThread::Sleep( duration );
  }
}

/// sends a DTMF tone.  Default behavior uses RFC2833
void SoftPhoneInterface::SendDTMF(
  const OString & tone, 
  int duration
)
{
  if( IsRFC2833Enabled() )
    SendRFC2833Tone( tone, duration );
  else
    SendINFOTone( tone, duration );
}



/// sets RFC 2833 as the default method for sending DTMF
void SoftPhoneInterface::SetEnableRFC2833( BOOL enable )
{
  PWaitAndSignal lock( m_ConfigMutex );
  m_Config->SetBoolean( "Audio", "DTMFAsRFC2833", enable );
}

/// check if RFC 2833 is the default method for sending DTMF
BOOL SoftPhoneInterface::IsRFC2833Enabled()const
{
  PWaitAndSignal lock( m_ConfigMutex );
  return m_Config->GetBoolean( "Audio", "DTMFAsRFC2833", TRUE );
}


/// enable hashing of RTP and SIP over the wire
/// take note that this is a none standard mechanism.
/// this will not work for servers other than opensbc.
/// Default encryption is XOR with "GS" key
void SoftPhoneInterface::SetEnableEncryption( 
  BOOL enable,
  Encryption::Engine::Type type, 
  const OString & key
)
{
  PWaitAndSignal lock( m_ConfigMutex );
  m_Config->SetBoolean( "Account", "Encrypted", enable );
  m_SoftPhone->SetEncryption( enable );

  if ( enable )
  {
    Encryption::Engine::m_Type  = type;
    Encryption::Engine::m_Key   = key;
  }
}

/// check if xor hashing for RTP and SIP is enabled
BOOL SoftPhoneInterface::IsEncryptionEnabled()const
{
  PWaitAndSignal lock( m_ConfigMutex );
  return m_Config->GetBoolean( "Account", "Encrypted", FALSE );
}

/// enable Privacy (RFC3325)
void SoftPhoneInterface::SetEnablePrivacy( BOOL enable )
{
  PWaitAndSignal lock( m_ConfigMutex );
  m_Config->SetBoolean( "Account", "Privacy", enable );
  m_SoftPhone->SetPrivacy( enable );
}

/// check if Privacy (RFC3325) is enabled
BOOL SoftPhoneInterface::IsPrivacyEnabled()const
{
  PWaitAndSignal lock( m_ConfigMutex );
  return m_Config->GetBoolean( "Account", "Privacy", FALSE );
}

/// Set the maximum registration keep-alive interval
void SoftPhoneInterface::SetRegistrarTTL( 
  const PTimeInterval & ttl 
)
{
  PWaitAndSignal lock( m_ConfigMutex );
  m_Config->SetInteger( "Account", "RegistrarTTL", ttl.GetSeconds() );
  m_SoftPhone->SetRegistrarTTL( ttl.GetSeconds() );
}

/// Returns the maximum registration keep alive interval
PTimeInterval SoftPhoneInterface::GetRegistrarTTL()const
{
  PWaitAndSignal lock( m_ConfigMutex );
  int seconds = m_Config->GetInteger( "Account", "RegistrarTTL", 3600 );
  PTimeInterval ttl( 0, seconds );
  return  ttl;
}

void SoftPhoneInterface::SetRequireRegistration(
  BOOL require
)
{
  PWaitAndSignal lock( m_ConfigMutex );
  m_Config->SetBoolean( "Account", "RequireRegistration", require );
}

BOOL SoftPhoneInterface::WillRequireRegistration()const
{
  PWaitAndSignal lock( m_ConfigMutex );
  return m_Config->GetBoolean( "Account", "RequireRegistration", FALSE );
}

/// Returns the interface address used by the softphone
OString SoftPhoneInterface::GetInterfaceAddress()const
{
  PWaitAndSignal lock( m_ConfigMutex );
  return (const char *)m_Config->GetString( "Transport", "InterfaceAddress", "sip:*:5060" );
}


/// Sets the interface address the softphone will bind to
void SoftPhoneInterface::SetInterfaceAddress( 
  const OString & iface 
)
{
  PWaitAndSignal lock( m_ConfigMutex );
  m_Config->SetString( "Transport", "InterfaceAddress", iface );
  m_SoftPhone->SetInterfaceAddress( iface );
}

void SoftPhoneInterface::WatchRegistrationState(
  const OString & stateAgent,
  const OString & resourceURI,
  const PTimeInterval & expires,
  const OString & authUser,
  const OString & authPassword
  
)
{
  m_SoftPhone->WatchRegistrationState( stateAgent, resourceURI, expires, authUser, authPassword );
}

void SoftPhoneInterface::StopWatchingRegistrationState(
  const OString & resourceURI
)
{
  m_SoftPhone->StopWatchingRegistrationState( resourceURI );
}

void SoftPhoneInterface::_OnWritePacketLog( PObject &, INT _log )
{
  if( _log == 0 )
    return;

  OString * line = reinterpret_cast<OString*>( _log );
  if( line != NULL )
  {
    OStringStream log;
    log << "----------------" << PTimer::Tick() << "----------------\r\n" << *line;
    Event_WritePacketLog( log.str() );
  }
}

void SoftPhoneInterface::_OnReadPacketLog( PObject & , INT _log )
{
  if( _log == 0 )
    return;

  OString * line = reinterpret_cast<OString*>( _log );
  if( line != NULL )
  {
    OStringStream log;
    log << "----------------" << PTimer::Tick() << "----------------\r\n" << *line;
    Event_ReadPacketLog( log.str() );
  }
}

int SoftPhoneInterface::GetState()const
{ 
  return const_cast<SoftPhoneInterface&>(*this).GetPhoneManager()->GetState(); 
}

/// Worker thread for playing an audible busy sound.
void SoftPhoneInterface::PlayBusyTone( PThread &, INT param )
{
  OString       *device = reinterpret_cast<OString *>( param );
  PSoundChannel *playChannel;

  if ( device == NULL )
  {
    // NULL -> Send audio to the softphone's default audio player device
    playChannel = new PSoundChannel( GetDefaultSpeaker(), PSoundChannel::Player );
  } else
  {
    // Send audio to specified sound device
    playChannel = new PSoundChannel( *device, PSoundChannel::Player );
    
    delete device;  // We don't need this anymore
  }

  if ( playChannel && playChannel->IsOpen() )
  {
    while( m_PlayBusyTone )
      m_BusyToneFile->PlaySound( *playChannel );
  }

  if ( playChannel )
    delete playChannel;
}

/// Plays an audible busy sound on the softphone's default audio player.
void SoftPhoneInterface::PlayBusyTone()
{
  if( !m_PlayBusyTone )
  {
    m_PlayBusyTone = TRUE;
    PThread::Create( PCREATE_NOTIFIER( PlayBusyTone ) );
  }
}

/// Plays an audible busy sound on a specified sound channel.
void SoftPhoneInterface::PlayBusyTone( const OString & device )
{
  OString *param = new OString( device );
  // Note: param is destroyed in PlayBusyTone worker thread

  if( !m_PlayBusyTone )
  {
    m_PlayBusyTone = TRUE;
    PThread::Create( PCREATE_NOTIFIER( PlayBusyTone ), reinterpret_cast<INT>( param ) );
  }
}

void SoftPhoneInterface::StopBusyTone()
{
  m_PlayBusyTone = FALSE;

  if( m_BusyToneFile != NULL )
    m_BusyToneFile->StopPlaying();
}

/// Worker thread for playing an audible dialtone sound.
void SoftPhoneInterface::PlayDialTone( PThread &, INT param )
{
  OString       *device = reinterpret_cast<OString *>( param );
  PSoundChannel *playChannel;

  if ( device == NULL )
  {
    // NULL -> Send audio to the softphone's default audio player device
    playChannel = new PSoundChannel( GetDefaultSpeaker(), PSoundChannel::Player );
  } else
  {
    // Send audio to specified sound device
    playChannel = new PSoundChannel( *device, PSoundChannel::Player );
        
    delete device;  // We don't need this anymore
  }

  m_DialToneFile->StopPlaying();
  if ( playChannel && playChannel->IsOpen() )
  {
    while( m_PlayDialTone )
      m_DialToneFile->PlaySound( *playChannel );
  }

  if ( playChannel )
    delete playChannel;
}

/// Plays an audible dialtone sound on the softphone's default audio player.
void SoftPhoneInterface::PlayDialTone()
{
  if( !m_PlayDialTone )
  {
    m_PlayDialTone = TRUE;
    PThread::Create( PCREATE_NOTIFIER( PlayDialTone ) );
  }
}

/// Plays an audible dialtone sound on a specified sound channel.
void SoftPhoneInterface::PlayDialTone( const OString & device )
{
  OString *param = new OString( device );
  // Note: param is destroyed in PlayDialTone worker thread

  if( !m_PlayDialTone )
  {
    m_PlayDialTone = TRUE;
    PThread::Create( PCREATE_NOTIFIER( PlayDialTone ), reinterpret_cast<INT>( param ) );
  }
}

void SoftPhoneInterface::StopDialTone()
{
  m_PlayDialTone = FALSE;
  if( m_DialToneFile != NULL )
    m_DialToneFile->StopPlaying();
}

/// Worker thread for playing an audible ringback sound.
void SoftPhoneInterface::PlayRingBackTone( PThread &, INT param )
{
  OString       *device = reinterpret_cast<OString *>( param );
  PSoundChannel *playChannel;

  if ( device == NULL )
  {
    // NULL -> Send audio to the softphone's default audio player device
    playChannel = new PSoundChannel( GetDefaultSpeaker(), PSoundChannel::Player );
  } else
  {
    // Send audio to specified sound device
    playChannel = new PSoundChannel( *device, PSoundChannel::Player );
    
    delete device;  // We don't need this anymore
  }
  
  if ( playChannel && playChannel->IsOpen() )
  {
    while( m_PlayRingBackTone )
      m_RingBackToneFile->PlaySound( *playChannel );
  }

  if ( playChannel )
    delete playChannel;
}

/// Plays audible ringback tone on the softphone's default audio player.
void SoftPhoneInterface::PlayRingBackTone()
{
  if( !m_PlayRingBackTone )
  {
    m_PlayRingBackTone = TRUE;
    PThread::Create( PCREATE_NOTIFIER( PlayRingBackTone ) );
  }
}

/// Plays audible ringback tone on a specified sound channel.
void SoftPhoneInterface::PlayRingBackTone( const OString & device )
{
  OString *param = new OString( device );
  // Note: param is destroyed in PlayRingBackTone worker thread

  if( !m_PlayRingBackTone )
  {
    m_PlayRingBackTone = TRUE;
    PThread::Create( PCREATE_NOTIFIER( PlayRingBackTone ), reinterpret_cast<INT>( param ) );
  }
}

void SoftPhoneInterface::StopRingBackTone()
{
  m_PlayRingBackTone = FALSE;
  if( m_RingBackToneFile != NULL )
    m_RingBackToneFile->StopPlaying();
}

double SoftPhoneInterface::RPCGetAccountBalance(
    const OString & server,
    const OString & account
)
{
  return m_SoftPhone->RPCGetAccountBalance( server, account );
}

void SoftPhoneInterface::DisableSilenceDetection()
{
  m_SoftPhone->SetSilenceDetectParams(
    OpalSilenceDetector::Params( OpalSilenceDetector::NoSilenceDetection ) );
}

void SoftPhoneInterface::EnableAdaptiveSilenceDetection(
  unsigned adaptivePeriod,
  unsigned signalDeadband,
  unsigned silenceDeadband
)
{
  m_SoftPhone->SetSilenceDetectParams(
    OpalSilenceDetector::Params( OpalSilenceDetector::AdaptiveSilenceDetection,
                                 0,
                                 signalDeadband,
                                 silenceDeadband,
                                 adaptivePeriod ) 
  );
}

void SoftPhoneInterface::EnableFixedSilenceDetection(
  unsigned threshold,
  unsigned signalDeadband,
  unsigned silenceDeadband
)
{
  m_SoftPhone->SetSilenceDetectParams(
    OpalSilenceDetector::Params( OpalSilenceDetector::FixedSilenceDetection,
                                 threshold,
                                 signalDeadband,
                                 silenceDeadband )
  );
}

void SoftPhoneInterface::SetAudioJitterDelay(
  unsigned minDelay, 
  unsigned maxDelay
)
{
  m_SoftPhone->SetAudioJitterDelay( minDelay, maxDelay );
}
































